/* 
##################################################################
 Copyright (c) 2004 PPARC, CERN, on behalf of the EU EGEE project.

 Creates the Catalogs' tables for Oracle

 author: Krzysztof Nienartowicz  <krzysztof.nienartowicz@cern.ch>
			

 version: $Id: SchemaOracleFM.sql,v 1.14 2005/10/14 15:08:33 knienart Exp $

 This version assumes NTFS-like ACL logic. No hierarchical checks
 are done to determine acccess. only dir/node level acl check is done.
 
 This file is to be used from sqlplus.
   
##################################################################

$Revision: 1.14 $
$Name: glite-data-catalog-service-fr_R_2_1_15 $

*/

--DROP TABLE fc_file CASCADE CONSTRAINTS;

CREATE TABLE fc_file (						
			parent_id integer DEFAULT 0 not null,		-- could be eventually 0 for root or null
			file_id integer not null,
			name varchar2(1024) not null,					-- could be autogen for dangling guids, parent_id is predefined then too
			TYPE char(1) not null,							-- dir, file, symlink, virtualdir: d,f,s,z,v, could be constraint'ed
			guid varchar2(36) not null,  					-- depending on the symlink implementation, zeroed if null			
			cTime timestamp(0) default systimestamp not null,   --or taken from the serivce
			mTime timestamp(0) default systimestamp not null,
			file_size decimal(21) default 0 not null,	--or rather from rc
			lfn_lifetime integer default 0 not null,--required existence in seconds,starting from time_of_creation, 0 if to be deleted manually
			file_checksum CHAR(32) default '0' NOT NULL ,-- md5 as described in rfc1321,was DEFAULT 00000000000000000000000000000000
			status decimal(1) default 0 not null,
			principal_id integer not null,  				-- assumption we pass principal_id to the qry so no star join,weak assoc as principals may disappear, also means each principal has its path.
			group_id integer not null,  					--refrences principal_id from principal tbl
			userperm decimal(3) not null,					-- [r|w|x|l|d|p|sm|gm] - 8 bits vector, up to 256dec, always exist 3x per guid
			groupperm decimal(3) not null,				-- [r|w|x|l|d|p|sm|gm] - 8 bits vector, up to 256dec, always exist 3x per guid
			otherperm decimal(3) not null,				-- [r|w|x|l|d|p|sm|gm] - 8 bits vector, up to 256dec, always exist 3x per guid
			dangling_ok	char(1) default '0' not null, -- lfn not required for such guids so dont check for dangling refs, means we allow lfnless operations			
      data varchar2(4000),	                     -- link path, vdir statement, metadata(?)
			CONSTRAINT pck_file_constraint PRIMARY KEY (parent_id,file_id), -- for fast lookup. guids could be in pk too
			CONSTRAINT uq_guid unique (guid),
			CONSTRAINT uq_name_parent unique(name, parent_id)			
			)			
			organization INDEX
			COMPRESS 1
			including data
			overflow;

/* create unique index fc_perms_idx on fc_file(guid,principal_id,group_id) */
/* ALTER TABLE fc_file MOVE MAPPING TABLE
CREATE BITMAP INDEX idx_fc_bitmap_type ON fc_file(TYPE) */


create or replace synonym fc for fc_file;

/*******************************************************
*	replica table
*   ---------------- Not used in Fireman. ------------------
*	
*	1-n guid-surl mapping.
*	Permissions per guid.
*	This should be in a 3rd form but compression on guid in iot should work in a similar way
*  Later on we may think of integrating permissions into main rc/fc table, but 
* 	this would require coming back to 3rd form as permission values would have to be repeated.
*
* Note:
*  We could reuse file_id if we knew tables stay always in the same server/schema 
*  and there's no orthogonal access from both fc_file and rc_replica
*  Otherwise we want to avoid joins whenever possible and pay the storage/index size price
*
*  FM notice: changed to hold bp values for faster insert/lookup - no more fas dependency.
*  still uses acl to store acls.
********************************************************/


--DROP TABLE rc_replica CASCADE CONSTRAINTS ;
CREATE TABLE rc_replica (
			guid varchar2(36) not null,  	
			file_checksum CHAR(32) NOT NULL,	-- md5 as described in rfc1321,was DEFAULT '00000000000000000000000000000000' 
			status decimal(1) default 0 not null,			
			file_size decimal(21) default 0 not null,			
			principal_id integer not null,  -- assumption we pass principal_id to the qry so no star join,weak assoc as principals may disappear, also means each principal has its path.
			group_id integer not null,  --refrences principal_id from principal tbl
			userperm decimal(3) not null,	-- [r|w|x|l|d|p|sm|gm] - 8 bits vector, up to 256dec, always exist 3x per guid
			groupperm decimal(3) not null,	-- [r|w|x|l|d|p|sm|gm] - 8 bits vector, up to 256dec, always exist 3x per guid
			otherperm decimal(3) not null,	-- [r|w|x|l|d|p|sm|gm] - 8 bits vector, up to 256dec, always exist 3x per guid
			dangling_ok	char(1) default '0' not null, -- lfn not required for such guids so dont check for dangling refs, means we allow lfnless operations			
			cTime timestamp(0) default systimestamp , --or taken from the serivce or stored in fc, used only for dangling
			mTime timestamp(0) default systimestamp ,	
			constraint pck_replica_constraint primary key (guid)
			)
			organization INDEX
			;

create or replace synonym rcg for rc_replica ;


/* rc_surl table */
--drop table rc_surl cascade constraints;

create table rc_surl (
	guid varchar2(36) not null,
	surl varchar2(1024) not null,			
	master_replica decimal(1) default 0 not NULL, --only one permitted per guid	
	surl_lifetime integer default 0 not null,	-- required existence in seconds,starting from time_of_creation, 0 if to be deleted manually			
	se varchar2(255) not null, --should be 255 according to posix -- to be index autogen'ed
	cTime  timestamp(0) default systimestamp not null,
	mTime timestamp(0) default systimestamp not null,			
	constraint pck_surl_constraint primary key(guid,surl)
	)
	organization index
	compress 1
	;
  
create unique index idx_uq_surlse on rc_surl(surl) 
  ;

create or replace synonym rcs for rc_surl ;

/***************************************************** 
			principal table
	used almost always for dn-> internal id lookup 
*****************************************************/

--DROP TABLE perm_principal CASCADE CONSTRAINTS;


				
CREATE TABLE perm_principal (
            principal_id integer not null,
            name varchar2(1024) not null,	-- cert dn
            p_type char(1), 				-- user/group
            creation_time timestamp(0) default systimestamp not null,
            CONSTRAINT pck_principaluq_constraint PRIMARY KEY (principal_id),
            CONSTRAINT uq_principal UNIQUE(name)
        )
		organization INDEX;
				



-- dependency holder for the def table
create OR REPLACE type ACLEntryO as object
	(
	principal varchar2(1024),
	perm SMALLINT	
	)
	;
/

create OR REPLACE type ACLList as TABLE of ACLEntryO;
/

create or replace type BasicPermissionO as object
	(
	userName varchar2(250),
	groupName varchar2(250),
	userPerm smallint,
	groupPerm smallint,
	otherPerm smallint 
	)
	not final
	;
/

create or replace type PermissionO under BasicPermissionO
	(
		acl ACLList
	);
/

/* side table 1:1 relation to keep defaults/policies for user/group/id, 
	used rarely 
*/
create table perm_principal_defaults (
				principal_id integer,
				guid varchar2(36) not null,	-- for FAS checks - only permitted may change it (at fas)
				defaultCreateLFN char(1) default '0',
				defaultCreateLFNPattern varchar2(1024),
				defaultParent integer not null,
				perm PermissionO,
				CONSTRAINT pck_def_constraint PRIMARY KEY (principal_id) )
				organization index
				nested table perm.acl store as nested_acl_default_tbl
				;


alter table perm_principal_defaults
ADD CONSTRAINT fk_principal_defaults FOREIGN KEY (principal_id) REFERENCES perm_principal(principal_id)
		ON DELETE CASCADE	;
		

/* CREATE bitmap INDEX idx_principal_bitmap_type ON principal(p_type); */


/*****************************************************
* 			perm_basic_permission table 
*
* ---------- 'almost' NOT USED IN FIREMAN -----------
*
*	Stores basic, ux like permissions for guids
* 	keyd by guid - big ut allows detachement from fc/rc.
*  There must be [always] an entry for fc/rc guids.
*
* Note:
* 	it could/should be subtable of rc_replica/fc_file as 
* 	it these two bind guids to permissions. 
*	we probably don't need reversed lookup on principal
*	so nesting would do. 
* 	problem with porting and scattering pemission data
* 	over two possible separated/distributed tables, 
*	making fas hard to optimise.
* the same renders impossible to use foreign keys on guid.
*
* FM note:
* 	only external (FAS) entries are sotred in BP, the rest goes to rc
*
******************************************************/

--DROP TABLE perm_basic_permission cascade constraints ;

create TABLE perm_basic_permission (
		guid varchar2(36) not null,  	--omni-key
		principal_id integer not null,  -- assumption we pass principal_id to the qry so no star join,weak assoc as principals may disappear, also means each principal has its path.
		group_id integer not null,  --refrences principal_id from principal tbl
		userperm decimal(3) not null,	-- [r|w|x|l|d|p|sm|gm] - 8 bits vector, up to 256dec, always exist 3x per guid
		groupperm decimal(3) not null,	-- [r|w|x|l|d|p|sm|gm] - 8 bits vector, up to 256dec, always exist 3x per guid
		otherperm decimal(3) not null,	-- [r|w|x|l|d|p|sm|gm] - 8 bits vector, up to 256dec, always exist 3x per guid
		CONSTRAINT pck_basic_perm_constraint PRIMARY KEY (guid, principal_id, group_id)
		)
		organization index          
		;

--we must have uniqueness create unique index uq_guid on perm_basic_permission(guid)		
--ALTER TABLE perm_basic_permission	
--	ADD CONSTRAINT fk_basic_perm_principal FOREIGN KEY (principal_id) REFERENCES perm_principal(principal_id)
--		ON DELETE CASCADE
--	ADD CONSTRAINT fk_basic_perm_group FOREIGN KEY (group_id) REFERENCES perm_principal(principal_id)
--		ON DELETE CASCADE

/* synonyms */
create or replace synonym bp for perm_basic_permission 
;

create or replace synonym acl for perm_acl_permission 
;

create or replace synonym pp for perm_principal 
;

create or replace synonym pd for perm_principal_defaults
;



--	
/*****************************************************
* 			perm_acl_permission table 
*	
*	Stores extended, optional permissions for guids
* 	keyd by guid too. expensive, but separate
*	index/joins would be expensive too.
*
*  It's an optional entry for fc/rc guids.
******************************************************/
--drop table perm_acl_permission ;

create TABLE perm_acl_permission (	
		guid varchar2(36) not null,  	--omni-key
		principal_id integer not null,-- assumption we pass principal_id to the qry so no star join,weak assoc as principals may disappear, also means each principal has its path.						
		perm decimal(3) not null,	    -- [r|w|x|l|d|p|sm|gm] - 8 bits vector, up to 256dec, always exist 3x per guid
		CONSTRAINT pck_acl_perm_constraint PRIMARY KEY (guid, principal_id)
		)
		organization index    
		compress 1
		;
		

/*          FC_SCHEMA_VERSION table
  ###################################################

 Table to know the current version of the schema
 Insert the current version of the schema incremented every script run
*/

DECLARE
BEGIN
execute immediate 'INSERT INTO fm_schema_version(version_tag)
	(SELECT nvl(''$Name: glite-data-catalog-service-fr_R_2_1_15 $'',''head'') from dual)'
	;
EXCEPTION
WHEN OTHERS THEN
EXECUTE IMMEDIATE 'CREATE TABLE fm_schema_version (
         	version_tag varchar2(1024),
					creation timestamp(0) default systimestamp not NULL,
					constraint pck_version_constraint primary key(version_tag,creation)
					)
					organization INDEX
					COMPRESS 1';

commit;
END;
/

/* auto-incrementation of the primary keys, should be moved into the code */


/* principal_id==0 is a default one p*/
--DROP SEQUENCE fc_principal_id_seq ;
CREATE SEQUENCE fc_principal_id_seq minvalue 0 START WITH 0 cache 100;


--DROP SEQUENCE fc_file_id_seq ;
CREATE SEQUENCE fc_file_id_seq START WITH 5 cache 10000;



--DROP SEQUENCE fc_perm_id_seq ;
CREATE SEQUENCE fc_perm_id_seq START WITH 1 cache 10000;

create sequence meta_schema_id_seq start with 1 cache 10;

create sequence meta_att_seq_id start with 1 cache 100;


--CREATE OR REPLACE TRIGGER fc_principal_id_auto_inc
--BEFORE INSERT ON perm_principal
--FOR EACH ROW
--WHEN (new.principal_id IS NULL)
--BEGIN
--  SELECT fc_principal_id_seq.nextval
--  INTO:new.principal_id from dual;
--END;

/*
CREATE OR REPLACE TRIGGER fc_file_id_auto_inc
BEFORE INSERT ON fc_file
FOR EACH ROW
WHEN (new.file_id IS NULL)
BEGIN
  SELECT fc_file_id_seq.nextval
  INTO:new.file_id from dual;
END;
*/


CREATE OR REPLACE TRIGGER acl_auto_delete
after delete ON bp
FOR EACH ROW

BEGIN
	delete from acl
		where guid=:old.guid;
		/* we may want some error logic here to avoid general errs for itemized lists..*/
END;
/
CREATE OR REPLACE TRIGGER rcs_auto_delete
after delete ON fc
FOR EACH ROW

BEGIN
	delete from rcs
		where guid=:old.guid;
   delete from acl
      where guid=:old.guid;
		/* we may want some error logic here to avoid general errs for itemized lists..*/
END;
/

/* CREATE OR REPLACE TRIGGER perm_perm_id_auto_inc
BEFORE INSERT ON perm_basic_permission
FOR EACH ROW
WHEN (new.perm_id IS NULL)
BEGIN
  SELECT fc_perm_id_seq.nextval
  INTO:new.perm_id from dual;
END;

*/

/** 	inserting root mount	into fc_file, its ':' and 'tmp' dir (increments seq too!)   [r|w|x|l|d|p|sm|gm] */
insert into fc_file(parent_id,file_id,guid,name, type,principal_id,group_id,	userperm,groupperm,otherperm)
		values (0,1,'00000000-0000-0000-0000-000000000000',':','d',0,0,bin_to_num(1,1,1,1,1,1,1,1),0,bin_to_num(0,1,0,1,0,0,0,1))
;
insert into fc_file(parent_id,file_id,guid,name, type,principal_id,group_id,	userperm,groupperm,otherperm)
		values (1,2,'00000000-0000-0000-0000-000000000001','tmp','d',0,0,0,0,bin_to_num(1,1,1,1,0,0,0,1)) 
;
insert into fc_file(parent_id,file_id,guid,name, type,principal_id,group_id,	userperm,groupperm,otherperm)
		values (1,3,'00000000-0000-0000-0000-000000000002','free','d',0,0,0,0,bin_to_num(1,1,1,1,0,0,1,1))
;

insert into fc_file(parent_id,file_id,guid,name, type,principal_id,group_id,	userperm,groupperm,otherperm)
		values (1,4,'00000000-0000-0000-0000-000000000003','user','d',0,0,255,bin_to_num(0,1,0,1,0,0,0,1),bin_to_num(0,1,0,1,0,0,0,0))
;

--/* hack for testing - creating predefined test directory with open permissions so does not clash with api-java tests with no superuser*/
--insert into fc_file(parent_id,file_id,guid,name, type,principal_id,group_id,	userperm,groupperm,otherperm)
--		values (1,fc_file_id_seq.nextval,'00000000-0000-0000-0000-000000000004','test-dir','d',0,0,255,bin_to_num(0,1,0,1,0,0,0,1),bin_to_num(1,1,1,1,1,1,1,1))
--;
--insert into fc_file(parent_id,file_id,guid,name, type,principal_id,group_id,	userperm,groupperm,otherperm)
--		values (1,fc_file_id_seq.nextval,'00000000-0000-0000-0000-000000000005','BaseDir','d',0,0,255,bin_to_num(0,1,0,1,0,0,0,1),bin_to_num(1,1,1,1,1,1,1,1))
--;



/* metadata init */


commit
;

create table meta_schema_attributes
	(
		schema_id integer default 0 not null, --guid
		schema_guid varchar2(36) not null,
		schema_name varchar2(1024) not null,
		constraint pk_attrib_sch_constraint primary key (schema_id),
		constraint uq_schema_att_name unique(schema_name),
		constraint uq_schema_guid unique(schema_guid)
	)
	organization index
	;

create or replace synonym msch for meta_schema_attributes;
-- on insert into this table we create partition to hold all the values related to this schema

insert into msch values(0,'00000000-0000-0000-0001-000000000000','default schema');

create or replace synonym matt for meta_attributes;


create table meta_attributes
	(
		schema_id number default 0 not null
					constraint fk_schema
					references meta_schema_attributes(schema_id)
					on delete cascade,
		att_id 		integer not null,
		att_name	varchar2(255),
		type_id 	varchar2(255), -- for now as text, change to mapping later to provide uniqueness if needed
		constraint pk_attrib_constraint primary key(schema_id,att_id),
		constraint uq_attrib_name unique(att_name, schema_id)
	)
	organization index
	compress 1
	--partition by schema_id
	;
  
  
create table meta_attrib_values
	(
		schema_id integer default 0 not null,
		att_id integer not null,
		parent_id integer not null, --partition by owner also?
		file_id integer not null,
		-- can be any type - we may switch to objects later
		val varchar2(255) not null,
		constraint pk_attrib_val primary key (schema_id,parent_id,file_id,att_id),
		constraint fk_atts foreign key (schema_id,att_id) references meta_attributes on delete cascade,
		constraint fk_fc foreign key (parent_id,file_id) references fc_file(parent_id,file_id) on delete cascade
	)
	organization index
	compress 3
	partition by list(schema_id)
	(
		partition main_partition values(0),
		partition others_partition values(default)
	)
	;
  
create or replace synonym mval for meta_attrib_values;