CREATE OR REPLACE 
PACKAGE PERMISSION_UTIL AS
/**
  $Name: glite-data-catalog-service-fr_R_2_1_15 $
  $Header: /cvs/jra1mw/org.glite.data.catalog-service-fr/config/oracle/package.permission.sql,v 1.8 2005/10/26 10:20:57 knienart Exp $

	Package to handle internal and external permissions logic.

*/

/* masks for permissions bit checks */

writemask constant pls_integer default 128;
readmask  constant pls_integer default 64;
execmask  constant pls_integer default 32;
listmask  constant pls_integer default 16;
removemask constant pls_integer default 8;
permmask  constant pls_integer default 4;
setmetamask constant pls_integer default 2;
getmetamask constant pls_integer default 1;


writeFileOp constant pls_integer default 1;
writeDirOp constant pls_integer default 2;
readFileOp constant pls_integer default 3;
readDirOp constant pls_integer default 4;
excecuteFileOp constant pls_integer default 5;
listFileOp constant pls_integer default 6;
listDirOp constant pls_integer default 7;
removeFileOp constant pls_integer default 8;
permissionOp constant pls_integer default 9;
setmetaOp constant pls_integer default 10;
getmetaOp constant pls_integer default 11;


homeDirectoryRoot constant varchar2(50):='/user';

/* generic types */
type returnR is record
	(
	guid perm_basic_permission.GUID%type,
	result integer := 0,
	errString varchar2(100)
	);
type returnT is table of returnR index by binary_integer;

--type gGuidsT is table of perm_basic_permission.GUID%type INDEX  by binary_integer;


/* errors */

accessDeniedErrStr constant varchar2(50) default 'Access Denied';
accessDeniedErrVal constant integer default 2;

NoSuchFileStr constant varchar2(50) default 'No such file or directory.';
NoSuchFileVal constant integer default 3;

cannotGetDeafultPermissionStr constant varchar2(70) default 'Cannot get default permission. Be sure you have proper rights.';
cannotGetDeafultPermissionVal constant integer default 4;

noMasterReplicaGivenStr constant varchar2(50) default 'No master replica given';
noMasterReplicaGivenVal constant integer default 5;

generalInternalBulkErrorStr constant varchar2(50) default 'General internal bulk error';
generalInternalBulkErrorVal  constant integer default 6;

fileOrDirAlreadyExistsStr constant varchar2(50) default 'File or dir already exists';
fileOrDirAlreadyExistsVal  constant integer default 7;

missingDataStr constant varchar2(50) default 'Missing data to proceed.';
missingDataVal  constant integer default 8;

generalInternalErrorStr constant varchar2(150) default 'General internal error. Contact manufacturer or a lawyer.';
generalInternalErrorVal  constant integer default 9;

InvalidParentStr constant varchar2(50) default 'Invalid parent specified.';
InvalidParentVal constant integer default 10;

directoryNotEmptyStr constant varchar2(30) default 'Directory not empty.';
directoryNotEmptyVal constant integer default 11;

invalidArgumentDirStr constant varchar2(50) default 'Operation not valid for directory';
invalidArgumentDirVal constant integer default 12;

loopInSymlinksDetectedStr constant varchar2(50) default 'Detected loop during symlink traversal.';
loopInSymlinksDetectedVal constant integer default 13;

 


/* Fundamental permission atom	*/
superUsers listNum;

--
/*	acl list type	*/
--type PermissionEntryT is array(1000) of PermissionEntryR ;
/* types used to provide principal data to internal calls.
 * by definition there must be at least one item for the principals dn
 * and optional group dn's
 */
--type principalListT is varray(100000) of varchar2(250);
--type principalListIdT is varray(100000) of integer;

type PermStringIntPair is record
	(
	item1 varchar2(250),
	item2 INTEGER
	)
	;
type PermStringIntPairT is table of PermStringIntPair index by binary_integer;

type dataAndStatus is record
	(
	guid perm_basic_permission.guid%type,
	permitted decimal(1)
	)
	;
type dataAndStatusT is table of dataAndStatus index by binary_integer;


/****************************************************************************************************
*
*	Methods
*
*****************************************************************************************************/

function addGuid(
			 securityToken in principalToken,
			 permissions in PermissionEntryT,
          outerrors out nocopy returnOT)
			return returnO
;

function setGuidPermission(
			 securityToken in principalToken,
			 permissions in PermissionEntryT,
          outerrors out nocopy returnOT)
			return returnR
;

function getGuidPermission(
			 securityToken in principalToken,
			 guids in guidOT,
			 permissions out nocopy PermissionEntryT,
          outerrors out nocopy returnOT)
			return returnR
;

function removeGuid(
			 securityToken in principalToken,
			 guids in guidOT,
          outerrors out nocopy returnOT)
			return returnR
;

-- adds users to the catalog
--function addUser(
--			 securityToken in principalToken,
--        PermStringIntPairOT(
--          outerrors out nocopy returnOT)
--			return returnR
--;

procedure refreshSuperUsers;
/*************************************************************************
*
*			Utility functions
*
**************************************************************************/

/* returns list of ids for passed principalList to be used in subsequent calls
	tbd what happens if some names dont match existing users.
	unconditional creation would create huge security hole, on the oher hand
	we trust security provider, so all principalnames should be ok and if
	one does not exist should be created.
  in or our of curent transaction

*/
function getPrincipalIdIUW(pname in perm_principal.name%type, ptype in char, insertop in pls_integer)
	return perm_principal.principal_id%type
;

procedure  getPrincipalIds(securityToken in principalToken, userId out nocopy perm_principal.principal_id%type,
	groupIds out nocopy PermStringIntPairT,insertop in pls_integer)
	;
/* obj version */
procedure getPrincipalIdsO(securityToken in principalToken,
	userId out nocopy perm_principal.principal_id%type,
	groupIds out nocopy PermStringIntPairOT,
	insertop in pls_integer,
  superuser in boolean:=false)
   deterministic
;
/* returns id for passed name, type */
function getPrincipalId(pname in perm_principal.name%type, ptype in char,	insertop in pls_integer)
	return perm_principal.principal_id%type
  deterministic
	;
/*	 returns name - reversed getPrincipalId */
function getPrincipalNameStr(pid in integer)
	return varchar2
	deterministic
;
PRAGMA RESTRICT_REFERENCES (getPrincipalNameStr, WNDS, RNPS, WNPS, RNPS, TRUST);


/* returns array of [guid,permitted] pairs where permitted says if operation is permitted for a given [guid,op] pair*/
procedure checkPrincipalPermissions(
				securityToken in principalToken,
				data in out nocopy dataAndStatusT,
				userPrincipal out nocopy perm_principal.principal_id%type,
				principals out nocopy PermStringIntPairT,
				op in pls_integer)
;
/* stripped version */
procedure checkPrincipalPermissions(
				securityToken in principalToken,
				data in out nocopy dataAndStatusT,
				op in pls_integer)
;

/* one conversion call less version */
procedure checkPrincipalPermissions(
				securityToken in principalToken,
				guids in guidOT,
				data out nocopy dataAndStatusT,
				op in pls_integer)
;

/* returns 1 if operation op permitted for supplied permVal permission value, zero otherwise */
function checkPerm
(
	permVal in bp.USERPERM%type,
	op in pls_integer
	)
	return pls_integer
	deterministic
	;
PRAGMA RESTRICT_REFERENCES (checkPerm, WNDS, RNPS, WNPS, RNPS, TRUST);

function getOpMask(op in pls_integer)
return pls_integer
deterministic
;

PRAGMA RESTRICT_REFERENCES (getOpMask, WNDS, RNPS, WNPS, RNPS, TRUST);

function isOkBasicPerm(op in pls_integer, u in pls_integer, g in pls_integer, o in pls_integer)
	return decimal
	deterministic
;
PRAGMA RESTRICT_REFERENCES (isOkBasicPerm, WNDS, RNPS, WNPS, RNPS, TRUST);

function isOk(u in pls_integer, g in pls_integer, o in pls_integer)
	return pls_integer
	deterministic
;

PRAGMA RESTRICT_REFERENCES (isOk, WNDS, RNPS, WNPS, RNPS, TRUST);

function isOk(u in pls_integer, g in pls_integer, o in pls_integer,
	p_id in pls_integer,p_id_in in pls_integer,
	g_id in pls_integer)
	return pls_integer
	deterministic
;
PRAGMA RESTRICT_REFERENCES (isOk, WNDS, RNPS, WNPS, RNPS, TRUST);

function isOkAclJoin(u in pls_integer, g in pls_integer, o in pls_integer,
	p_id in pls_integer,p_id_in in pls_integer,
	g_id in pls_integer,
	principals in PermStringIntPairOT,
	iguid in perm_acl_permission.guid%type,
	op in pls_integer
	)
	return pls_integer
;

function isOkAclMaskJoin(u in pls_integer, g in pls_integer, o in pls_integer,
	p_id in pls_integer,p_id_in in pls_integer,
	g_id in pls_integer,
	principals in PermStringIntPairOT,
	iguid in perm_acl_permission.guid%type,
	mask in pls_integer
	)
	return pls_integer
;
function isOkAcl(u in pls_integer, g in pls_integer, o in pls_integer,
	p_id in pls_integer,p_id_in in pls_integer,
	g_id in pls_integer,
	principals in PermStringIntPairOT,
	iguid in perm_acl_permission.guid%type,
	op in pls_integer
	)
	return pls_integer
--  deterministic
	;
 
--PRAGMA RESTRICT_REFERENCES (isOkAcl, RNPS, WNPS, RNPS, TRUST);


function isOkAclX (
  res in ResolverO,
	p_id_in in pls_integer,
	principals in PermStringIntPairOT,
	op in pls_integer
	)
	return resolverO	
  ;


function isOkAclXMask (
  res in ResolverO,
	p_id_in in pls_integer,
	principals in PermStringIntPairOT,
	mask in pls_integer
	)
	return resolverO	
  ;


function isOkAcl(
  resolver in ResolverT,  
	p_id_in in pls_integer,	
	principals in PermStringIntPairOT,	
	op in pls_integer
	)
	return returnOT
  ;

function isOkAclMask(u in pls_integer, g in pls_integer, o in pls_integer,
	p_id in pls_integer,p_id_in in pls_integer,
	g_id in pls_integer,
	principals in PermStringIntPairOT,
	iguid in perm_acl_permission.guid%type,
	mask in pls_integer
	)
	return pls_integer
	;

function isOkAcl2(u in pls_integer, g in pls_integer, o in pls_integer,
	p_id in pls_integer,p_id_in in pls_integer,
	g_id in pls_integer,	gexists in pls_integer,
	principals in PermStringIntPairOT,
	iguid in perm_acl_permission.guid%type,
	op in pls_integer
	)
	return PermissionEntryO
;

function isOkAclPermission(u in pls_integer, g in pls_integer, o in pls_integer,
	p_id in pls_integer,p_id_in in pls_integer,
	g_id in pls_integer,	gexists in pls_integer,
	principals in PermStringIntPairOT,
	iguid in perm_acl_permission.guid%type,
	op in pls_integer
	)
	return PermissionO
;

function maximum(a in pls_integer, b in pls_integer)
	return decimal
	deterministic
;
PRAGMA RESTRICT_REFERENCES (maximum, WNDS, RNPS, WNPS, RNPS, TRUST);


function getPrincipalNameS(pids in principalListIdT)
	return principalNameListT
	pipelined
;

function getPrincipalName(pid in integer)
	return principalName
;

function getACL(g in varchar2)
	return ACLList
;

PRAGMA RESTRICT_REFERENCES (getAcl, WNDS, RNPS, WNPS, TRUST);

function translateAclToSimpleAcl
  ( nameAcl in ACLList)
  return ACLSimpleIdList
  deterministic
  ;
procedure translateAclToSimpleAcl
  ( nameAcl in  ACLList, outRet in out nocopy ACLSimpleIdList)
  deterministic
;

/***************************** testing junk ****************************/
function getprincipalNameListT
	return principalNameListT
;
--
--function generateSecToken
--	return principalToken
--;
--
--function generateData
--	return dataAndStatusT
--	;
function generateDataO(k in integer)
	return dataAndStatusOT
;

-- pass in security context, permEntry to set, get
--procedure testSecurityArray(
--	securityList in principalListT,
--	permEntry  in PermissionEntryT,
--	outPerm out nocopy PermissionEntryT,
--	dataStatusT out nocopy dataAndStatusT)
--	;
function genPermEntry(limit in number)
	return PermissionEntryT;
function getPermEntryACLIdList(limit in number)
	return ACLIdList;
function getPermStringIntPairOT
	return PermStringIntPairOT;
function getSuperUsers
  return listNum;

END;
