create or replace 
PACKAGE BODY PERMISSION_UTIL AS

/* TODO:
  1. create trigger to del from acl while removing or rather put del code from acl
  2. be sure ordering is not lost ie we dont mess up with index in the input arrays
		when errors appear (checkPrincipalPermissions usage)
*/


globalDefaultGuid CONSTANT VARCHAR(36) DEFAULT '10000000-0000-0000-0000-000000000000';


type namesT is table of varchar2(250);
subtype bpR is perm_basic_permission%rowtype;
--	type bpT is table of bpR index by binary_integer;
type bpT is varray(200000) of bpR;
type aclT is table of perm_acl_permission%rowtype index by binary_integer;
type aclAT is varray(200000) of perm_acl_permission%rowtype ;
type principalNameType is record (name varchar2(250), p_type char(1));
type principalNameTypeT is table of principalNameType;

type pairR is record (a integer,b integer);
type pairsT is table of pairR index by varchar2(36);
type guidIdxT is table of integer index by varchar2(36);
type nameidAssocT is table of integer index by varchar2(250);

type PermSerializerR is record
	(
		guid varchar2(36),
		userName varchar2(250),
		groupName varchar2(250),
		userPerm decimal(2),
		groupPerm decimal(2),
		otherPerm decimal(2),
		ok decimal(1),
		acl ACLList
	);
type PermSerializerT is table of PermSerializerR index by binary_integer;


type bulkPrincipalidT is varray(200000) of perm_basic_permission.principal_id%type;
type bulkGroupIdT is varray(200000) of perm_basic_permission.group_id%type;
type bulkUserPT is varray(200000) of perm_basic_permission.userperm%type;
type bulkGroupPT is varray(200000) of perm_basic_permission.groupperm%type;
type bulkOtherPT is varray(200000) of perm_basic_permission.otherperm%type;
type bulkGuidT is varray(200000) of perm_basic_permission.guid%type;
type bulkPermittedT is varray(200000) of decimal(1);

--type bulkACLTypeT is table of perm_acl_permission.acl_type%type;

/* internal types */
--type permR is record
--	(
--	readPerm char(1),
--	writePerm char(1),
--	execPerm char(1),
--	list char(1),
--	remove char(1),
--	permission char(1)
--
--	)
--	;
subtype permR is decimal(2);
/* ux like permission, always present, does not refer to the schema object with similar name. */


--type BasicPermissionR is record
--	(
--	userName varchar2(250),
--	groupName varchar2(250),
--	userPerm permR,
--	groupPerm permR,
--	otherPerm permR
--	)
--	;

type ACLEntryR is record
	(
	principal varchar2(250),
	principalPerm permR,
	principalType char(1)
	)
	;

type ACLtbl is table of ACLEntryR index by binary_integer;

/* array of ACL entries */
type ACLArrT is varray(1000) of ACLEntryR;

/* permission which has all perms there	*/
--type PermissionR is record
--	(
--		basicperm BasicPermissionR ,
--		acl ACLArrT
--	)
--	;

/* utility to pass permissions around with the key	it describes*/
--type PermissionEntryR is record
--	(
--		Item varchar2(1024),
--		perm PermissionR
--	)
--	;
/*internal types end*/
FUNCTION generateGuid
	RETURN VARCHAR2
	AS
	lguid VARCHAR2(36);
	BEGIN
		lguid:=sys_guid;
		RETURN SUBSTR(lguid,1,8)||'-'||
		SUBSTR(lguid,9,4)||'-'||
		SUBSTR(lguid,13,4)||'-'||
		SUBSTR(lguid,17,4)||'-'||
		SUBSTR(lguid,21,12);
	END;



procedure debugOn
as
begin
   begin
	null;
  DBMS_DEBUG_JDWP.CONNECT_TCP( 'pcegeekn.cern.ch', '5003' );
	exception
		when others then
		null;
	end;
end;


/* conversion functions */
FUNCTION dataAndStatusT2SQL(aPlsqlItem permission_util.dataAndStatusT)
RETURN dataAndStatusOT
IS
aSqlItem dataAndStatusOT:=dataAndStatusOT();
BEGIN
	-- initialize the table
	aSqlItem.EXTEND(aPlsqlItem.COUNT);
		FOR I IN 1..aPlsqlItem.LAST LOOP
		    aSqlItem(I):=dataAndStatusO(aPlsqlItem(I).guid,aPlsqlItem(I).permitted);
		END LOOP;
	RETURN aSqlItem;
END dataAndStatusT2SQL;


FUNCTION dataAndStatusT2PL(sqlItem dataAndStatusOT)
RETURN permission_util.dataAndStatusT
IS
aPLItems permission_util.dataAndStatusT;
aPLItem permission_util.dataAndStatus;
BEGIN
		FOR I IN 1..sqlItem.LAST LOOP
			aPLItem.guid:=sqlItem(I).guid;
			aPLItem.permitted:=sqlItem(I).permitted;
		    aPLItems(I):=aPLItem;
		END LOOP;
	RETURN aPLItems;
END dataAndStatusT2PL;


/******************************************************************
*
*								METHODS
*
******************************************************************/

/* creates new mapping:
	Deserializes permissions object and ingests them into bp and acl tables.

	exceptions:
		guid already exist
	notes:
		everybody can insert new guid permissions...
*/


function addGuid(
			 securityToken in principalToken,
			 permissions in PermissionEntryT,
          outerrors out nocopy returnOT)
			return returnO
as

	ret returnO:=returnO(0);

	userPrincipal perm_principal.principal_id%type;
	principals PermStringIntPairT;	-- dn and internal id records


	guidMarker pairR;
	guidMarkerAcl pairsT;


	aclR perm_acl_permission%rowtype;

	bulkBP bpT:=bpT();
	bulkAcl aclT;
	bulkAclNew aclAT:=aclAT();


	checkdata dataAndStatusT;

	k pls_integer:=1;
	n pls_integer;
	j pls_integer:=0;

	/* errors */
	nerrors pls_integer;
	nerrorsBp pls_integer;
	bulk_errors EXCEPTION;
	PRAGMA exception_init(bulk_errors, -24381);
	starterr pls_integer:=1;
begin
	outerrors:=returnOT();

	/* for all permissions items do:
		get item(guid)
			get basic_perms
				insert bp
			get acl
				insert acl
	*/

-- do name-> id translation
	getPrincipalIds(securityToken,userPrincipal,principals,1);

 --TODO: change iuw to bulk check and insert with one call using arrays for this.
	-- create users if not existing by getprincipalId
	bulkBP.extend(permissions.count);
	for i in 1..permissions.count loop
		bulkBP(i).guid:=permissions(i).item;
		bulkBP(i).principal_id:=getPrincipalIdIUW(permissions(i).perm.username,'',1);
		bulkBP(i).group_id:=getPrincipalIdIUW(permissions(i).perm.groupname,'',1);
		bulkBP(i).userperm:=permissions(i).perm.userPerm;
		bulkBP(i).groupperm:=permissions(i).perm.groupPerm;
		bulkBP(i).otherperm:=permissions(i).perm.otherPerm;

		if permissions(i).perm.acl is not null and permissions(i).perm.acl.count>0 then
		 guidMarker.a:=j+1;
		 guidMarker.b:=j+permissions(i).perm.acl.count;
		 guidMarkerAcl(permissions(i).item):=guidMarker;

			for ac in 1..permissions(i).perm.acl.count loop
			 j:=j+1;
			 bulkAcl(j).guid:=permissions(i).item;
--			 bulkAcl(j).principal_id:=getPrincipalId(permissions(i).perm.acl(ac).principal,permissions(i).perm.acl(ac).acl_type);
			 bulkAcl(j).principal_id:=getPrincipalIdIUW(permissions(i).perm.acl(ac).principal,'',1); --change it to one call per all perms
			 bulkAcl(j).perm:=permissions(i).perm.acl(ac).perm;
--			 bulkAcl(j).acl_type:=permissions(i).perm.acl(ac).acl_type;
--			 DBMS_OUTPUT.PUT_LINE('acls: '||permissions(i).perm.acl(ac).principal);
			end loop;
		end if;
--		DBMS_OUTPUT.PUT_LINE('acls: '||bulkACl.count);
	end loop;


	begin
		forall i in 1..bulkBp.count save exceptions
			insert into perm_basic_permission
				values bulkBP(i);
	exception
		WHEN bulk_errors THEN
		nerrorsBp := SQL%BULK_EXCEPTIONS.COUNT;
		outerrors.extend(nerrorsBp);
		FOR i IN 1..nerrorsBp LOOP
			outerrors(i):=
			returnO
					(
						permissions(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).item,
						SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
						permission_util.generalInternalBulkErrorVal,
						substr(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||i ,1,100)
					);
			if bulkAcl.count>0 then
			bulkAcl.delete(guidMarkerAcl(permissions(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).item).a,
							   guidMarkerAcl(permissions(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).item).b);
			end if;
--			DBMS_OUTPUT.PUT_LINE('acls after error: '||bulkACl.count||', error:'||outerrors(i).errString);
		END LOOP;
			--make it sparse again

			if bulkACl.count >0 then
			j:=0;
			for z in 1..bulkACl.last loop
				if bulkAcl.exists(z) then
					bulkAclNew.extend;
					j:=j+1;
					bulkAclNew(j):=bulkAcl(z);
				end if;
			end loop;
			end if;

	end;

	--create outstanding acl entries
	-- beware - will not be inserted these of which of bp fails - double error msgs...
	-- we should delete err items from bulkACl..:1:2:3


	if bulkAcl.count>0 then

	forall i in 1..bulkAcl.LAST save exceptions
		insert into perm_acl_permission
			values bulkAcl(i);

	elsif bulkAclNew.count>0 then

		forall i in 1..bulkAclNew.LAST save exceptions
			insert into perm_acl_permission
				values bulkAclNew(i);
	end if;


	return ret;

exception
	WHEN bulk_errors THEN
	nerrors := SQL%BULK_EXCEPTIONS.COUNT;
	outerrors.extend(nerrors);
	FOR i IN 1..nerrors LOOP
		outerrors(i+nerrorsBp):=
			returnO
				(
				bulkACL(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid,
				SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
				permission_util.generalInternalBulkErrorVal,
				substr(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| 'at index '||i ,1,100)
				);
	END LOOP;

	return ret;
	when others then
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
		return ret;

end;


/**********************************************************************************
*		Sets permissions for exisiting guids.
*		If guid does not exist, or no permission to update it: set error
*
*		No check is done if double guids exist, resulting in possible exceptions
************************************************************************************/
function setGuidPermission(
			 securityToken in principalToken,
			 permissions in PermissionEntryT,
          outerrors out nocopy returnOT)
			return returnR
as
	ret returnR;


	userPrincipal perm_principal.principal_id%type;
	principals PermStringIntPairT;	-- dn and internal id records


	--bp
	bulkPrincipalId bulkPrincipalIdT:=bulkPrincipalIdT();
	bulkGroupId     bulkGroupIdT:=bulkGroupIdT();
	bulkUserP       bulkUserPT:=bulkUserPT();
	bulkGroupP      bulkGroupPT:=bulkGroupPT();
	bulkOtherP      bulkOtherPT:=bulkOtherPT();
	bulkGuid        bulkGuidT:=bulkGuidT();

	--acl
	bulkACLPrincipalId bulkPrincipalIdT;
	bulkACLUserP       bulkUserPT;
	bulkACLGuid        bulkGuidT;
--	bulkACLType 	 	 bulkACLTypeT;




	bulkBP bpT:=bpT();
	bulkAcl ACLIdList:=ACLIdList();
	bulkAclNew ACLIdList:=ACLIdList();

	guidMarker pairR;
	guidMarkerAcl pairsT;


	checkdata dataAndStatusT;
	guididx guidIdxT;
	AclstoDelete PermStringIntPairT;
--	type perm

	k pls_integer:=1;
	n pls_integer;
	j pls_integer:=0;

	/* errors */
	nerrors pls_integer;
	nerrorsBp pls_integer:=0;
	bulk_errors EXCEPTION;
	PRAGMA exception_init(bulk_errors, -24381);
	starterr pls_integer:=1;

begin

	outerrors:=returnOT();
/* for all permissions items do:
		get item(guid)
			check if permitted to change permissions
			get basic_perms
				update bp
			get acl
				update acl
	*/

	/*userName varchar2(250),
	groupName varchar2(250),
	userPerm decimal(2),
	groupPerm decimal(2),
	otherPerm decimal(2)
	acl ACLList*/
-- do name-> id translation

	/* check permissions for guids/users */
	for i in 1..permissions.count loop
		checkdata(i).guid:=permissions(i).item;
		guididx(checkdata(i).guid):=i;
	end loop;

checkPrincipalPermissions(securityToken,
				checkdata,
				userPrincipal,
				principals,
				permissionOp);

	-- create users if not existing by getprincipalId
	bulkGuid.extend(permissions.count);
	bulkPrincipalId.extend(permissions.count);
	bulkGroupId.extend(permissions.count);
	bulkUserP.extend(permissions.count);
	bulkGroupP.extend(permissions.count);
	bulkOtherP.extend(permissions.count);


	for i in 1..permissions.count loop
	 if checkdata(i).permitted=1 then
		n:=guididx(checkdata(i).guid);
--		dbms_output.put_line('i:'||i||', '||permissions(n).item||', '||checkdata(i).permitted||'cdata: '||checkdata(i).guid);

--			permissions(n).perm.groupPerm||', '||permissions(n).perm.otherPerm);

		bulkGuid(k):=permissions(n).item;
		bulkPrincipalId(k):=getPrincipalIdIUW(permissions(n).perm.username,'',1);
		bulkGroupId(k):=getPrincipalIdIUW(permissions(n).perm.groupname,'',1);
		bulkUserP(k):=permissions(n).perm.userPerm;
		bulkGroupP(k):=permissions(n).perm.groupPerm;
		bulkOtherP(k):=permissions(n).perm.otherPerm;
		k:=k+1;

		if permissions(n).perm.acl.count>0 then
		 guidMarker.a:=j+1;
		 guidMarker.b:=j+permissions(n).perm.acl.count;
		 guidMarkerAcl(permissions(n).item):=guidMarker;

			bulkAcl.extend(permissions(n).perm.acl.count);
			--DBMS_OUTPUT.PUT_LINE('acl count:'||permissions(n).perm.acl.count);
			for ac in 1..permissions(n).perm.acl.count loop
			 j:=j+1;
			 bulkAcl(j):=ACLEntryIdO(
			 permissions(n).item,
--			 getPrincipalId(permissions(n).perm.acl(ac).principal,permissions(n).perm.acl(ac).acl_type),
			 getPrincipalIdIUW(permissions(n).perm.acl(ac).principal,'',1),
			 permissions(n).perm.acl(ac).perm
--			 , permissions(n).perm.acl(ac).acl_type
			 );
--			 bulkAcl(j).guid:=permissions(n).item;
--			 bulkAcl(j).principal_id:=getPrincipalId(permissions(n).perm.acl(ac).principal,permissions(n).perm.acl(ac).acl_type);
--			 bulkAcl(j).perm:=permissions(n).perm.acl(ac).perm;
--			 bulkAcl(j).acl_type:=permissions(n).perm.acl(ac).acl_type;
--			 DBMS_OUTPUT.PUT_LINE('acls: '||permissions(n).perm.acl(ac).principal);
			end loop;
		end if;
	 else
			nerrorsBp:=nerrorsBp+1;
			outerrors.extend();
			outerrors(nerrorsBp):=
					returnO
						(checkdata(i).guid,
						guididx(checkdata(i).guid),
						permission_util.accessDeniedErrVal ,
						permission_util.accessDeniedErrStr ||' at index: '||i);

	 end if; --permitted
--	 DBMS_OUTPUT.PUT_LINE('acls: '||bulkACl.count);
	end loop;


	bulkGuid.trim(nerrorsBp);
	bulkPrincipalId.trim(nerrorsBp);
	bulkGroupId.trim(nerrorsBp);
	bulkUserP.trim(nerrorsBp);
	bulkGroupP.trim(nerrorsBp);
	bulkOtherP.trim(nerrorsBp);


	begin
		forall i in 1..bulkGuid.count save exceptions
			update perm_basic_permission
			set
				principal_id=bulkPrincipalId(i),
				group_id=bulkGroupId(i),
				userperm=bulkUserP(i),
				groupperm=bulkGroupP(i),
				otherperm=bulkOtherP(i)
				where
				guid=bulkGuid(i)
				--returning userperm bulk collect into bulkUserP
				;


	exception
		WHEN bulk_errors THEN
		nerrors := SQL%BULK_EXCEPTIONS.COUNT;
		outerrors.extend(nerrors);
		FOR i IN 1..nerrors LOOP
			outerrors(i+nerrorsBp):=
				returnO
					(
					permissions(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).item,
					SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
					permission_util.generalInternalBulkErrorVal,
					substr(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| 'at index '||i ,1,100)
					);
			bulkAcl.delete(guidMarkerAcl(permissions(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).item).a,
							   guidMarkerAcl(permissions(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).item).b);
			--DBMS_OUTPUT.PUT_LINE('acls after error: '||bulkACl.count);
		END LOOP;
		nerrorsBp:=nerrorsBp+nerrors;

			--make it sparse again

			if bulkACl.count >0 then
			bulkAclNew:=ACLIdList();
			j:=0;
			for z in 1..bulkACl.last loop
				if bulkAcl.exists(z) then
					bulkAclNew.extend;
					j:=j+1;
					bulkAclNew(j):=bulkAcl(z);
				end if;
			end loop;
			end if;

	end;

	/*
	for acls - delete existing ones first :4:5
				- delete errorenous ones too :6:7
	*/

	/* delete ones for a given guid not in the input set...
			(delete all for now)
	*/

--items to delete - exist in acl but not in input
delete from acl
where
(guid, principal_id) in
(
  select /*+ cardinality( d 100 ) */ distinct a.guid, a.principal_id
  from
  acl a,
  table(cast(bulkAcl as ACLIdList)) d
  where
    a.guid=d.guid
  minus
  select /*+ cardinality( d 100 ) */ a.guid, a.principal_id
  from
  acl a,
  table(cast(bulkAcl as ACLIdList)) d
  where
    a.guid=d.guid and a.principal_id=d.principal_id
 )
 ;

--items to update or insert
-- if error occurs here we have some general problem... we could switch it to dynamic bulk sql later

	merge /*+ cardinality (S 100) */ into acl D
	using (select guid,principal_id,perm from table(cast(bulkACl as ACLIdList))) S
	on (S.principal_id=D.principal_id and S.guid=D.guid)
	when matched then update set perm=S.perm
	when not matched then insert values(S.guid,S.principal_id,S.perm)
  ;

	return ret;

exception
	WHEN bulk_errors THEN --never reaches here..
	nerrors := SQL%BULK_EXCEPTIONS.COUNT;
	FOR i IN nerrorsBp+1..nerrors+nerrorsBp LOOP
		outerrors(i+nerrorsBp):=
					returnO
					(
					bulkACL(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid,
					SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
					permission_util.generalInternalBulkErrorVal,
					substr(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| 'at index '||i ,1,100)
					);
	END LOOP;
	return ret;
--	when others then
--		ret.result:=SQLCODE;
--		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('other error:'||ret.errString);
--		return ret;

end;

/* retrieve all permission user has right to read (permissions) */
function getGuidPermission(
			 securityToken in principalToken,
			 guids in guidOT,
			 permissions out nocopy PermissionEntryT,
          outerrors out nocopy returnOT)
			return returnR
as
	ret returnR;

	userPrincipal perm_principal.principal_id%type;
	principalsO PermStringIntPairOT;	-- dn and internal id objects

	checkdata dataAndStatusT;
	checkdataO dataAndStatusOT;

	guididx guidIdxT;

	tmp PermSerializerT;
  errItem pls_integer:=0;
begin
	outerrors:=returnOT();
	/* check if permitted for these guids
		for those which permiited
			get * from bp
				for guid row from bp
					get all from acl, build acl array
	*/


	getPrincipalIdsO(securityToken, userPrincipal, principalsO,0);
	for i in 1..guids.count loop
		guididx(guids(i).guid):=i;
	end loop;

select /*+ use_nl(inP) cardinality( i 100 ) cardinality (inP 10)  */
			isOkAcl2(bp.userperm,bp.groupperm,bp.otherperm,
			bp.principal_id,userPrincipal,
			bp.group_id,item2,
	      principalsO,
			i.guid,readFileOp
		) val
 bulk collect into permissions
  from
 ( (table(guids) i left outer join bp on i.guid=bp.guid)
			  left outer join table(principalsO) inP
				 on bp.group_id=inP.item2)
			;

	/* fill up with errors */

	for i in 1..permissions.last loop
		if permissions(i).perm is null then
		outerrors.extend();
		outerrors(i):=returnO
			(
			permissions(i).item,
			guididx(i),
			accessDeniedErrVal,
			accessDeniedErrStr
			);
		end if;
	end loop;


	return ret;
end;

/******************************** remove guid **************************/
function removeGuid(
			 securityToken in principalToken,
			 guids in guidOT,
          outerrors out nocopy returnOT)
			return returnR
as
--	ret returnO:=returnO('',0,'');
	ret returnR;
	nerrors pls_integer;
	nerrorsBp pls_integer:=0;
	k pls_integer:=1;
	n pls_integer;

	userPrincipal perm_principal.principal_id%type;
	principals PermStringIntPairT;	-- dn and internal id records

	checkdata dataAndStatusT;
	checkdataO dataAndStatusOT;
	bulkGuid        bulkGuidT:=bulkGuidT();
	bulkPermitted bulkPermittedT:=bulkPermittedT();
	guididx guidIdxT;

	bulk_errors EXCEPTION;
	PRAGMA exception_init(bulk_errors, -24381);

begin
	outerrors:=returnOT();
	/* check if permitted
		for those which permitted
			remove from bp
				(all from acl removed via constraint)
				dont check if stale perm_principal entries exist.
	*/
	/* check permissions for guids/users */
	bulkGuid.extend(guids.count);
	bulkPermitted.extend(guids.count);
	for i in 1..guids.count loop
		checkdata(i).guid:=guids(i).guid;
		guididx(checkdata(i).guid):=i;
	end loop;
	checkPrincipalPermissions(securityToken,
				checkdata,
				userPrincipal,
				principals,
				removeFileOp);
	for i in 1..checkdata.last loop
	 if checkdata(i).permitted=1 then
		n:=guididx(checkdata(i).guid);
	--	dbms_output.put_line('i:'||i||', '||guids(n).guid);
		bulkGuid(k):=guids(n).guid;
		k:=k+1;
	 else		--could be checked for not existence here too (=2)
		nerrorsBp:=nerrorsBp+1;
		outerrors.extend;
		outerrors(nerrorsBp):=
					returnO
					(
					checkdata(i).guid,
					n,
					permission_util.accessDeniedErrVal ,
					permission_util.accessDeniedErrStr ||' at index: '||i
					);
	 end if; --permitted

	end loop;
	bulkGuid.trim(nerrorsBp);

	forall i in 1..bulkGuid.count save exceptions
		delete from bp
			where
				guid=bulkGuid(i);
  --delte from acl as well, or use trigger


	return ret;

exception
	WHEN bulk_errors THEN
	nerrors := SQL%BULK_EXCEPTIONS.COUNT;
	outerrors.extend(nerrors);
	FOR i IN 1..nerrors LOOP

		outerrors(i+nerrorsBp):=
					returnO
					(
					bulkguid(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX),
					SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
					permission_util.generalInternalBulkErrorVal,
					substr(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| 'at index '||i ,1,100)
					);
	END LOOP;
	return ret;
	when others then
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
		DBMS_OUTPUT.PUT_LINE('other error:'||ret.errString);
		return ret;

end;

/*******************************************************************
* 								UTILITIES

NOTE: Object type do not work with bulk collect into!!!!

*******************************************************************/

/* returns name,principal_id pairs for principal names */
procedure getPrincipalIds(securityToken in principalToken,
	userId out nocopy perm_principal.principal_id%type,
	groupIds out nocopy PermStringIntPairT,
	insertop in pls_integer)

as
newNames namesT;
newPrincipals PermStringIntPairT;
gSize pls_integer;
begin

	begin
		/* get user id */
		select principal_id
			into userId
		from perm_principal
			where
			name=securityToken.user_name
	--		and p_type='u'
			;
	exception
		when no_data_found then
		if insertop<>0 then
				insert into perm_principal(principal_id,name,p_type--,defaultUser,defaultGroup
				)
					values(fc_principal_id_seq.nextval,securityToken.user_name,'u'--,0,0
					)
					returning principal_id into userId;
		end if;
	end;

	/* get ids of all user's groups'*/
select n.name, p.principal_id
	bulk collect into groupids
		from
		perm_principal p,
		table(cast(securityToken.group_names as principalNameListT)) n
		where
		n.name=p.name
		;
		/* do sth if input names dont match existing ones - exception or create them....*/
	if groupids.count<>securityToken.group_names.count then

--then select....
		select name
		bulk collect into newNames
		from
				table(cast(securityToken.group_names as principalNameListT)) a
					where not exists
						(select name from perm_principal where name=a.name);
		--dbms_output.put_line('new count: '||newnames.count);
		--always creates nonzero collection
		--dbms_output.put_line('count'||newnames.count);
		if insertop<>0 then
		forall i in 1..newNames.last
			insert into perm_principal(principal_id,name,p_type --,defaultUser,defaultGroup
			)
				values(fc_principal_id_seq.nextval,newNames(i),'g'--,0,0
				)
				returning name,principal_id bulk collect into newprincipals
				;
		end if;
	if groupids.count=0 then
		groupids:=newprincipals;
	else
		--extend it with new values
		gSize:=groupids.count;
		for i in 1+gSize..newprincipals.last+gSize loop
			groupids(i):=newprincipals(i-gSize);
		end loop;
	end if;
   end if;

end;


/* returns name,principal_id pairs for principal names
Object version*/
procedure getPrincipalIdsO(securityToken in principalToken,
	userId out nocopy perm_principal.principal_id%type,
	groupIds out nocopy PermStringIntPairOT,
	insertop in pls_integer,
  superuser in boolean)

as

userType pp.p_type%type:='u';
begin

	begin
		/* get user id */
		select principal_id
			into userId
		from perm_principal
			where
			name=securityToken.user_name
	--		and p_type='u'
			;
	exception
		when no_data_found then
     declare
     ret returnO;
     outerrors returnOT:=returnOT();
     users PermStringIntPairOT ;
     homeDirPath varchar2(32000);
     -- TODO: move it to the admin package
     begin

		 if insertop=1 and securityToken is not null and securityToken.user_name is not null then

      if superuser=true then userType:='a'; end if;
			insert into perm_principal(principal_id,name,p_type--,defaultUser,defaultGroup
				)
					values(fc_principal_id_seq.nextval,securityToken.user_name,userType)
					returning principal_id into userId;

        --initialise defaults for the new user
        -- do this only for the primary
        -- others would use global defaults as usual unless sbd dont
        -- set defaults for them

       -- create home directory as superuser

       -- created directories should have perms for principal only
       -- group should be the 1st from the 1st from the list
       --TODO: may change principalnameListT to listNum globally...
       select PermStringIntPairO(null,principal_id)
        bulk collect into users
        from pp where name='Admin';

        
       homeDirPath := replace(homeDirectoryRoot||securityToken.user_name,' ','_')||'/default';
       ret:=fireman.mkdirParents (
          null, -- we used default token for this so no need for sec token
          userId,
          users,
          lfnt(homeDirPath),
          0,--get permissions from defaults
          outerrors,
          true
          );
       -- check errors, if exists already mkdirparents skips error..

       declare
       permGuid fc.guid%type;
       parentDirId fc.parent_id%type;
       defaultPermission PermissionO ; -- taken from defaults
       perm PermissionEntryT;
       begin
       permGuid:=generateGuid;
       SELECT /*+ cardinality (inP 5)*/
        s.perm
        COLLECT INTO defaultPermission
        FROM
          pd s
          WHERE  s.guid = globalDefaultGuid
          ;

       -- get the default parent after construction
       begin
          SELECT file_id      
          collect into parentDirId
          FROM
          FC_FILE f
          WHERE LEVEL>fireman.getNoOfPathDelimiters(homeDirPath) --take both but just one permisson
          START WITH parent_id=0 AND file_id=1
          CONNECT BY  
          PRIOR file_id = parent_id --!!!! check if 'd' only valid paths..
          AND  NAME=fireman.getpathtoken(LEVEL-1,homeDirPath)
          ;
        exception 
          when NO_DATA_FOUND then      
            parentDirId:=1;
        end;

       	INSERT INTO pd(principal_id,guid,perm,defaultParent)
          VALUES(userId,permGuid,defaultPermission,parentDirId );
                                                          -- create it in the /VO/userPrincipal, dont really check for permissions
                                                          -- use it
                                                          -- defaultParent, hack for now

          -- This should come from global defaults
        -- change global permission to be of the owner
        defaultPermission.userName:=securityToken.user_name;
        defaultPermission.groupName:=securityToken.user_name;
        perm:=
        PermissionEntryT
        (PermissionEntryO(permGuid,defaultPermission));
        --set security on it
        ret:=addGuid(
             securityToken,
             perm,
             outerrors);
      end ;

		else
			userId:=null;
		end if;
    end;
	end;

	/* get ids of all user's groups'*/

  select /*+ cardinality (n 5) */ PermStringIntPairO(n.name, p.principal_id)
    bulk collect into groupids
      from
      perm_principal p,
      table(securityToken.group_names) n
      where
      n.name=p.name
      ;
		/* do sth if input names dont match existing ones - exception or create them....*/
	if (securityToken is not null) and (groupids is null or groupids.count<>securityToken.group_names.count) and insertop=1 then
  declare
  newNames namesT;
  newPrincipals PermStringIntPairOT;
  gSize pls_integer;
  begin
    if superuser=true then userType:='a'; else userType:='g'; end if;
--then select....
		select name
		bulk collect into newNames
		from
				table(cast(securityToken.group_names as principalNameListT)) a
					where not exists
						(select name from perm_principal where name=a.name);
		--dbms_output.put_line('new count: '||newnames.count);
		--always creates nonzero collection
		--dbms_output.put_line('count'||newnames.count);


		forall i in 1..newNames.last
			insert into perm_principal(principal_id,name,p_type--,defaultUser,defaultGroup
			)
				values(fc_principal_id_seq.nextval,newNames(i),userType--,0,0
				)
				returning PermStringIntPairO(name,principal_id)
				bulk collect into newprincipals
				;

    -- for new users do not create home dirs - only originating user
    -- TODO: decide what to do..


  	if groupids.count=0 then
      groupids:=newprincipals;
    else
		--extend it with new values
		gSize:=groupids.count;
		groupids.extend(gSize);
		for i in 1+gSize..newprincipals.last+gSize loop
			groupids(i):=newprincipals(i-gSize);
		end loop;
    end if;
    end;
		end if; --insertop

end;

-- FIXME: change check order, null type first
/* returns id for passed name, type */
function getPrincipalId(pname in perm_principal.name%type, ptype in char, insertop in pls_integer)
	return perm_principal.principal_id%type
as
  PRAGMA AUTONOMOUS_TRANSACTION;
  pnameTmp perm_principal.name%type;
	ret perm_principal.principal_id%type;
begin
  if pname is null and ptype='g' then pnameTmp:='egee-group';else pnameTmp:=pname;end if;
	case ptype
		when 'u' then
			select principal_id
				into ret
			from perm_principal
				where
				name=pnameTmp
				and p_type='u'
				;
		when 'g' then
			select principal_id
				into ret
			from perm_principal where
				name=pnameTmp
				and p_type='g';
		else
			select principal_id
				into ret
			from perm_principal where
				name=pnameTmp 		;
	 end case;
	 return ret;
exception
	when NO_DATA_FOUND then
  begin
	if insertop<>0 then
		insert into perm_principal(principal_id,name,p_type --,defaultUser,defaultGroup
		)
			values(fc_principal_id_seq.nextval,pnameTmp,ptype--,0,0
			)
			returning principal_id into ret;
      commit;
	end if;
		return ret;
  exception
    when DUP_VAL_ON_INDEX then
      case ptype
      when 'u' then
        select principal_id
          into ret
        from perm_principal
          where
          name=pnameTmp
          and p_type='u'
          ;
      when 'g' then
        select principal_id
          into ret
        from perm_principal where
          name=pnameTmp
          and p_type='g';
      else
        select principal_id
          into ret
        from perm_principal where
          name=pnameTmp 		;
     end case;
     return ret;
  end;

end;


/* generate user in same trans */
function getPrincipalIdIUW(pname in perm_principal.name%type, ptype in char, insertop in pls_integer)
	return perm_principal.principal_id%type
as
  pnameTmp perm_principal.name%type;
	ret perm_principal.principal_id%type;
begin
  if pname is null then pnameTmp:='nobody';else pnameTmp:=pname;end if;
	case ptype
		when 'u' then
			select principal_id
				into ret
			from perm_principal
				where
				name=pnameTmp
				and p_type='u'
				;
		when 'g' then
			select principal_id
				into ret
			from perm_principal where
				name=pnameTmp
				and p_type='g';
		else
			select principal_id
				into ret
			from perm_principal where
				name=pnameTmp 		;
	 end case;
	 return ret;
exception
	when NO_DATA_FOUND then
  begin
	if insertop<>0 then
		insert into perm_principal(principal_id,name,p_type --,defaultUser,defaultGroup
		)
			values(fc_principal_id_seq.nextval,pnameTmp,ptype--,0,0
			)
			returning principal_id into ret;
	end if;
		return ret;
  exception
    when DUP_VAL_ON_INDEX then
      case ptype
      when 'u' then
        select principal_id
          into ret
        from perm_principal
          where
          name=pnameTmp
          and p_type='u'
          ;
      when 'g' then
        select principal_id
          into ret
        from perm_principal where
          name=pnameTmp
          and p_type='g';
      else
        select principal_id
          into ret
        from perm_principal where
          name=pnameTmp 		;
     end case;
     return ret;
  end;

end;

/************ getacl obj for a given guid ************/
function getACL(g in varchar2)
	return ACLList
as
	ret ACLList:=ACLList();
	tmp ACLtbl; --plsql
begin
	select ACLEntryO(p.name,acl.perm)
	bulk collect into ret
	from acl,perm_principal p
		where guid=g and p.principal_id=acl.principal_id;
	return ret;
end;


/* returns name,type for a given principal_id	*/
function getPrincipalName(pid in integer)
	return principalName
as
tmpR principalNameType;
begin

	select p.name,p.p_type into tmpR.name,tmpR.p_type
		from perm_principal p
			where
			principal_id=pid;

	return principalName(tmpR.name);
end;



/* returns name,type for a given principal_id	*/
function getPrincipalNameStr(pid in integer)
	return varchar2
as
-- nameU perm_principal.NAME%type;
nameU varchar2(255);

begin

	select p.name into nameU
		from perm_principal p
			where
			principal_id=pid;

  return nameU;
exception
		when no_data_found then
		return 'nobody';
	end;  
	





/* returns name,type for a given principal_id	*/
function getPrincipalNameS(pids in principalListIdT)
	return principalNameListT
	pipelined

as
tmp principalNameTypeT;
ret principalNameListT:=principalNameListT();
param principalName;
begin

	select name,p_type bulk collect into tmp
		from
			perm_principal p,
			table(cast(pids as principalListIdT)) input
			where
			principal_id=input.val;

	ret.extend(tmp.count);
	for i in 1..tmp.count loop
		param:=principalName(tmp(i).name);
--		ret(i):=;
		pipe row(param);
	end loop;

	return;
end;


/* get acllist for a given guid list acllist by aclilist


/* check permissions of principal for a list of guid param, op operation to be checked against issuing user
	return [gguid,permok] array
	Note:
		Ownership is an issue
*/
procedure checkPrincipalPermissions(
				securityToken in principalToken,
				data in out nocopy dataAndStatusT,
				userPrincipal out nocopy perm_principal.principal_id%type,
				principals out nocopy PermStringIntPairT,
				op in pls_integer)
as

principalsO PermStringIntPairOT;	-- dn and internal id objects

dataO dataAndStatusOT;					--	guid and if permitted flag objects
dataOTmp dataAndStatusOT:=dataAndStatusOT();				--	guid and if permitted flag objects

allBasicOk pls_integer:=1;
k pls_integer:=0;
one pls_integer:=0;


begin
	/* transalte names->internal ids	*/
	getPrincipalIdsO(securityToken, userPrincipal, principalsO,0);
	dataO := dataAndStatusT2SQL(data);

/* get basic permission permissions	*/
/* Returns status for all, even nonexisting, and no permitted ones.
  Works but may be expensive for bigger input sets

  1- 0k
  0 - not permitted
  2 - does not exist
*/


select /*+ cardinality( i 100 ) cardinality (inP 10)  */
	i.guid,
	isOkAcl(userperm,groupperm,otherperm,
		bp.principal_id,userPrincipal,item2,
      principalsO,
		i.guid,op
		) permitted
 bulk collect into data
  from
 ( (table(dataO) i left outer join bp on i.guid=bp.guid)
			  left outer join table(principalsO) inP
				 on bp.group_id=inP.item2)
			;


end;
/* new, external  version */

/* check permissions of principal for a list of guid param, op operation to be checked against issuing user
	return [gguid,permok] array
	Note:
		Ownership is an issue
*/
procedure checkPrincipalPermissions(
				securityToken in principalToken,
				data in out nocopy dataAndStatusT,
				op in pls_integer)
as

principalsO PermStringIntPairOT;	-- dn and internal id objects

	dataO dataAndStatusOT;					--	guid and if permitted flag objects
	dataOTmp dataAndStatusOT:=dataAndStatusOT();				--	guid and if permitted flag objects

	userPrincipal  perm_principal.principal_id%type;
	principals  PermStringIntPairT;

	allBasicOk pls_integer:=1;
	k pls_integer:=0;
	one pls_integer:=0;

begin
	/* transalte names->internal ids	*/
	getPrincipalIdsO(securityToken, userPrincipal, principalsO,0);


	dataO := dataAndStatusT2SQL(data);


/* get basic permission permissions	*/
/* Returns status for all, even nonexisting, and no permitted ones.
  Works but may be expensive for bigger input sets

  1- 0k
  0 - not permitted
  2 - does not exist
*/

select /*+  cardinality( i 100 ) cardinality (inP 10)  */
	i.guid,
	isOkAcl(userperm,groupperm,otherperm,
		bp.principal_id,userPrincipal,item2,
      principalsO,
		i.guid,op
		) permitted
 bulk collect into data
  from
 ( (table(dataO) i left outer join bp on i.guid=bp.guid)
			  left outer join table(principalsO) inP
				 on bp.group_id=inP.item2)
			;


end;
/* end of new version */

/* version with guidO so no conversion needed */
/* check permissions of principal for a list of guid param, op operation to be checked against issuing user
	return [gguid,permok] array
	Note:
		Ownership is an issue
*/
procedure checkPrincipalPermissions(
				securityToken in principalToken,
				guids in guidOT,
				data out nocopy dataAndStatusT,
				op in pls_integer)
as

   principalsO PermStringIntPairOT;	-- dn and internal id objects

	dataO dataAndStatusOT;					--	guid and if permitted flag objects
	dataOTmp dataAndStatusOT:=dataAndStatusOT();				--	guid and if permitted flag objects

	userPrincipal  perm_principal.principal_id%type;
	principals  PermStringIntPairT;

	allBasicOk pls_integer:=1;
	k pls_integer:=0;
	one pls_integer:=0;

begin
	/* transalte names->internal ids	*/
	getPrincipalIdsO(securityToken, userPrincipal, principalsO,0);



/* get basic permission permissions	*/
/* Returns status for all, even nonexisting, and no permitted ones.
  Works but may be expensive for bigger input sets

  1- 0k
  0 - not permitted
  2 - does not exist
*/


select /*+  cardinality( i 100 ) cardinality (inP 10)  */
	i.guid,
	isOkAcl(userperm,groupperm,otherperm,
		bp.principal_id,userPrincipal,item2,
      principalsO,
		i.guid,op
		) permitted
 bulk collect into data
  from
 ( (table(guids) i left outer join bp on i.guid=bp.guid)
			  left outer join table(principalsO) inP
				 on bp.group_id=inP.item2)
			;

end;

/* end of guidOT version */

/* version with errors and guidOT handling so no conversion needed */
/* check permissions of principal for a list of guid param, op operation to be checked against issuing user
	return [gguid,permok] array
	Note:
		Ownership is an issue
*/
procedure checkPrincipalPermissions(
				securityToken in principalToken,
				guids in guidOT,
				data out nocopy dataAndStatusT,
				outerrors out nocopy returnOT,
				op in pls_integer)
as

--	principalsO PermStringIntPairOT:=PermStringIntPairOT();	-- dn and internal id objects
	principalsO PermStringIntPairOT;	-- dn and internal id objects

	dataO dataAndStatusOT;					--	guid and if permitted flag objects
	dataOTmp dataAndStatusOT:=dataAndStatusOT();				--	guid and if permitted flag objects

	userPrincipal  perm_principal.principal_id%type;
	principals  PermStringIntPairT;

	allBasicOk pls_integer:=1;
	k pls_integer:=0;
	one pls_integer:=0;

begin
	/* transalte names->internal ids	*/
	getPrincipalIdsO(securityToken, userPrincipal, principalsO,0);


/* get basic permission permissions	*/
/* Returns status for all, even nonexisting, and no permitted ones.
  Works but may be expensive for bigger input sets

  1- 0k
  0 - not permitted
  2 - does not exist
*/

select /*+  cardinality( i 100 ) cardinality (inP 10)  */
	i.guid,
	isOkAcl(userperm,groupperm,otherperm,
		bp.principal_id,userPrincipal,item2,
      principalsO,
		i.guid,op
		) permitted
 bulk collect into data
  from
 ( (table(guids) i left outer join bp on i.guid=bp.guid)
			  left outer join table(principalsO) inP
				 on bp.group_id=inP.item2)
			;

	--handle errors
	--we must take into account input order was lost...
	for i in 1..data.last loop
		if data(i).permitted=0 then
			outerrors.extend;
			outerrors(i):=
					returnO
					(
					data(i).guid,
					0,	--temporal solution. we should probably remember input guids ordering but callee does it anyway
					accessDeniedErrVal,
					accessDeniedErrStr
					);
		else
			/*data is ok...
			if we just rebuild data here we have to repeat that loop outside
			to deserialize proper items...
			*/
			null;
		end if ;
	end loop;

end;

/* end of version with errors and guidOT version */


/* decoder for [permission,operation] returns binary value -allowed/not allowed */
/* operations are:
	write - create new or write to or change stats, add new surl(replicas)
	read - read the file, for dir get dir stat and
	execute -  not used for items, for dirs traverse...
	list - read stats of files,
	remove - deletion of the guid, deletion of the surl
	permission - change permissions

*/

function checkPerm
(
permVal in bp.USERPERM%type
,op in pls_integer
)
	return pls_integer
	deterministic
as
permitted pls_integer;

BEGIN
	case op
		when writeFileOp then permitted := bitand(permVal,writemask);
		when writeDirOp then  permitted := bitand(permVal,writemask);
		when readFileOp then  permitted := bitand(permVal,readmask);
		when excecuteFileOp then  permitted := bitand(permVal,execmask);
		when listFileOp then  permitted := bitand(permVal,listmask);
		when listDirOp then  permitted := bitand(permVal,listmask);
		when removeFileOp then  permitted := bitand(permVal,removemask);
		when permissionOp then  permitted := bitand(permVal,permmask);
	end case;
	--or readDir
if (permitted>0) then
	return 1;
else
	return 0;
end if;

end;


function getOpMask(op in pls_integer)
return pls_integer
deterministic
as
begin
	case op
		when writeFileOp then return writemask;
		when writeDirOp then  return writemask;
		when readFileOp then  return readmask;
		when readDirOp then  return readmask;
		when excecuteFileOp then  return execmask;
		when listFileOp then  return listmask;
		when listDirOp then  return listmask;
		when removeFileOp then  return removemask;
		when permissionOp then  return permmask;
    when setMetaOp then  return setmetamask;
    when getMetaOp then  return getmetamask;
	end case;
end;

/******** complex check for all perms isOkBasicPerm 	********/
function isOkBasicPerm(op in pls_integer, u in pls_integer, g in pls_integer, o in pls_integer)
	return decimal
	deterministic
as
permitted pls_integer;
BEGIN
	if (checkPerm(u,op)>0 or checkPerm(g,op)>0 or checkPerm(o,op)>0) then return 1;	end if;
	return 0;
end;

function isOk(u in pls_integer, g in pls_integer, o in pls_integer)
	return pls_integer
	deterministic
as
permitted pls_integer;
BEGIN
	if (u=1 or g=1 or o=1) then return 1;
	elsif (u is null) then return 2;
	else
		return 0;
	end if;
end;

function isOk(u in pls_integer, g in pls_integer, o in pls_integer,
	p_id in pls_integer,p_id_in in pls_integer,
	g_id in pls_integer)
	return pls_integer
	deterministic
as
permitted pls_integer;
BEGIN
	if ((u=1 and p_id=p_id_in)
	or (g=1 and g_id is not null)
	or o=1)
	then return 1;
	elsif u is null then return 2;
	else
		return 0;
	end if;
end;

--original one
function isOkAclJoin(u in pls_integer, g in pls_integer, o in pls_integer,
	p_id in pls_integer,p_id_in in pls_integer,
	g_id in pls_integer,
	principals in PermStringIntPairOT,
	iguid in perm_acl_permission.guid%type,
	op in pls_integer
	)
	return pls_integer
as
permitted pls_integer;
mask pls_integer;
BEGIN
  debugOn();
	mask:=getOpMask(op);
	if ((sign(bitand(u,mask))=1 and p_id=p_id_in)
	or (sign(bitand(g,mask))=1 and g_id is not null)
	or sign(bitand(o,mask))=1)
	then return 1;
	elsif u is null then return 2;
	elsif u is not null then-- check acl
  SELECT COUNT(*)
    into permitted
    FROM dual WHERE EXISTS
    (
    SELECT /*+ cardinality (inP2 5) */  NULL
      FROM
        acl a,
        TABLE(principals) inP2
        WHERE
        (a.guid=iguid
        AND (inP2.item2=a.principal_id OR a.principal_id=p_id_in)
        AND SIGN(BITAND(perm,mask))=1))
    or exists
    (
    select /*+ cardinality (inP2 5) cardinality (su 5) */ null
    from
    TABLE(superUsers) su,
    TABLE(principals) inP2
				where su.column_value=inP2.item2 or su.column_value=p_id_in)
    ;

	return permitted;

	else
		return 0;
	end if;
end;

function isOkAclMaskJoin(u in pls_integer, g in pls_integer, o in pls_integer,
	p_id in pls_integer,p_id_in in pls_integer,
	g_id in pls_integer, --group_id from 
	principals in PermStringIntPairOT,
	iguid in perm_acl_permission.guid%type,
	mask in pls_integer
	)
	return pls_integer
as
  permitted pls_integer;
BEGIN
  --debugOn();
 	if ((bitand(u,mask)=mask) and (p_id=p_id_in))
	or ((bitand(g,mask)=mask) and (g_id is not null))
	or (bitand(o,mask)=mask)
	then return 1;
	elsif u is null then return 2;
	elsif u is not null then-- check acl
  SELECT COUNT(*)
    into permitted
    FROM dual WHERE EXISTS
    (
    SELECT /*+ cardinality (inP2 10) */  NULL
      FROM
        acl a,
        TABLE(principals) inP2
        WHERE
        (a.guid=iguid
        AND (inP2.item2=a.principal_id OR a.principal_id=p_id_in)
        AND SIGN(BITAND(perm,mask))=1))
    or exists
    (
    select /*+ cardinality (inP2 5) cardinality (su 5) */ null
    from
    TABLE(superUsers) su,
    TABLE(principals) inP2
				where su.column_value=inP2.item2 or su.column_value=p_id_in)
    ;
	return permitted;
	else
		return 0;
	end if;
end;




function isOkAcl(u in pls_integer, g in pls_integer, o in pls_integer,
	p_id in pls_integer,p_id_in in pls_integer,
	g_id in pls_integer,
	principals in PermStringIntPairOT,
	iguid in perm_acl_permission.guid%type,
	op in pls_integer
	)
	return pls_integer
	--deterministic
as
permitted pls_integer;
mask pls_integer;
BEGIN

	mask:=getOpMask(op);
	if ((sign(bitand(u,mask))=1 and p_id=p_id_in)
	or (sign(bitand(g,mask))=1 and g_id is not null)
	or sign(bitand(o,mask))=1)
	then return 1;
	elsif u is null then return 2;
	elsif u is not null then-- check acl
  SELECT COUNT(*)
    into permitted
    FROM dual WHERE EXISTS
    (
    SELECT /*+ cardinality (inP2 10) */  NULL
      FROM
        acl a,
        TABLE(principals) inP2
        WHERE
        (a.guid=iguid
        AND (inP2.item2=a.principal_id OR a.principal_id=p_id_in)
        AND SIGN(BITAND(perm,mask))=1))
    or exists
    (
    select /*+ cardinality (inP2 5) cardinality (su 5) */ null
    from
    TABLE(superUsers) su,
    TABLE(principals) inP2
				where su.column_value=inP2.item2 or su.column_value=p_id_in)
    ;
  
	return permitted;

	else
		return 0;
	end if;
end;


-- should return error array and return type
-- return type should be resolver array?
function isOkAcl(
  resolver in ResolverT,  
	p_id_in in pls_integer,	
	principals in PermStringIntPairOT,	
	op in pls_integer
	)
	return returnOT
--	deterministic
as
ret returnOT;
mask pls_integer;
BEGIN

	mask:=getOpMask(op);
  
--lfns(i),i,Permission_Util.accessDeniedErrVal,Permission_Util.accessDeniedErrStr  
  
  select 
  case 
    when ((sign(bitand(i.userperm,mask))=1 and i.principal_id=p_id_in)
          or (sign(bitand(i.groupperm,mask))=1 and i.group_id is not null)
          or sign(bitand(i.otherperm,mask))=1)
          then returnO(1)
    when i.userperm is null then returnO(2)
    else
      (SELECT returnO(COUNT(*))
        FROM dual WHERE 
        EXISTS
        (
        SELECT /*+ cardinality (inP2 10) */  NULL
          FROM
            acl a,
            TABLE(principals) inP2
            WHERE
            (a.guid=i.guid
            AND (i.group_id=a.principal_id OR a.principal_id=p_id_in)
            AND SIGN(BITAND(perm,mask))=1))
        or exists
        (
        select /*+ cardinality (inP2 5) cardinality (su 5) */ null
        from
        TABLE(superUsers) su,
        TABLE(principals) inP2
            where su.column_value=inP2.item2 or su.column_value=p_id_in
        )
      ) 
  end --case
   as rets
  bulk collect into ret 
  from 
  table(resolver) i;
    
  return ret;
end;

function isOkAclX (
  res in ResolverO,
	p_id_in in pls_integer,
	principals in PermStringIntPairOT,
	op in pls_integer
	)
	return resolverO
as
mask pls_integer;
ret resolverO;

BEGIN

	mask:=permission_util.getOpMask(op);
	if ((sign(bitand(res.userperm,mask))=1 and res.principal_id=p_id_in)
	or (sign(bitand(res.groupperm,mask))=1 and res.group_id is not null)
		or sign(bitand(res.otherperm,mask))=1)
	then return resolverO(res,1);
	elsif res.file_id is null then return res;
  elsif res.permission=permission_util.loopInSymlinksDetectedVal then    
  return res;
	elsif res.userperm is not null then-- check acl
  SELECT resolverO(res,COUNT(*))
    into ret
    FROM dual WHERE 
    EXISTS
    (
    SELECT /*+ cardinality (inP2 10) */  NULL
      FROM
        acl a,
        TABLE(principals) inP2
        WHERE
        (a.guid=res.guid
        AND (inP2.item2=a.principal_id OR a.principal_id=p_id_in)
        AND SIGN(BITAND(perm,mask))=1)
    )
     or exists
    (
      select /*+ cardinality (inP2 5) cardinality (su 5) */ null
      from
      TABLE(superUsers) su,
      TABLE(principals) inP2
          where su.column_value=inP2.item2 or su.column_value=p_id_in
     )      
    ;
    	return ret;
	else
		ret.permission:=0;
		return ret;
	end if;

end;


function isOkAclXMask (
  res in ResolverO,
	p_id_in in pls_integer,
	principals in PermStringIntPairOT,
	mask in pls_integer
	)
	return resolverO	
as

ret resolverO;
BEGIN

  debugon();
	if ((sign(bitand(res.userperm,mask))=1 and res.principal_id=p_id_in)
	or (sign(bitand(res.groupperm,mask))=1 and res.group_id is not null)
		or sign(bitand(res.otherperm,mask))=1)
	then return resolverO(res,1);
	elsif res.file_id is null then return res;
  elsif res.permission=permission_util.loopInSymlinksDetectedVal then    
  return res;
	elsif res.userperm is not null then-- check acl
  SELECT resolverO(res,COUNT(*))
    into ret
    FROM dual WHERE EXISTS
    (
    SELECT /*+ cardinality (inP2 5) */  NULL
      FROM
        acl a,
        TABLE(principals) inP2
        WHERE
        (a.guid=res.guid
        AND (inP2.item2=a.principal_id OR a.principal_id=p_id_in)
        AND SIGN(BITAND(perm,mask))=1))
         or exists
        (
        select /*+ cardinality (inP2 5) cardinality (su 5) */ null
        from
        TABLE(superUsers) su,
        TABLE(principals) inP2
            where su.column_value=inP2.item2 or su.column_value=p_id_in
        )
      
    ;
    	return ret;
	else
		ret.permission:=0;
		return ret;
	end if;

end;



--
--function isOkAcl(
--  resolver in ResolverT,  
--	p_id_in in pls_integer,	
--	principals in PermStringIntPairOT,	
--	op in pls_integer
--	)
--	return returnOT
----	deterministic
--as
--ret returnOT;
--mask pls_integer;
--BEGIN
--
--	mask:=getOpMask(op);
--  
----lfns(i),i,Permission_Util.accessDeniedErrVal,Permission_Util.accessDeniedErrStr  
--  
--  select 
--  case 
--    when ((sign(bitand(i.userperm,mask))=1 and i.principal_id=p_id_in)
--          or (sign(bitand(i.groupperm,mask))=1 and i.group_id is not null)
--          or sign(bitand(i.otherperm,mask))=1)
--          then returnO(1)
--    when i.userperm is null then returnO(2)
--    else
--      (SELECT returnO(COUNT(*))
--        FROM dual WHERE 
--        EXISTS
--        (
--        SELECT /*+ cardinality (inP2 10) */  NULL
--          FROM
--            acl a,
--            TABLE(principals) inP2
--            WHERE
--            (a.guid=i.guid
--            AND (i.group_id=a.principal_id OR a.principal_id=p_id_in)
--            AND SIGN(BITAND(perm,mask))=1))
--        or exists
--        (
--        select /*+ cardinality (inP2 5) cardinality (su 5) */ null
--        from
--        TABLE(superUsers) su,
--        TABLE(principals) inP2
--            where su.column_value=inP2.item2 or su.column_value=p_id_in
--        )
--      ) 
--  end --case
--   as rets
--  bulk collect into ret 
--  from 
--  table(resolver) i;
--    
--  return ret;
--end;


function isOkAclMask(u in pls_integer, g in pls_integer, o in pls_integer,
	p_id in pls_integer,p_id_in in pls_integer,
	g_id in pls_integer,
	principals in PermStringIntPairOT,
	iguid in perm_acl_permission.guid%type,
	mask in pls_integer
	)
	return pls_integer
as
permitted pls_integer;
BEGIN
  debugOn();
 	if (bitand(u,mask)=mask and p_id=p_id_in)
	or (bitand(g,mask)=mask and g_id is not null)
	or (bitand(o,mask)=mask)
	then return 1;
	elsif u is null then return 2;
	elsif u is not null then-- check acl
  SELECT COUNT(*)
    into permitted
    FROM dual WHERE 
    exists 
    (
      select null
         from
          table(principals) inP2
          where inP2.item2=g_id AND SIGN(BITAND(g,mask))=1
    )
    or EXISTS
    (
      SELECT /*+ cardinality (inP2 10) */  NULL
        FROM
          acl a,
          TABLE(principals) inP2
          WHERE
          (a.guid=iguid
          AND (inP2.item2=a.principal_id OR a.principal_id=p_id_in)
          AND SIGN(BITAND(perm,mask))=1)
    )
    or exists
    (
    select /*+ cardinality (inP2 5) cardinality (su 5) */ null
    from
    TABLE(superUsers) su,
    TABLE(principals) inP2
				where su.column_value=inP2.item2 or su.column_value=p_id_in
    )
    ;

	return permitted;

	else
		return 0;
	end if;
end;

function isOkAcl2(u in pls_integer, g in pls_integer, o in pls_integer,
	p_id in pls_integer,p_id_in in pls_integer,
	g_id in pls_integer,	gexists in pls_integer,
	principals in PermStringIntPairOT,
	iguid in perm_acl_permission.guid%type,
	op in pls_integer
	)
	return PermissionEntryO
--	deterministic
as
permitted pls_integer;
mask pls_integer;
BEGIN

	mask:=getOpMask(op);
	if ((sign(bitand(u,mask))=1 and p_id=p_id_in)
	or (sign(bitand(g,mask))=1 and gexists is not null)
	or sign(bitand(o,mask))=1)
	then return
		PermissionEntryO(iguid,
		PermissionO(
			getPrincipalNameStr(p_id),
			getPrincipalNameStr(g_id),
			u, g, o,
			(getAcl(iguid))));
	elsif u is null then return PermissionEntryO(iguid,null);
	elsif u is not null then-- check acl
  SELECT COUNT(*)
    into permitted
    FROM dual WHERE EXISTS
    (
    SELECT /*+ cardinality (inP2 10) */  NULL
      FROM
        acl a,
        TABLE(principals) inP2
        WHERE
        (a.guid=iguid
        AND (inP2.item2=a.principal_id OR a.principal_id=p_id_in)
        AND SIGN(BITAND(perm,mask))=1))
    or exists
    (
    select /*+ cardinality (inP2 5) cardinality (su 5) */ null
    from
    TABLE(superUsers) su,
    TABLE(principals) inP2
				where su.column_value=inP2.item2 or su.column_value=p_id_in)
    ;

		if permitted<>0 then
		return PermissionEntryO(iguid,
		PermissionO(
			getPrincipalNameStr(p_id),
			getPrincipalNameStr(g_id),
			u, g, o,
			(getAcl(iguid))));
		else
			return PermissionEntryO(iguid,null);
		end if;
	else
		return PermissionEntryO(iguid,null);
	end if;
end;

function isOkAclPermission(u in pls_integer, g in pls_integer, o in pls_integer,
	p_id in pls_integer,p_id_in in pls_integer,
	g_id in pls_integer,	gexists in pls_integer,
	principals in PermStringIntPairOT,
	iguid in perm_acl_permission.guid%type,
	op in pls_integer
	)
	return PermissionO
--	deterministic
as
permitted pls_integer;
mask pls_integer;
per PermissionO;
BEGIN
	mask:=getOpMask(op);
	if ((sign(bitand(u,mask))=1 and p_id=p_id_in)
	or (sign(bitand(g,mask))=1 and gexists is not null)
	or sign(bitand(o,mask))=1)
	then
	per:=PermissionO(
			getPrincipalNameStr(p_id),
			getPrincipalNameStr(g_id),
			u, g, o,
			getAcl(iguid));

	return per;

	elsif u is null then

	return null;

	elsif u is not null then-- check acl
-- create cursor, exit on first succesfull or return null obj
-- or leave it as max query - we dont know what would be faster, max() on average for sure.

  SELECT COUNT(*)
    into permitted
    FROM dual WHERE EXISTS
    (
    SELECT /*+ cardinality (inP2 10) */  NULL
      FROM
        acl a,
        TABLE(principals) inP2
        WHERE
        (a.guid=iguid
        AND (inP2.item2=a.principal_id OR a.principal_id=p_id_in)
        AND SIGN(BITAND(perm,mask))=1))
    or exists
    (
    select /*+ cardinality (inP2 5) cardinality (su 5) */ null
    from
    TABLE(superUsers) su,
    TABLE(principals) inP2
				where su.column_value=inP2.item2 or su.column_value=p_id_in)
    ;

		if permitted<>0 then
		return
		PermissionO(
			getPrincipalNameStr(p_id),
			getPrincipalNameStr(g_id),
			u, g, o,
			(getAcl(iguid)));
		else
			return null;
		end if;
	else
		return null;
	end if;
end;


/**************** C++ max() with nulls   ************/
function maximum(a in pls_integer, b in pls_integer)
	return decimal
	deterministic
as
begin
	if a>b or b is null then return a;
		else
		return b;
	end if;
end;

/***************** translate in bulk *************/
function translateAclToSimpleAcl
  ( nameAcl in  ACLList)
  return ACLSimpleIdList
  deterministic
  as
  outret ACLSimpleIdList;
  begin
    if nameACL is null or nameACl.count=0 then return null;end if;
    select /*+ cardinality (a 2) */
    ACLEntrySimpleId(p.principal_id, a.perm)
    bulk collect into outRet
    from table(nameACl) a , perm_principal p
    where p.name=a.principal;
    -- TODO: check if all from input are mapped - throw an error if not
    return outRet;
  end;

procedure translateAclToSimpleAcl
  ( nameAcl in  ACLList, outRet in out nocopy ACLSimpleIdList)
  deterministic
  as
  begin
    if nameACL is null or nameACl.count=0 then return ;end if;
    select /*+ cardinality (a 2) */
    ACLEntrySimpleId(p.principal_id, a.perm)
    bulk collect into outRet
    from table(nameACl) a , perm_principal p
    where p.name=a.principal;

  end;


/******************************************************************
*
*
*								testing junk
*
*
*******************************************************************/


/************ dbug/testing junk *****************/
function getprincipalNameListT
	return principalNameListT
as

begin
return principalNameListT
	(
	principalName('user_110'),
	principalName('kakaugrp1'),
	principalName('kakaugrp2')
	)
	;
end;

function getPermStringIntPairOT
	return PermStringIntPairOT
as
begin

	return PermStringIntPairOT
		(
      PermStringIntPairO('whatever',431),
			PermStringIntPairO('whatever',4),
			PermStringIntPairO('whatever',2),
			PermStringIntPairO('whatever',3),
			PermStringIntPairO('whatever',10),
      PermStringIntPairO('whatever',10000),
		PermStringIntPairO('whatever',10237)

		);

end;

function getSuperUsers
  return listnum
as 
begin
  return superUsers;
end;




procedure testSecurityArray(
	securityList in principalListT,
	permEntry  in PermissionEntryT,
	outPerm out nocopy PermissionEntryT,
	dataStatusT out nocopy dataAndStatusT)
	as
		resPerm pls_integer;

	type gT is table of pls_integer index by binary_integer;
	g gT;

	begin
		null;
		select userperm bulk collect into g
		from perm_basic_permission bp
		where bitand(bp.userperm,readmask)=readmask;
		-- securityList - holds list of dns, minimum one
		-- permEntry - item and Permission for it
		-- outPerm - array
--		forall k in securityList.first..securityList.last
--			merge into basic_permission bp
--			insert into perm_basic_permission values(
--				permEntry.item,
--				1,
--				1,
--				permEntry.PermissionR.basicperm. ,
--		acl ACLArrT
--);
--	userName varchar2(250),
--	groupName varchar2(250),
--	userPerm permR,
--	groupPerm permR,
--	otherPerm permR

	null;
	dbms_output.put_line('size'||g.count);

	end;


function generateSecToken
	return principalToken
as
ret principalToken;
begin
	return null;
end;


/********* generate fake guid data	****************/
function generateData
	return dataAndStatusT
as
ret dataAndStatusT;
d dataAndStatus;
begin


	d.guid:='E2D16159-3D70-3B9D-E034-0003BA2F87BE';
	d.permitted:=0;

	ret(1):=d;
	d.guid:='774F509-E352-2BE4-E034-0003BA2F87BE';
	ret(2):=d;
	return ret;
end;


/* public generation */
function generateDataO(k in integer)
	return dataAndStatusOT
as
f integer;
tmp dataAndStatusOT;
begin
  f:=k;
  if f<5 then f:=5; end if;
	select dataAndStatusO(guid, mod(rownum,2)) checked
	bulk collect into tmp
		from
		fc
		where
		rownum<=f;


  TMP(1).GUID:='DOESNOTEXIST2cb-9606-e0f7c6ba478a';
  TMP(1).permitted:=null;
  TMP(2).GUID:='DOESNOTEXIST2cb-9606-e0f7c6ba478a';
  TMP(2).permitted:=null;
  tmp(3).guid:=tmp(4).guid;


	tmp(4).guid:='E2D16159-3D70-3B9D-E034-0003BA2F87BE';
	tmp(4).permitted:=0;

	tmp(5):=tmp(4);
	tmp(5).guid:='774F509-E352-2BE4-E034-0003BA2F87BES';

	return tmp;
end;


/*************** gen  permissionentry	**********/
function genPermEntry(limit in number)
	return PermissionEntryT
as
ret PermissionEntryT:=PermissionEntryT();
tmp dataAndStatusT;
begin
	select distinct guid,0 checked
	bulk collect into tmp
		from
		bp
		where
		rownum<=limit group by guid;


	ret.extend(tmp.count);
	for i in 1..tmp.count loop

	ret(i):=
		PermissionEntryO
		 (
		 tmp(i).guid,
			PermissionO('user_11','group_10150',63,33,1,
				ACLList(
					ACLEntryO('sidvicious_0',0),
					ACLEntryO('user_98',1),
					ACLEntryO('user_99',63)
						 )
					  )
		 );
	end loop;

	return ret;
end;

function getPermEntryACLIdList(limit in number)
	return ACLIdList
as
ret ACLIdList:=ACLIdList();
t PermissionEntryT;
j pls_integer:=0;
begin
	t:=genPermEntry(limit);

	for i in 1..t.count loop
	ret.extend(t(i).perm.acl.count);

			for ac in 1..t(i).perm.acl.count loop
			 j:=j+1;
			 ret(j):=ACLEntryIdO(
			 t(i).item,
--			 getPrincipalId(t(i).perm.acl(ac).principal,t(i).perm.acl(ac).acl_type),
			 getPrincipalIdIUW(t(i).perm.acl(ac).principal,'',0),
			 t(i).perm.acl(ac).perm
--			 ,t(i).perm.acl(ac).acl_type
			 );
--			 bulkAcl(j).guid:=permissions(i).item;
--			 bulkAcl(j).principal_id:=getPrincipalId(permissions(i).perm.acl(ac).principal,permissions(i).perm.acl(ac).acl_type);
--			 bulkAcl(j).perm:=permissions(i).perm.acl(ac).perm;
--			 bulkAcl(j).acl_type:=permissions(i).perm.acl(ac).acl_type;

			end loop;

	end loop;
--	DBMS_OUTPUT.PUT_LINE('ret size '||ret.count);
	return ret;
end;

function getAClList(item in varchar2)
	return ACLList
	as
ret ACLList:=ACLList();
begin
	return ret;
end;

procedure refreshSuperUsers
as
begin
select principal_id
  bulk collect into superUsers
  from perm_principal p
  where p_type='a';
end refreshSuperUsers;

begin
  refreshSuperUsers;
END;


