create or replace
PACKAGE BODY "FIREMAN" AS

/*
  $Name: glite-data-catalog-service-fr_R_2_1_15 $


*/

/********** rc types ***************/


globalDefaultGuid CONSTANT VARCHAR(36) DEFAULT '10000000-0000-0000-0000-000000000000';

/* serialisation types */
TYPE pairR IS RECORD (a INTEGER,b INTEGER);
--TYPE tripleR IS RECORD (a INTEGER,b INTEGER, c INTEGER);
TYPE pairsT IS TABLE OF pairR INDEX BY VARCHAR2(36);
--TYPE triplesT IS TABLE OF tripleR INDEX BY VARCHAR2(36);
TYPE bigPairsT IS TABLE OF pairR INDEX BY VARCHAR2(32000);



/* perms */
TYPE bulkPrincipalidT IS TABLE OF fc.principal_id%TYPE INDEX BY BINARY_INTEGER;
TYPE bulkGroupIdT IS TABLE OF fc.group_id%TYPE INDEX BY BINARY_INTEGER;
TYPE bulkUserPT IS TABLE OF fc.userperm%TYPE INDEX BY BINARY_INTEGER;
TYPE bulkGroupPT IS TABLE OF fc.groupperm%TYPE INDEX BY BINARY_INTEGER;
TYPE bulkOtherPT IS TABLE OF fc.otherperm%TYPE INDEX BY BINARY_INTEGER;
TYPE bulkGuidT IS TABLE OF fc.guid%TYPE INDEX BY BINARY_INTEGER;
type fileIdT is table of fc.file_id%TYPE INDEX BY pls_integer;



TYPE GuidSurlR IS RECORD
	(
	guid VARCHAR2(36),
	surl VARCHAR2(1024)
	)
	;
TYPE GuidSurlT IS TABLE OF GuidSurlR INDEX  BY BINARY_INTEGER;
TYPE GuidStatExtR IS RECORD
	(
		guid RC_REPLICA.GUID%TYPE,
		permitted DECIMAL(1),
		file_checksum RC_REPLICA.file_checksum%TYPE,
		status RC_REPLICA.status%TYPE,
		file_size RC_REPLICA.file_size%TYPE,
		cTime RC_REPLICA.cTime%TYPE,
		mTime RC_REPLICA.mTime%TYPE
	)
	;
TYPE GuidStatExtT IS TABLE OF GuidStatExtR INDEX  BY BINARY_INTEGER;

TYPE SurlEntryExtR IS RECORD
	(
	 guid RC_REPLICA.GUID%TYPE,
	 permitted DECIMAL(1),
	 surl VARCHAR2(1024),
	 master_replica SMALLINT,	--tb added
	 se VARCHAR2(128),
	 cTime TIMESTAMP,
	 mTime TIMESTAMP
	)
	;
TYPE SurlEntryExtRT IS TABLE OF SurlEntryExtR INDEX  BY BINARY_INTEGER;


TYPE surlEntryTablesT IS TABLE OF surlEntryT INDEX BY BINARY_INTEGER;

/* mapping to RCEntryO for select'ing  */
TYPE RCEntryR IS RECORD (
  guid VARCHAR2(36),
	guidStat GuidStatO,
	surlStats surlEntryT,
	permission PermissionO
  );
TYPE outEntries IS TABLE OF RCEntryR INDEX BY BINARY_INTEGER;

TYPE RCEntryRT IS TABLE OF RCEntryO INDEX BY BINARY_INTEGER;

TYPE outGuidStatsT IS TABLE OF GuidStatO INDEX BY BINARY_INTEGER;
TYPE outSurlStatsT IS TABLE OF SurlEntryT INDEX BY BINARY_INTEGER;
TYPE outPermissionsT IS TABLE OF PermissionO INDEX BY BINARY_INTEGER;


/*			for rcg				*/
TYPE gGuidsT IS TABLE OF fc.GUID%TYPE INDEX  BY BINARY_INTEGER;
TYPE file_checksumT IS TABLE OF fc.file_checksum%TYPE INDEX  BY BINARY_INTEGER;
TYPE statusT IS TABLE OF fc.status%TYPE INDEX  BY BINARY_INTEGER;
TYPE mTimeT IS TABLE OF fc.mTime%TYPE INDEX  BY BINARY_INTEGER;
TYPE file_sizeT IS TABLE OF fc.file_size%TYPE INDEX  BY BINARY_INTEGER;
TYPE cTimeT IS TABLE OF fc.cTime%TYPE INDEX  BY BINARY_INTEGER;


/*			for rcs				*/
TYPE ssurlT IS  TABLE OF RC_SURL.surl%TYPE INDEX  BY BINARY_INTEGER;
TYPE bulkPermittedT IS TABLE OF DECIMAL(1) INDEX  BY BINARY_INTEGER;

SUBTYPE rc_g IS fc%ROWTYPE;
TYPE rc_gT IS TABLE OF rc_g INDEX BY BINARY_INTEGER;
SUBTYPE rc_s IS RC_SURL%ROWTYPE;
TYPE rc_sT IS TABLE OF rc_s INDEX BY BINARY_INTEGER;

TYPE fc_T IS TABLE OF fc%ROWTYPE INDEX BY BINARY_INTEGER;
SUBTYPE fc_R IS fc%ROWTYPE;
TYPE acl_T IS TABLE OF acl%ROWTYPE INDEX BY BINARY_INTEGER;
TYPE surl_T IS  TABLE OF RC_SURL%ROWTYPE INDEX  BY BINARY_INTEGER;


TYPE tripleR IS RECORD (a INTEGER,b INTEGER, c INTEGER);
TYPE triplesT IS TABLE OF tripleR INDEX BY VARCHAR2(36);

/* for meta */

type meta_attT is table of matt%rowtype;
type meta_att_idT is table of matt.ATT_ID%type;
type meta_attNamesT is table of matt.ATT_NAME%type;

procedure transformFcDataDir(
  inFc in out fc_T,
  fc_sRec in out fc_R,
  inAcl in out acl_T,
  tmpPermission in out PermissionSimpleO,
  aclMarkers in out triplesT,
  aclMarker in out tripleR,
  aclCounter in out PLS_INTEGER,
  inputidx in pls_integer,
  inFCcounter in pls_integer,
  name in fc.name%type,
  fileId in fc.file_id%type,
  parentId in fc.parent_id%type,
  principalId in fc.principal_id%TYPE,
  timings in timestamp
);


/* symilnk private */
function getSymlinkInt (path in varchar2, looper in out nocopy fileIdT)
	return ResolverO
;

/********* end of rc types *****************/

/* debug */
PROCEDURE DEBUGON
AS
BEGIN
  BEGIN
	DBMS_DEBUG_JDWP.CONNECT_TCP( 'pcegeekn.cern.ch', '5003' );
--  null;
	EXCEPTION
		WHEN OTHERS THEN
		NULL;
	END;
END;

/************************************  *************************/
FUNCTION getNoOfPathDelimiters(inPath IN VARCHAR2)
	RETURN PLS_INTEGER
  deterministic
AS
INT PLS_INTEGER:=0;
startI INTEGER:=1;
offset INTEGER:=1;
BEGIN
	LOOP
	startI:=INSTR(SUBSTR(inPath,offset),'/',1,1);
	offset:=startI+offset;
	IF startI >0 THEN	INT:=INT+1;	ELSE EXIT;END IF;
	END LOOP;
	RETURN INT;
END;


FUNCTION getNoOfPathDelimitersRoot(inPath IN VARCHAR2)
	RETURN PLS_INTEGER
deterministic
AS
INT PLS_INTEGER:=0;
startI INTEGER;
offset INTEGER:=1;
BEGIN
   IF inPath='/' THEN RETURN 0; END IF;
	LOOP
	startI:=INSTR(SUBSTR(inPath,offset),'/',1,1);
	offset:=startI+offset;
	IF startI >0 THEN	INT:=INT+1;	ELSE EXIT;END IF;
	END LOOP;
	RETURN INT;
END;


/************************************ Internal permission helpers ***************************/
PROCEDURE checkPermission(path IN VARCHAR2,
   userPrincipal IN PERM_PRINCIPAL.principal_id%TYPE, principals IN PermStringIntPairOT,
   permCheck OUT NOCOPY FCRStatusT
   )
AS
BEGIN
		SELECT /* +use_nl(f) +cardinality (b 2) */
				--permisison-
				FCRStatusO(
            -1,
            Permission_Util.isOkAcl(userperm,groupperm,otherperm,
						principal_id,userPrincipal,
						group_id,
						principals,
						guid, --input guid
						Permission_Util.writeFileOp),
            file_id,
            parent_id,
            NULL
            )
         BULK COLLECT INTO permCheck
			FROM
			FC_FILE f
			WHERE LEVEL>getNoOfPathDelimiters(path)-1
			START WITH parent_id=0 AND file_id=1 --and name=getpathtoken(1,i.lfn)
			CONNECT BY PRIOR file_id = parent_id --and type='d'
			  AND NAME=getpathtoken(LEVEL-1,path)
		;
END;


PROCEDURE checkPrincipalPermissions(
            userPrincipal PERM_PRINCIPAL.principal_id%TYPE,
            principals PermStringIntPairOT,
            guids IN guidOT,
				DATA IN OUT NOCOPY dataAndStatusOT,
				op IN PLS_INTEGER)
AS

   principalsO PermStringIntPairOT;	-- dn and internal id objects

	dataO dataAndStatusOT;					--	guid and if permitted flag objects

	allBasicOk PLS_INTEGER:=1;
	k PLS_INTEGER:=0;
	one PLS_INTEGER:=0;

BEGIN

/* get basic permission permissions	*/
/* Returns status for all, even nonexisting, and no permitted ones.
  Works but may be expensive for bigger input sets

  1- 0k
  0 - not permitted
  2 - does not exist
*/
  SELECT /*+  cardinality( i 100 ) cardinality (inP 10)  */
    dataAndStatusO(i.guid,
    Permission_Util.isOkAclJoin(userperm,groupperm,otherperm,
      bp.principal_id,userPrincipal,item2,
        principalsO,
      i.guid,op
      ))
   BULK COLLECT INTO DATA
    FROM
    (TABLE(guids) i LEFT OUTER JOIN fc bp ON i.guid=bp.guid
          LEFT OUTER JOIN TABLE(principalsO) inP
           ON bp.group_id=inP.item2)
        ;
END;

PROCEDURE checkPrincipalPermissions(
            securityToken IN principalToken,
				DATA IN OUT NOCOPY dataAndStatusOT,
				op IN PLS_INTEGER)
AS

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

   dataIn dataAndStatusOT:=dataAndStatusOT();

	allBasicOk PLS_INTEGER:=1;
	k PLS_INTEGER:=0;
	one PLS_INTEGER:=0;

BEGIN

   Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);
/* get basic permission permissions	*/
/* Returns status for all, even nonexisting, and no permitted ones.
  Works but may be expensive for bigger input sets

  1- 0k
  0 - not permitted
  2 - does not exist
*/
   dataIn :=DATA;

SELECT /*+  cardinality( i 100 ) cardinality (inP 10)  */
	dataAndStatusO(i.guid,
	Permission_Util.isOkAclJoin(userperm,groupperm,otherperm,
		bp.principal_id,userPrincipal,item2,
      principals,
		i.guid,op
		))
 BULK COLLECT INTO DATA
  FROM
  (TABLE(dataIn) i LEFT OUTER JOIN fc bp ON i.guid=bp.guid
			  LEFT OUTER JOIN TABLE(principals) inP
				 ON bp.group_id=inP.item2)
			;
END;

PROCEDURE checkPrincipalPermissions(
            securityToken IN principalToken,
            guids IN guidOT,
				DATA IN OUT NOCOPY dataAndStatusOT,
				op IN PLS_INTEGER)
AS

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	dataO dataAndStatusOT;					--	guid and if permitted flag objects

	allBasicOk PLS_INTEGER:=1;
	k PLS_INTEGER:=0;
	one PLS_INTEGER:=0;

BEGIN

   Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);
/* get basic permission permissions	*/
/* Returns status for all, even nonexisting, and no permitted ones.
  Works but may be expensive for bigger input sets

  1- 0k
  0 - not permitted
  2 - does not exist
*/
SELECT /*+  cardinality( i 100 ) cardinality (inP 10)  */
	dataAndStatusO(i.guid,
	Permission_Util.isOkAclJoin(userperm,groupperm,otherperm,
		bp.principal_id,userPrincipal,item2,
      principals,
		i.guid,op
		))
 BULK COLLECT INTO DATA
  FROM
  (TABLE(guids) i LEFT OUTER JOIN fc bp ON i.guid=bp.guid
			  LEFT OUTER JOIN TABLE(principals) inP
				 ON bp.group_id=inP.item2)
			;
END;




/* Internal helper to generate guids
	Used for deafultuserperms storage */
FUNCTION generateGuid
	RETURN VARCHAR2
	AS
	lguid VARCHAR2(36);
	BEGIN
		lguid:=sys_guid;
		RETURN SUBSTR(lguid,1,8)||'-'||
		SUBSTR(lguid,9,4)||'-'||
		SUBSTR(lguid,13,4)||'-'||
		SUBSTR(lguid,17,4)||'-'||
		SUBSTR(lguid,21,12);
	END;

PROCEDURE translatePermtoSimple(
    basic IN OUT NOCOPY PermissionO,
    simple IN OUT NOCOPY PermissionSimpleO
  )
AS

BEGIN

  simple.group_id:=Permission_Util.getPrincipalIdIUW(basic.groupname,null,1);
  simple.userperm:=basic.userperm;
  simple.groupperm:=basic.groupperm;
  simple.otherperm:=basic.otherperm;
  Permission_Util.translateAclToSimpleAcl(basic.acl,simple.acl);

END;

--function translatePermtoSimple(
--    basic in out nocopy PermissionO,
--    insertOp in pls_integer
--  )
--  return PermissionSimpleO
--  deterministic
--as
--  simple PermissionSimpleO;
--begin
--  simple:=:=PermissionSimpleO(
--  simple.group_id:=permission_util.getPrincipalId(basic.groupname,'g',insertOp);
--  simple.userperm:=basic.userperm;
--  simple.groupperm:=basic.groupperm;
--  simple.otherperm:=basic.otherperm;
--  simple.acl:=basic.acl;
--
--end;

/*********************************** rc **************************************/
/* default, hidden guid to operate on global default permissions */
/************** internal helpers ***************/
FUNCTION getDefaultPermission(securityToken IN principalToken,
        outEntry OUT NOCOPY PermissionO,
        outerrors OUT NOCOPY returnOT)
RETURN returnO
AS

	ret returnO:=returnO(0);
BEGIN
  DEBUGON();
	ret:=getDefaultPrincipalPermission(securityToken,securityToken.user_name,outEntry,outerrors);
	IF ret.result=0 AND outEntry IS NOT NULL THEN
		RETURN ret;
	ELSE
		ret:=getDefaultGlobalPermission(securityToken,outEntry,outerrors);
		IF ret.result=0 AND outEntry IS NOT NULL THEN RETURN ret; END IF;
	END IF;

	RETURN ret;

END;


/*******************************************************************************
  Superusers set on the table, reinit of the global variable
*******************************************************************************/
/* function called at least at the begining do superuser elaboration on the package */
function SetSuperusers(
        names in lfnT)
  return returnO
as
	/* return */
	ret returnO:=returnO(0);
  inToken principalToken:=principalToken(null,principalNameListT());
  /* fake stuff */
  userId perm_principal.principal_id%type;
	groupIds PermStringIntPairOT:=PermStringIntPairOT();
begin
  --merge on the pp table with 'a' type
    -- if in the list update to a
    -- insert if not in the list (and create home dir)
    -- update with regular user 'u' type on existing ones in pp but not in the list
    -- always preserve internal Admin

   debugOn();
   Permission_Util.refreshSuperusers;
  --inToken.group_names.extend(names.count);
  for i in 1..names.count loop
    inToken.user_name:=names(i);
    Permission_Util.getPrincipalIdsO(inToken,	userId, groupIds,1,true);
  end loop;



  update pp set p_type =
    (nvl2((select 'a' from table(names) i where i.column_value=name) ,'a',decode(p_type,'a','u',p_type)))
        where (p_type='a' or name in (select column_value from table(names))) and name <> 'Admin';

  Permission_Util.refreshSuperusers;
  commit;
  return ret;
end SetSuperusers;



/*******************************************************************************/
/*public void create(FRCEntry[] entries) throws CatalogException, AuthorizationException, ExistsException, NotExistsException, InvalidStatException;*/
FUNCTION FRCCreate(securityToken IN principalToken,
          entries IN FRCEntryT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	permCheck FCCreateT;
	inFc fc_T;
	inAcl acl_T;
   inSurl surl_T;

   rc_sRec 	rc_s;
   fc_sRec fc_R;

   /* const */
   timings TIMESTAMP(0);


   guidMarkerSurlPerms triplesT;

   idxPos INTEGER;

   aclMarkers triplesT;
   aclMarker tripleR;
    k PLS_INTEGER:=0;
   j PLS_INTEGER:=0;
   sCounter PLS_INTEGER:=1;
   aclCounter PLS_INTEGER:=0;

   masterSurlGuard PLS_INTEGER:=0;
   defaultPermFetched BOOLEAN:=FALSE;
   defaultPermission PermissionO;
   tmpPermission PermissionSimpleO;

	/* errors */
	nerrors PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
  errGuid VARCHAR2(36);
  errSurl rcs.surl%TYPE;
  nBadItems PLS_INTEGER:=0;
  debugMsg VARCHAR2(500);

	/* return */
	ret returnO:=returnO(0);


  /*tsting */
  ftable FCTableT;
  atable ACLIdList;
BEGIN

	   --execute immediate 'alter session set events ''10046 trace name context forever, level 12''';
   DEBUGON();
   Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);

	-- remember order

	-- think if entries are too big for input, we could switch to regular lfn checks
	-- starts from the hidden root, so perms are evaluated for '/bla' like paths
  -- TODO: do 'chdir' - \ duplicates on check then, use single vals from
  --       the check while constructing the input

  -- create ass arr for input dirs
  --    create path array, strip it, make unique
  --    get info for this array
  -- deserialize using input arr:
  --    for all in input, find match from the checkarr, initialise permCheck

	SELECT /*+cardinality(ins 10) use_nl(q)*/
	 CASE NVL(d.val.permitted,2)
	   WHEN 1 THEN
	   --ok, can create if file does not exist
	   CASE nvl2(q.parent_id,1,0)
	   	WHEN 0 THEN FCCreateO(
			   r,
	   		 d.val.file_id,
			   fc_file_id_seq.NEXTVAL,
			   			   nvl2(permission,
                PermissionSimpleO(
                  userPrincipal,--principal_id
                  --Permission_Util.getprincipalId(i.permission.groupname,'g',1), -- group_id
                  null, -- do the lookup/creation outside
                  i.permission.userperm,i.permission.groupperm,i.permission.otherperm,
                  -- TODO: create nonexisting names or throw error on nonexistence
                  Permission_Util.translateAclToSimpleAcl(i.permission.acl)), --transalte acls to ids if there are any, don not create users here..
				  d.val.permission)
			   ) --ok, get the data - parent id, new file_id and parental permissions
        ELSE NULL END
	 ELSE NULL END AS val,
     CASE NVL(d.val.permitted,2)  -- fetch permission values checked in th
	   WHEN 1 THEN
	   CASE nvl2(q.parent_id,1,0)   -- check for parents existence
	   	WHEN 1 THEN returnO
               (i.lfn,r,
               Permission_Util.fileOrDirAlreadyExistsVal,
               Permission_Util.fileOrDirAlreadyExistsStr --err, file already there
               )
        ELSE NULL  END
	   WHEN 0 THEN returnO
               (i.lfn,r,
               Permission_Util.accessDeniedErrVal,  --no right to write
               Permission_Util.accessDeniedErrStr
                )
	   WHEN 2 THEN 	returnO(i.lfn,r,
               Permission_Util.InvalidParentVal, --no parent directory exists
               Permission_Util.InvalidParentStr
                )
	   END AS err		--err, parent not there
     BULK COLLECT INTO permCheck,outerrors
	    FROM
      (((SELECT SUBSTR(lfn,1,INSTR(lfn,'/',-1,1)) dir, lfn, permission, ROWNUM AS r FROM TABLE(entries) ins)
      i LEFT OUTER JOIN (
      SELECT  (SELECT
            FCRStatusExtO
            (
              lfn,
              --permisisons check-
              Permission_Util.isOkAcl(
                      userperm,groupperm,otherperm,
                      principal_id,userPrincipal,
                      group_id,
                      principals,
                      guid, --input guid
                      Permission_Util.writeFileOp),
              file_id,
              parent_id,
              PermissionSimpleO
                (userPrincipal,group_id,userperm,groupperm,otherperm,
                   CAST(MULTISET
                        (
                          SELECT p.principal_id,acl.perm
                            FROM acl,PERM_PRINCIPAL p
                              WHERE guid=f.guid AND p.principal_id=acl.principal_id
                        ) AS ACLSimpleIdList
                       )
                )
            )
            FROM
            FC_FILE f
            WHERE LEVEL=Fireman.getNoOfPathDelimiters(lfn)
            START WITH parent_id=0 AND file_id=1 --and name=getpathtoken(1,i.lfn)
            CONNECT BY PRIOR file_id = parent_id AND type='d' -- check if 'd' only valid paths..
              AND NAME=Fireman.getpathtoken(LEVEL-1,lfn))
            AS val
              FROM
              (
                SELECT DISTINCT SUBSTR(lfn,1,INSTR(lfn,'/',-1,1)) lfn
                  FROM TABLE(entries) i
              )
            )
            d ON i.dir=d.val.lfn)
          LEFT OUTER JOIN fc q
            ON parent_id=d.val.file_id AND NAME=SUBSTR(lfn,INSTR(i.lfn,'/',-1,1)+1)
         )
        ;
      -- make it sparse
      j:=permCheck.COUNT;

    -- TODO: MAYBE SHOULD GO INSIDE OF THE NEXT LOOP or change the lookup creation to bulk
    FOR i IN 1..j LOOP
      IF permCheck(i) IS NULL THEN
        permCheck.DELETE(i);
      ELSE
        outerrors.DELETE(i);
        -- the other way could be doing full join on the input users vs. existing ones
        -- followed by the creation for non-exisiting ones
        -- we assume input ordering was preserved here!
        permCheck(i).permission.group_id:=Permission_Util.getprincipalIdIUW(entries(i).permission.groupname,null,1);
      END IF;
    END LOOP;

   timings:=SYSTIMESTAMP();
    FOR i IN 1..permCheck.COUNT LOOP
				--ok to go
        idxPos:=permCheck(i).idx;
				fc_sRec.parent_id:=permCheck(i).parent_id;
				fc_sRec.file_id:=permCheck(i).file_id;

				fc_sRec.NAME:= SUBSTR(entries(idxPos).lfn,INSTR(entries(idxPos).lfn,'/',-1)+1);--get the last part!
				fc_sRec.type:=entries(idxPos).lfnstat.lfntype; --either d or f (for now)
				fc_sRec.guid:=entries(idxPos).guid;

				fc_sRec.cTime:=NVL(entries(idxPos).lfnStat.creationTime,timings);
				fc_sRec.mTime:=NVL(entries(idxPos).lfnStat.mTime,timings);

				fc_sRec.file_size:=entries(idxPos).lfnStat.fileSize;
				fc_sRec.lfn_lifetime:=entries(idxPos).lfnStat.validitytTime;
				fc_sRec.file_checksum:= ' '; --entries(i).guidstat.checkSum;
				fc_sRec.status:=0;            --entries(i).guidstat.status; -- TODO: proper init ? defaults?


        /*
        take parent permission with acls or provided permission if not empty
        if there are no permissions provided (always skip ownership),
          take them from the parent directory
        if provided use them (except userperm)
        */
        -- permisson already fetched with the qry
        tmpPermission:=permCheck(i).permission;

  			-- permissions logic
        /* always take ownership (owner, not group?) from the token */
        fc_sRec.principal_id:=tmpPermission.principal_id;
				fc_sRec.group_id:=tmpPermission.group_id;
				fc_sRec.userperm:=tmpPermission.userperm;
				fc_sRec.groupperm:=tmpPermission.groupperm;
				fc_sRec.otherperm:=tmpPermission.otherperm;

				fc_sRec.dangling_ok:=0;
--				fc_sRec.DATA:=entries(i).lfnstat.DATA; --alter table
        --fetch surl and acl data too

        --surls
        IF entries(idxPos).surlstats IS NOT NULL THEN
        	guidMarkerSurlPerms(entries(idxPos).guid).a:=sCounter;
      		guidMarkerSurlPerms(entries(idxPos).guid).b:=sCounter+entries(idxPos).surlstats.COUNT;
          guidMarkerSurlPerms(entries(idxPos).guid).c:=i;
          FOR j IN 1..entries(idxPos).surlstats.COUNT LOOP --must have at least one element
               rc_sRec.guid:=entries(idxPos).guid;
               rc_sRec.surl:=entries(idxPos).surlstats(j).surl;
               rc_sRec.mTime:=NVL(entries(idxPos).surlstats(j).modifyTime,timings);
--                  0,timings,entries(i).surlstats(j).modifyTime,
--                  entries(i).surlstats(j).modifyTime);
               rc_sRec.cTime:=NVL(entries(idxPos).surlstats(j).creationTime,timings);
               IF entries(idxPos).surlstats(j).master_replica=1 AND masterSurlGuard=0 THEN
                  rc_sRec.master_replica:=1;
                  masterSurlGuard:=1;
               ELSE
                  rc_sRec.master_replica:=0;
               END IF;
               rc_sRec.se:=NVL(entries(idxPos).surlstats(j).se,'dummy'); -- TODO: define proper SE in the 2tier
               rc_sRec.surl_lifetime:=0;	--not used yet
               inSurl(sCounter):=rc_sRec;
               sCounter:=sCounter+1;
          END LOOP; --surls
            IF masterSurlGuard=0 THEN
            -- raise error no master_surl has been specified
               nerrors:=nerrors+1;
               outerrors.EXTEND;
               outerrors(nerrors):=returnO
                        (
                        entries(idxPos).lfn,idxPos,
                        Permission_Util.noMasterReplicaGivenVal,
                        Permission_Util.noMasterReplicaGivenStr
                        );
               inSurl.DELETE(
               guidMarkerSurlPerms(fc_sRec.guid).a,
               guidMarkerSurlPerms(fc_sRec.guid).b
               );
            END IF;
          END IF;--surls exist
          inFc(i):=fc_sRec;  --add lfn to the story
          -----------------------
          -- handle acl's -------
          BEGIN --acls
          IF  tmpPermission.acl IS NOT NULL AND tmpPermission.acl.COUNT>0 THEN
           aclMarker.a:=aclCounter+1;
           aclMarker.b:=aclCounter+tmpPermission.acl.COUNT;
           aclMarker.c:=i;
           aclMarkers(fc_sRec.guid):=aclMarker;

           FOR ac IN 1..tmpPermission.acl.COUNT LOOP
              aclCounter:=aclCounter+1;
              inAcl(aclCounter).guid:=fc_sRec.guid;
              inAcl(aclCounter).principal_id:=tmpPermission.acl(ac).principal_id; --change it to one call per all perms
              inAcl(aclCounter).perm:=tmpPermission.acl(ac).perm;
          END LOOP;
          END IF;
          END; --acls
          masterSurlGuard:=0;

	END LOOP;

   IF inFc.COUNT=0 THEN
      ret.result:=Permission_Util.missingDataVal;
      ret.errString:=Permission_Util.missingDataStr;
      RETURN ret;
   END IF;
	--insert into surl,acl,fc (name inserts already during deserialization)
   ------------ fc insert -----------
   BEGIN
      forall i in 1..inFc.last save exceptions
         insert into fc
            values inFc(i);

   EXCEPTION
      WHEN bulk_errors THEN
         nbulkErr := SQL%BULK_EXCEPTIONS.COUNT;
         outerrors.EXTEND(nbulkErr);
         FOR i IN 1..nbulkErr LOOP
            outerrors(nerrors+i):=
               returnO
                  (
                  entries(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).lfn, --entries always preserve order:1
                  SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
                  Permission_Util.generalInternalBulkErrorVal,
                  SUBSTR('Guid/LFN related data. '||
                  SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
                  );
            --delete from ingested surls, acl to be ingested
            errGuid:=inFc(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid;
            inSurl.DELETE(guidMarkerSurlPerms(errGuid).a,guidMarkerSurlPerms(errGuid).b);
            IF aclMarkers.EXISTS(errGuid) THEN
               inAcl.DELETE(aclMarkers(errGuid).a,aclMarkers(errGuid).b);
            END IF;
         END LOOP;
         nerrors:=nbulkErr+nerrors;
	END; --exception
   ------------ surl insert ---------
   BEGIN
   --TODO: think if surls must be there...
--   if inSurl.count=0 then
--      ret.result:=permission_util.missingDataVal;
--      ret.errString:=permission_util.missingDataStr;
--   end if;

    FORALL i IN 1..inSurl.COUNT save EXCEPTIONS
         INSERT INTO rcs
            VALUES inSurl(i);

   EXCEPTION
      WHEN bulk_errors THEN
         nbulkErr := SQL%BULK_EXCEPTIONS.COUNT;
         outerrors.EXTEND(nbulkErr);
         FOR i IN 1..nbulkErr LOOP
            errSurl:=inSurl(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).surl;
            errGuid:=inSurl(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid;
            outerrors(nerrors+i):=
               returnO
                  (
                  errSurl,
                  guidMarkerSurlPerms(errGuid).c,
                  Permission_Util.generalInternalBulkErrorVal,
                  SUBSTR('Surl related data. '|| SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
                  );
--                  debugMsg:=substr(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
--                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024);
            --delete from input surl,acl
            inSurl.DELETE(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX);
            --in the future we may have optional success... then you'd:
--            guidMarkerSurlPerms(errGuid).c:=guidMarkerSurlPerms(errGuid).c+1;
--            if guidMarkerSurlPerms(errGuid).c >= guidMarkerSurlPerms(errGuid).b-
--                  guidMarkerSurlPerms(errGuid).a
--               then
--               delete from fc where guid = errGuid;
--            end if;
            --but today we just:
            DELETE FROM fc WHERE guid = errGuid;
            --acl is always optional so need for back-delete

            IF inSurl.EXISTS(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX) THEN
            inAcl.DELETE(
               aclMarkers(inSurl(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid).a,
               aclMarkers(inSurl(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid).b
               );
            END IF;
         END LOOP;
         nerrors:=nbulkErr+nerrors;
	END; --exception

   ------------ acl insert ---------
   BEGIN
      forall i in 1..inACl.count save exceptions
         insert into acl
            values inAcl(i);

   EXCEPTION
      WHEN bulk_errors THEN
         nbulkErr := SQL%BULK_EXCEPTIONS.COUNT;
         outerrors.EXTEND(nbulkErr);
         FOR i IN 1..nbulkErr LOOP
            errGuid:=inAcl(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid;
            outerrors(nerrors+i):=
               returnO
                  (
                  entries(aclMarkers(errGuid).c).lfn,
                  aclMarkers(errGuid).c,
                  Permission_Util.generalInternalBulkErrorVal,
                  SUBSTR('ACL related data. '||
                    SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
                  );
            -- in the future we may still keep such entry as valid (and not back-delete)
            -- nowadays:
            DELETE FROM fc WHERE guid=errGuid;
            DELETE FROM rcs WHERE guid=errGuid;
         END LOOP;
         nerrors:=nbulkErr+nerrors;
	END; --exception

   -- do escalation error check to set general resul as error o warnign eventually.
   --execute immediate 'alter session set events ''10046 trace name context off''';
   RETURN ret;
  EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
		RETURN ret;
END;


/*+ public FRCEntry[] readDir(String path, boolean withPermissions) throws CatalogException, AuthorizationException, NotExistsException, CatalogException;*/
FUNCTION readDir(securityToken IN principalToken,
        path IN VARCHAR2,
        pattern IN VARCHAR2,
        LIMIT IN PLS_INTEGER,
        offset IN PLS_INTEGER,
        wPerms IN PLS_INTEGER,
        outEntries OUT NOCOPY FRCEntryT,
        outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	parentCheck fc.file_id%TYPE;
  parentCheckVal PLS_INTEGER;

  /* path ops */
  newPath VARCHAR2(32000);

	/* return */
	ret returnO:=returnO(0);
BEGIN
   DEBUGON();
   outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);
   BEGIN
   --single resovling of symlinks too.

--      SELECT -- TODO: change to parent,file pair for the key
--      file_id,
--      Permission_Util.isOkAcl(userperm,groupperm,otherperm,
--                  principal_id,userPrincipal,
--                  group_id,
--                  principals,
--                  guid, --input guid
--                  Permission_Util.readDirOp)
--      COLLECT INTO parentCheck, parentCheckVal
--      FROM
--      FC_FILE f
--      WHERE LEVEL=getNoOfPathDelimitersRoot(path)+1
--      START WITH parent_id=0 AND file_id=1
--      CONNECT BY PRIOR file_id = parent_id
--      AND NAME =getpathtoken(LEVEL-1,path);
--

	 SELECT
    TREAT (res AS resolverO).file_id,
    TREAT (res AS resolverO).permission
   COLLECT INTO parentCheck, parentCheckVal
	  FROM(
  SELECT
	permission_util.isOkAclX(--check perms and return error or data as usual but using non-bulk check
	  NVL(CASE
            WHEN TYPE IN ( 'f', 'd')  THEN
              resolverO(parent_id,file_id,NAME,guid,principal_id,group_id,userperm,groupperm,otherperm) --also do permissions
			  	   WHEN TYPE ='s' THEN
              fireman.getSymlink(DATA) --also do permissions
					ELSE
					NULL
				END
			 ,resolverO(0,2))
			 ,userPrincipal,principals,Permission_Util.readDirOp) res
       FROM
      FC_FILE f
     WHERE LEVEL=fireman.getNoOfPathDelimitersRoot(path)+1
      START WITH parent_id=0 AND file_id=1
      CONNECT BY PRIOR file_id = parent_id -- check if 'd' only valid paths..
		AND NAME=fireman.getpathtoken(LEVEL-1,path)
		);

   EXCEPTION
      WHEN NO_DATA_FOUND THEN
         ret.result:=Permission_Util.NoSuchFileVal;
         ret.errString:=Permission_Util.NoSuchFileStr;
      RETURN ret;
   END;
   IF parentCheckVal=0 THEN
     ret.result:=Permission_Util.accessDeniedErrVal;
     ret.errString:=Permission_Util.accessDeniedErrStr;
     RETURN ret;
    elsif parentCheck=permission_util.loopInSymlinksDetectedVal then
      ret.result:=Permission_Util.loopInSymlinksDetectedVal;
      ret.errString:=Permission_Util.loopInSymlinksDetectedStr;
     RETURN ret;     
   END IF;


  IF path='/' THEN newPath:=NULL; ELSE newPath:=path; END IF;

  IF pattern IS NULL OR pattern='%' THEN
   --parent ok
  SELECT FRCEntryO
  (
      guid,
      guidstatO(file_size,ctime,file_checksum,status,mtime),
      NULL, --surl
      CASE wPerms
          WHEN 1 THEN permissionO
            (
             (SELECT p.NAME FROM PERM_PRINCIPAL p WHERE p.principal_id=b.principal_id),
             (SELECT p.NAME FROM PERM_PRINCIPAL p WHERE p.principal_id=b.group_id),
             userperm,groupperm,otherperm,
             CAST(MULTISET
              (
              SELECT
                p.NAME,acl.perm
                FROM acl,PERM_PRINCIPAL p
                WHERE guid=b.guid AND p.principal_id=acl.principal_id
              ) AS ACLList)
            )
          ELSE NULL END,
      newPath||'/'||NAME,
      lfnstatO(file_size,ctime,mTime,type,data,lfn_lifetime) --alter table
        )
  BULK COLLECT INTO outEntries
  FROM (
    SELECT principal_id,group_id,userperm,groupperm,otherperm,guid,
      file_size,ctime,file_checksum,status,mtime,
      NAME,
      type,data,lfn_lifetime, ROWNUM rnum
     FROM (
      SELECT
      principal_id,group_id,userperm,groupperm,otherperm,
      guid,
      file_size,ctime,file_checksum,status,mtime,
      NAME,
      type,data,lfn_lifetime
       FROM fc f WHERE parent_id = parentCheck
      ) a
    WHERE ROWNUM <= offset+LIMIT )  b
   WHERE rnum > offset
  AND Permission_Util.isOkAcl(userperm,groupperm,otherperm,
              principal_id,userPrincipal,
              group_id,
              principals,
              guid, --input guid
              Permission_Util.listFileOp)=1
   ;
   ELSE

   SELECT FRCEntryO
  (
      guid,
      guidstatO(file_size,ctime,file_checksum,status,mtime),
      NULL, --surl
      CASE wPerms
          WHEN 1 THEN permissionO
            (
             (SELECT p.NAME FROM PERM_PRINCIPAL p WHERE p.principal_id=b.principal_id),
             (SELECT p.NAME FROM PERM_PRINCIPAL p WHERE p.principal_id=b.group_id),
             userperm,groupperm,otherperm,
             CAST(MULTISET
              (
              SELECT
                p.NAME,acl.perm
                FROM acl,PERM_PRINCIPAL p
                WHERE guid=b.guid AND p.principal_id=acl.principal_id
              ) AS ACLList)
            )
          ELSE NULL END,
      newPath||'/'||NAME,
      lfnstatO(file_size,ctime,mTime,type,data,lfn_lifetime)
        )
  BULK COLLECT INTO outEntries
  FROM (
    SELECT principal_id,group_id,userperm,groupperm,otherperm,guid,
      file_size,ctime,file_checksum,status,mtime,
      NAME,
      type,data,lfn_lifetime, ROWNUM rnum
     FROM (
      SELECT
      principal_id,group_id,userperm,groupperm,otherperm,
      guid,
      file_size,ctime,file_checksum,status,mtime,
      NAME,
      type,data,lfn_lifetime
       FROM fc f
       WHERE
       parent_id = parentCheck
       AND NAME LIKE pattern ESCAPE '\'
      ) a
    WHERE ROWNUM <= offset+LIMIT ) b
   WHERE rnum > offset
  AND Permission_Util.isOkAcl(userperm,groupperm,otherperm,
              principal_id,userPrincipal,
              group_id,
              principals,
              guid, --input guid
              Permission_Util.listFileOp)=1
      ;

   END IF;
	RETURN ret;
END;



/* not that err outentries are nullified on the erroneous input index */
/*+ public FRCEntry[] listReplicas(String[] lfns, boolean withPermissions) throws CatalogException, AuthorizationException, NotExistsException;*/
FUNCTION listReplicas(securityToken IN principalToken,
			 lfns IN lfnT,
			 wPerms IN PLS_INTEGER,
			 outEntries OUT NOCOPY FRCEntryT,
       outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	parentCheck lfnCheckT;

	/* errors */
	j PLS_INTEGER:=0;


	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
   DEBUGON();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);



SELECT
 CASE
 	WHEN treat(res as resolverO).permission=1 AND treat(res as resolverO).ftype='f'
  THEN
  FRCEntryO(
	treat(res as resolverO).guid,guidstatO(treat(res as resolverO).file_size,
	treat(res as resolverO).ctime,treat(res as resolverO).file_checksum,
	treat(res as resolverO).status,treat(res as resolverO).mtime),
	CAST(MULTISET
			(
			SELECT s.surl,s.mTime, s.cTime,s.se,s.master_replica
				FROM RC_SURL s
				WHERE s.guid=treat(res as resolverO).guid
			) AS surlEntryT),--surlentry
   CASE wPerms
     WHEN 1 THEN permissionO(
	    (SELECT p.NAME FROM PERM_PRINCIPAL p WHERE	principal_id=treat(res as resolverO).principal_id),
    	(SELECT p.NAME FROM PERM_PRINCIPAL p WHERE	principal_id=treat(res as resolverO).group_id),
	     treat(res as resolverO).userperm,treat(res as resolverO).groupperm,treat(res as resolverO).otherperm,
      --permission_util.getacl(guid)
      CAST(MULTISET
         	(
         	SELECT
            p.NAME,acl.perm
            FROM acl,PERM_PRINCIPAL p
            WHERE
            guid=treat(res as resolverO).guid AND p.principal_id=acl.principal_id
			) AS ACLList)
      )
     ELSE permissionO(
    (SELECT p.NAME FROM PERM_PRINCIPAL p WHERE	principal_id=treat(res as resolverO).principal_id),
    (SELECT p.NAME FROM PERM_PRINCIPAL p WHERE	principal_id=treat(res as resolverO).group_id),
      treat(res as resolverO).userperm,
      treat(res as resolverO).groupperm,
      treat(res as resolverO).otherperm,NULL)
   END,
	treat(res as resolverO).fname,     --lfn
	lfnstatO(
		treat(res as resolverO).file_size,treat(res as resolverO).ctime,
		treat(res as resolverO).mTime,treat(res as resolverO).ftype,
		treat(res as resolverO).DATA,
		treat(res as resolverO).lfn_lifetime) --alter table
    )
    ELSE NULL END val,
	CASE treat(res as resolverO).permission
      when 1 then
        case treat(res as resolverO).ftype
          when 'd' then
          returnO
              (treat(res as resolverO).fname,treat(res as resolverO).idx,
               Permission_Util.invalidArgumentDirVal,  --no right to ask about dir
               Permission_Util.invalidArgumentDirStr
              )
        else
          null
        end
  	  WHEN 0 THEN returnO
              (treat(res as resolverO).fname,treat(res as resolverO).idx,
               Permission_Util.accessDeniedErrVal,  --no right to write
               Permission_Util.accessDeniedErrStr
              )
        WHEN 2 THEN returnO
              (treat(res as resolverO).fname,treat(res as resolverO).idx,
               Permission_Util.NoSuchFileVal, --no parent directory exists
               Permission_Util.NoSuchFileStr
              )
      WHEN permission_util.loopInSymlinksDetectedVal THEN   		
   		returnO(TREAT(res AS resolverO).fname,
        TREAT(res AS resolverO).idx,
        Permission_Util.loopInSymlinksDetectedVal, --no parent directory exists
        Permission_Util.loopInSymlinksDetectedStr
      )              
   		ELSE
        returnO
              (treat(res as resolverO).fname,treat(res as resolverO).idx,
               Permission_Util.NoSuchFileVal, --no parent directory exists
               Permission_Util.NoSuchFileStr
              )
   		END ret
  bulk collect into outEntries,outErrors
 FROM (
 SELECT
	permission_util.isOkAclX(--check perms and return error or data as usual but using non-bulk check
	  NVL((SELECT
	   		  CASE WHEN TYPE IN ( 'f', 'd')  THEN
         resolverO(idx,parent_id,file_id,
            lfn,guid,type,ctime,mtime,
            file_size,file_checksum,LFN_LIFETIME,status,
            principal_id,group_id,userperm,groupperm,otherperm,
            dangling_ok,
            data,-1)
			  	   WHEN TYPE ='s' THEN
				   	resolverO(idx,fireman.getSymlink(DATA)) --also do permissions
					ELSE
					NULL
				   END
  	   		    FROM
		            FC_FILE f
                WHERE LEVEL=case namelfn when ':' then 1 else 2 end -- always leaf or root if /
		            START WITH parent_id=z.val.parent_id AND file_id=z.val.file_id
		            CONNECT BY PRIOR file_id = parent_id -- check if 'd' only valid paths..
		              AND NAME=namelfn
			 ),resolverO(idx,2,lfn))
			 ,userPrincipal,principals,Permission_Util.listFileOp) RES
  FROM
  (SELECT
		SUBSTR(lfn,1,INSTR(lfn,'/',-1,1)) dir,
		nvl(SUBSTR(lfn,INSTR(lfn,'/',-1,1)+1),':') nameLfn,
		lfn,
		idx
    FROM
    (
      SELECT /*+ cardinality (iz, 50) */
        COLUMN_VALUE AS lfn,ROWNUM AS idx
        FROM
        TABLE(lfns) iz
    )
	) i
  LEFT JOIN
	(
	SELECT (
    SELECT updateFCSimpleO(parent_id,file_id,NULL,dir)
    FROM
      FC_FILE f
      WHERE LEVEL=Fireman.getNoOfPathDelimiters(dir)
      START WITH parent_id=0 AND file_id=1
      CONNECT BY PRIOR file_id = parent_id AND TYPE='d' -- check if 'd' only valid paths..
        AND NAME=Fireman.getpathtoken(LEVEL-1,dir)
        ) val
    FROM
    (
    SELECT /*+ cardinality (dirz, 50) */
      DISTINCT SUBSTR(column_value,1,INSTR(column_value,'/',-1,1)) dir
    FROM
    table(lfns)dirz
    )
	) z
	 ON (i.dir=z.val.item)
	 )
	 ;
   -- remove nulls from the output, make it sparse
   j:=outEntries.count;
    FOR i IN 1..j LOOP
      IF outEntries(i) IS NULL THEN
        outEntries.DELETE(i);
      ELSE
        outerrors.DELETE(i);
      END IF;
    END LOOP;




	RETURN ret;
END;

/*+ public String[] listLfn(String[] surls) throws CatalogException, AuthorizationException, NotExistsException;*/
FUNCTION listLfn(securityToken IN principalToken,
          surls IN surlT,
          outEntries OUT NOCOPY StringPairT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	tmpStat StringPairT;
	permSimple booleanT;

   surlMarker bigPairsT;

	/* errors */
	nerrors PLS_INTEGER:=0;


	/* return */
	ret returnO:=returnO(0);
BEGIN

   DEBUGON();

   outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);

   /* stupid but true: remember ordering for err hdling */
   FOR i IN 1..surls.LAST LOOP
      surlMarker(surls(i)).a:=i;
   END LOOP;


 SELECT /*+ cardinality( i 5 ) cardinality(inP 10) use_nl(inP) */
 	StringPairO(
   i.COLUMN_VALUE,
   CASE
   (Permission_Util.isOkAclJoin(bp.userperm,bp.groupperm,bp.otherperm,
      	bp.principal_id,userPrincipal,
			item2,
	      principals,
			s.guid, Permission_Util.listFileOp))
      WHEN 1 THEN  (
      SELECT
         reversePath(SYS_CONNECT_BY_PATH(NAME,'/'))
         FROM FC_FILE f
         WHERE NAME=':'
         START WITH f.guid = s.guid
         CONNECT BY PRIOR parent_id = file_id)
	  ELSE NULL END)
	BULK COLLECT INTO outEntries
	FROM
	 (((TABLE(surls) i LEFT OUTER JOIN rcs s ON s.surl=i.COLUMN_VALUE)
		LEFT OUTER JOIN fc bp  ON s.guid=bp.guid)
				  LEFT OUTER JOIN TABLE(principals) inP
				  ON bp.group_id=inP.item2)
   ;
   -- fill up errrors
   FOR i IN 1..outEntries.LAST LOOP
      IF outEntries(i).string2 IS NULL THEN
         nerrors:=nerrors+1;
         outErrors.EXTEND();
         outErrors(nerrors):=
            returnO(outEntries(i).string1,
            surlMarker(outEntries(i).string1).a,
            Permission_Util.accessDeniedErrVal,Permission_Util.accessDeniedErrStr);
      END IF;
   END LOOP;

	RETURN ret;
END;


/* nonsense call as no key nor index on the return */
/*+ public Stat[] getStat(String[] lfns) throws CatalogException, AuthorizationException, NotExistsException;*/
FUNCTION getStat(securityToken IN principalToken,
          lfns IN lfnT,
          outEntries OUT NOCOPY statT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	parentCheck lfnCheckT;

	/* errors */
	j PLS_INTEGER:=0;

	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);

   /* gets stat based on lfns, based on their type:
         for dirs return lfnstat
         for files surstat:1:2
   */
 SELECT
 CASE treat(res as resolverO).permission
 	WHEN 1 THEN
	lfnstatO(
		treat(res as resolverO).file_size,treat(res as resolverO).ctime,
		treat(res as resolverO).mTime,treat(res as resolverO).ftype,
		treat(res as resolverO).DATA,
		treat(res as resolverO).lfn_lifetime) --alter table
  ELSE NULL END val,
	CASE treat(res as resolverO).permission
      when 1 then null
  	  WHEN 0 THEN returnO
              (treat(res as resolverO).fname,treat(res as resolverO).idx,
               Permission_Util.accessDeniedErrVal,  --no right to write
               Permission_Util.accessDeniedErrStr
              )
      WHEN 2 THEN returnO
              (treat(res as resolverO).fname,treat(res as resolverO).idx,
               Permission_Util.NoSuchFileVal, --no parent directory exists
               Permission_Util.NoSuchFileStr
              )
      WHEN permission_util.loopInSymlinksDetectedVal THEN   		
   		returnO(TREAT(res AS resolverO).fname,
        TREAT(res AS resolverO).idx,
        Permission_Util.loopInSymlinksDetectedVal, --no parent directory exists
        Permission_Util.loopInSymlinksDetectedStr
      )              
   		else
   		returnO
              (treat(res as resolverO).fname,treat(res as resolverO).idx,
               Permission_Util.NoSuchFileVal, --no parent directory exists
               Permission_Util.NoSuchFileStr
              )
   		END ret
  bulk collect into outEntries,outErrors
 FROM (
 SELECT
	permission_util.isOkAclX(--check perms and return error or data as usual but using non-bulk check
	  NVL((SELECT
	   		  CASE WHEN TYPE IN ( 'f', 'd')  THEN
         resolverO(idx,parent_id,file_id,
            lfn,guid,type,ctime,mtime,
            file_size,file_checksum,LFN_LIFETIME,status,
            principal_id,group_id,userperm,groupperm,otherperm,
            dangling_ok,
            data,-1)
			  	   WHEN TYPE ='s' THEN
				   	resolverO(idx,fireman.getSymlink(DATA)) --also do permissions
					ELSE
					NULL
				   END
  	   		    FROM
		            FC_FILE f
                WHERE LEVEL=case namelfn when ':' then 1 else 2 end -- always leaf or root if /
		            START WITH parent_id=z.val.parent_id AND file_id=z.val.file_id
		            CONNECT BY PRIOR file_id = parent_id -- check if 'd' only valid paths..
		              AND NAME=namelfn
			 ),resolverO(idx,2,lfn))
			 ,userPrincipal,principals,Permission_Util.listDirOp) RES
  FROM
  (SELECT
		SUBSTR(lfn,1,INSTR(lfn,'/',-1,1)) dir,
		nvl(SUBSTR(lfn,INSTR(lfn,'/',-1,1)+1),':') nameLfn,
		lfn,
		idx
    FROM
    (
      SELECT /*+ cardinality (iz, 50) */
        COLUMN_VALUE AS lfn,ROWNUM AS idx
        FROM
        TABLE(lfns) iz
    )
	) i
  LEFT JOIN
	(
	SELECT (
    SELECT updateFCSimpleO(parent_id,file_id,NULL,dir)
    FROM
      FC_FILE f
      WHERE LEVEL=Fireman.getNoOfPathDelimiters(dir)
      START WITH parent_id=0 AND file_id=1
      CONNECT BY PRIOR file_id = parent_id AND TYPE='d' -- check if 'd' only valid paths..
        AND NAME=Fireman.getpathtoken(LEVEL-1,dir)
        ) val
    FROM
    (
    SELECT /*+ cardinality (dirz, 50) */
      DISTINCT SUBSTR(column_value,1,INSTR(column_value,'/',-1,1)) dir
    FROM
    table(lfns)dirz
    )
	) z
	 ON (i.dir=z.val.item)
	 )
	 ;
   -- remove nulls from the output, make it sparse
   j:=outEntries.count;
    FOR i IN 1..j LOOP
      IF outEntries(i) IS NULL THEN
        outEntries.DELETE(i);
      ELSE
        outerrors.DELETE(i);
      END IF;
    END LOOP;

	RETURN ret;
END;


/* does not make any sense now since no key (lfn/guid) is passed in*/
/*+ public void updateStat(Stat[] stats) throws CatalogException, AuthorizationException, InvalidStatException, NotExistsException;*/
FUNCTION updateStat(securityToken IN principalToken,
          stats IN statT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	parentCheck lfnCheckT;


	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);

   /* 2 types possbile - lfnstat or guidstat, update accordyingly */

	RETURN ret;
END;

/*+ public PermissionEntry[] getPermission(String[] lfns) throws CatalogException, AuthorizationException, NotExistsException;*/
-- TODO: heuristics to validate what to qry first or to have union...

FUNCTION getPermission(securityToken IN principalToken,
          items IN StringBooleanT,
          outEntries out nocopy PermissionEntryT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* errors */

  j pls_integer;
	/* return */
	ret returnO:=returnO(0);

BEGIN

  DEBUGON();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);

SELECT
 CASE TREAT(res AS resolverO).permission
 	WHEN 1 THEN
   	PermissionEntryO(
      TREAT(res AS resolverO).fname,
      permissionO
      (
	     (SELECT p.NAME FROM PERM_PRINCIPAL p WHERE	principal_id=TREAT(res AS resolverO).principal_id), --35sec
       (SELECT p.NAME FROM PERM_PRINCIPAL p WHERE	principal_id=TREAT(res AS resolverO).group_id),
         TREAT(res AS resolverO).userperm,TREAT(res AS resolverO).groupperm,TREAT(res AS resolverO).otherperm,
         CAST(MULTISET
         	(
         	SELECT
            p.NAME,acl.perm
            FROM acl,PERM_PRINCIPAL p
            WHERE guid=TREAT(res AS resolverO).guid AND p.principal_id=acl.principal_id
          ) AS ACLList)
      ))
  ELSE
      NULL
  END val,
	CASE TREAT(res AS resolverO).permission
      when 1 then null
  	  WHEN 0 THEN
	  returnO(
        TREAT(res AS resolverO).guid,
        TREAT(res AS resolverO).idx,
        Permission_Util.accessDeniedErrVal,  --no right to write
        Permission_Util.accessDeniedErrStr
      )
        WHEN 2 THEN
		returnO(TREAT(res AS resolverO).fname,
        TREAT(res AS resolverO).idx,
        Permission_Util.NoSuchFileVal, --no parent directory exists
        Permission_Util.NoSuchFileStr
      )
      WHEN permission_util.loopInSymlinksDetectedVal THEN   		
   		returnO(TREAT(res AS resolverO).fname,
        TREAT(res AS resolverO).idx,
        Permission_Util.loopInSymlinksDetectedVal, --no parent directory exists
        Permission_Util.loopInSymlinksDetectedStr
      )
   		ELSE
   		returnO(TREAT(res AS resolverO).fname,
        TREAT(res AS resolverO).idx,
        Permission_Util.NoSuchFileVal, --no extra-parent directory exists
        Permission_Util.NoSuchFileStr
      )
   		END rets
  BULK COLLECT INTO outEntries, outErrors
  FROM (
  SELECT --guid retrieval/check
	permission_util.isOkAclX(--check perms and return error or data as usual but using non-bulk check
	  		  nvl(CASE WHEN TYPE IN ('f','d')  THEN
            resolverO(idx,parent_id,file_id,
            inguid,guid,type,ctime,mtime,
            file_size,file_checksum,LFN_LIFETIME,status,
            principal_id,group_id,userperm,groupperm,otherperm,
            dangling_ok,
            data,-1)
			  	   WHEN TYPE ='s' THEN
				   	resolverO(idx,fireman.getSymlink(DATA)) --also do permissions
        	ELSE
            NULL
          END,
          resolverO(idx,2,inguid)) -- not found permission err propagation
			 ,userPrincipal,principals,Permission_Util.listFileOp) RES
        FROM
          (
            select indx idx ,STRING inguid,bp.* from
            (SELECT /*+ cardinality( i 10) */ i.*, rownum as indx
              FROM TABLE(items)  i)
            LEFT JOIN fc bp ON string=bp.guid WHERE boolval=1
          ) z
 	 UNION ALL
  SELECT --lfn side
	permission_util.isOkAclX(--check perms and return error or data as usual but using non-bulk check
	  NVL((SELECT
	   		  CASE WHEN TYPE IN ('f','d')  THEN
          resolverO(idx,parent_id,file_id,
            lfn,guid,type,ctime,mtime,
            file_size,file_checksum,LFN_LIFETIME,status,
            principal_id,group_id,userperm,groupperm,otherperm,
            dangling_ok,
            data,-1)
			  	   WHEN TYPE ='s' THEN
				   	resolverO(idx,fireman.getSymlink(DATA)) --also do permissions
					ELSE
					NULL
				   END
  	   		    FROM
		            FC_FILE f
                WHERE LEVEL=case namelfn when ':' then 1 else 2 end -- always leaf or root if /
		            START WITH parent_id=z.val.parent_id AND file_id=z.val.file_id
		            CONNECT BY PRIOR file_id = parent_id -- check if 'd' only valid paths..
		              AND NAME=namelfn
			 ),resolverO(idx,2,lfn))
			 ,userPrincipal,principals,Permission_Util.listFileOp) RES
  FROM
  (SELECT
		SUBSTR(lfn,1,INSTR(lfn,'/',-1,1)) dir,
		nvl(SUBSTR(lfn,INSTR(lfn,'/',-1,1)+1),':') nameLfn,
		lfn,
		idx
    FROM
    (
      select /*+ cardinality (iz, 50) */
        string as lfn, ROWNUM AS idx,boolval
          from TABLE(items)
    ) WHERE boolval=0
	) i
  LEFT JOIN
	(
	SELECT (
    SELECT updateFCSimpleO(parent_id,file_id,NULL,dir)
    FROM
      FC_FILE f
      WHERE LEVEL=Fireman.getNoOfPathDelimiters(dir)
      START WITH parent_id=0 AND file_id=1
      CONNECT BY PRIOR file_id = parent_id AND TYPE='d' -- check if 'd' only valid paths..
        AND NAME=Fireman.getpathtoken(LEVEL-1,dir)
        ) val
    FROM
    (
    SELECT /*+ cardinality (dirz, 50) */
      DISTINCT SUBSTR(dir,1,INSTR(dir,'/',-1,1)) dir
    FROM
    (select i.string as dir from TABLE(items) i WHERE i.boolval=0) dirz
    )
	) z
	 ON (i.dir=z.val.item)
	 )
  ;
  j:=outEntries.count;
  FOR i IN 1..j LOOP
    IF outEntries(i) IS NULL THEN
      outEntries.DELETE(i);
    else
      outErrors.DELETE(i);
    END IF;
  END LOOP;

	RETURN ret;
END;

/*+ public void setPermission(PermissionEntry[] permissions) throws CatalogException, AuthorizationException, NotExistsException, InvalidStatException;*/
FUNCTION setPermission(securityToken IN principalToken,
          permissions  IN PermissionEntryExtT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
   itemCheck FCRStatusIdxT;
   cnt pls_integer;

   pids bulkPrincipalidT;
   gids bulkGroupIdT;
   ups bulkUserPT;
   gps bulkGroupPT;
   ops bulkOtherPT;
   guids bulkGuidT;


   j PLS_INTEGER:=0;
   i pls_integer;
   run pls_integer:=0;


   bulkAcl ACLIdList:=ACLIdList();
   guidMarker pairR;
   guidMarkerIn bigPairsT;
   guidMarkerAcl pairsT;


  mask pls_integer;

  	/* errors */
	nerrors PLS_INTEGER:=0;
  nerrorsBp PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
   errGuid VARCHAR2(36);
   errSurl rcs.surl%TYPE;
   nBadItems PLS_INTEGER:=0;

	/* return */
	ret returnO:=returnO(0);
BEGIN
--   outerrors:=returnOT();
   Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);
   DEBUGON();
   /* consider updating using full bulk wo forall, just check returning for errors */
   /* what could be possible error there:1 - locks:2*/

   -- check if perms ok, report errors


   --get data for sel update, get data for

--   SELECT val
--   BULK COLLECT INTO itemCheck
--   FROM
--   (
--   SELECT
--   (
--   SELECT
--   updateFCSimpleO(parent_id,file_id,guid,i.COLUMN_VALUE)
-- 	FROM
--	FC_FILE f
--	WHERE LEVEL=getNoOfPathDelimitersRoot(i.COLUMN_VALUE)+1
--	AND Permission_Util.isOkAcl(userperm,groupperm,otherperm,
--					principal_id,userPrincipal,
--					group_id,
--			    principals,
--					guid, --input guid
--					Permission_Util.permissionOp)=1
--	START WITH file_id=1
--	CONNECT BY PRIOR file_id = parent_id
--	AND NAME =getpathtoken(LEVEL-1,i.COLUMN_VALUE)
--	) val
--   FROM TABLE(pipePermissionEntryExtV(permissions,0))  i
--   UNION ALL
--   SELECT /*+ use_nl(inP) cardinality( i 100 ) cardinality (inP 10)  */
--			DECODE(
--        Permission_Util.isOkAclJoin(bp.userperm,bp.groupperm,bp.otherperm,
--        bp.principal_id,userPrincipal,
--        bp.group_id,
--        principals,
--        i.item,Permission_Util.permissionOp),
--        1, updateFCSimpleO(parent_id,file_id,guid,guid),
--        NULL) val
--  FROM
-- (( TABLE(pipePermissionEntryExt(permissions,1)) i
--  LEFT OUTER JOIN fc bp ON i.item=bp.guid )
--			  LEFT OUTER JOIN TABLE(principals) inP
--				 ON bp.group_id=inP.item2)
--  )
--   ;
--  -- TODO: remove and change to rownum in the input
--  FOR i IN 1..permissions.LAST LOOP
--    guidMarkerIn(permissions(i).item).a:=i;
--  END LOOP;
--

  mask:=permission_util.getOpMask(Permission_Util.permissionOp);

  SELECT val, err
  BULK COLLECT INTO itemCheck,outErrors
  FROM
  (
   --check guid permissions
    SELECT /*+  cardinality( i 100 ) cardinality (inP 10) use_nl(inP) */
        case permitted
          when 1 then FCRStatusIdxO
        (guid,
        idx,
        permitted,
        file_id,
        parent_id,
        null) else null end val,
        case permitted
          when 1 then NULL
          when 0 then returnO(string,idx,Permission_Util.accessDeniedErrVal,Permission_Util.accessDeniedErrStr,0)
          when 2 then returnO(string,idx,Permission_Util.NoSuchFileVal,Permission_Util.NoSuchFileStr,0)          
          WHEN permission_util.loopInSymlinksDetectedVal THEN
                    returnO(string,idx,Permission_Util.loopInSymlinksDetectedVal,Permission_Util.loopInSymlinksDetectedStr)
        else null end as err
        from
        (
          select
          case
          when ((sign(bitand(userperm,mask))=1 and principal_id=userPrincipal)
            or (sign(bitand(groupperm,mask))=1 and group_id is not null)
            or sign(bitand(otherperm,mask))=1)
            then 1
          when userperm is null then 2
          when userperm is not null then-- check acl
            (SELECT COUNT(*)
          --    into permitted
              FROM dual WHERE
              EXISTS
              (
              SELECT /*+ cardinality (inP2 10) */  NULL
                FROM
                  acl a,
                  TABLE(principals) inP2
                  WHERE
                  (a.guid=guid
                  AND (inP2.item2=a.principal_id OR a.principal_id=userprincipal)
                  AND SIGN(BITAND(perm,mask))=1))
              or exists
                (
                select /*+ cardinality (inP2 5) cardinality (su 5) */ null
                from
                TABLE(permission_util.superUsers) su,
                TABLE(principals) inP2
                    where su.column_value=inP2.item2 or su.column_value=userPrincipal
                )
          )
		else 2
		end as permitted
    ,idx,string,parent_id,file_id,guid
          FROM
          (
          select /*+cardinality (i 20) */ itemtype boolval,item string,rownum as idx
          from
          TABLE(permissions)
          ) i
          LEFT OUTER JOIN fc bp ON (i.string=bp.guid) WHERE boolval=1
      )
      UNION ALL
      --lfn part
      SELECT
         (TREAT(res AS resolverO)).getFCRStatusIdxO() val,
         (TREAT(res AS resolverO)).getReturn() err
         FROM (
         SELECT
          permission_util.isOkAclXMask(--check perms and return error or data as usual but using non-bulk check
            NVL((SELECT
                  CASE WHEN TYPE IN ( 'f', 'd')  THEN
                 resolverO(idx,parent_id,file_id,
                    lfn,guid,type,ctime,mtime,
                    file_size,file_checksum,LFN_LIFETIME,status,
                    principal_id,group_id,userperm,groupperm,otherperm,
                    dangling_ok,
                    data,-1)
                     WHEN TYPE ='s' THEN
                    resolverO(idx,fireman.getSymlink(DATA)) --also do permissions
                  ELSE
                  NULL
                   END
                      FROM
                        FC_FILE f
                        WHERE LEVEL=case namelfn when ':' then 1 else 2 end -- always leaf or root if /
                        START WITH parent_id=z.val.parent_id AND file_id=z.val.file_id
                        CONNECT BY PRIOR file_id = parent_id -- check if 'd' only valid paths..
                          AND NAME=namelfn
               ),resolverO(idx,2,lfn))
               ,userPrincipal,principals,Permission_Util.permissionOp) RES
          FROM
          (SELECT
            SUBSTR(lfn,1,INSTR(lfn,'/',-1,1)) dir,
            nvl(SUBSTR(lfn,INSTR(lfn,'/',-1,1)+1),':') nameLfn,
            lfn,
            idx
            FROM
            (
              SELECT /*+ cardinality (iz, 50) */
                item AS lfn,itemtype, ROWNUM AS idx
                FROM
                TABLE(permissions) iz
            ) where itemtype=0
          ) i
          LEFT JOIN
          (
          SELECT (
            SELECT updateFCSimpleO(parent_id,file_id,NULL,dir)
            FROM
              FC_FILE f
              WHERE LEVEL=Fireman.getNoOfPathDelimiters(dir)
              START WITH parent_id=0 AND file_id=1
              CONNECT BY PRIOR file_id = parent_id AND TYPE='d' -- check if 'd' only valid paths..
                AND NAME=Fireman.getpathtoken(LEVEL-1,dir)
                ) val
            FROM
            (
            SELECT /*+ cardinality (dirz, 50) */
              DISTINCT SUBSTR(item,1,INSTR(item,'/',-1,1)) dir,itemtype
            FROM
            table(permissions) dirz
            ) where itemtype=0
          ) z
           ON (i.dir=z.val.item)
        )
      )

        ;


   cnt:=itemcheck.count;
    FOR idx IN 1..cnt LOOP
      IF itemCheck(idx) IS NULL THEN
        nerrors:=nerrors+1;
      ELSE
         outerrors.DELETE(idx);
         i:=itemCheck(idx).idx;
         run:=run+1;
         pids(run):=case permissions(i).perm.userName is null when true then
          userPrincipal
          else Permission_Util.getPrincipalIdIUW(permissions(i).perm.userName,null,1) end;
         gids(run):=case permissions(i).perm.groupName is null when true then
          principals(1).item2
          else Permission_Util.getPrincipalIdIUW(permissions(i).perm.groupName,null,1) end;
         ups(run):=permissions(i).perm.userperm;
         gps(run):=permissions(i).perm.groupperm;
         ops(run):=permissions(i).perm.otherperm;
         guids(run):=itemCheck(idx).guid;
         -- acls
         IF permissions(i).perm.acl IS NOT NULL THEN
          guidMarker.a:=j+1;
          guidMarker.b:=j+permissions(i).perm.acl.COUNT;
          guidMarkerAcl(itemCheck(idx).guid):=guidMarker;
          bulkAcl.EXTEND(permissions(i).perm.acl.COUNT);
           FOR ac IN 1..permissions(i).perm.acl.COUNT LOOP
             j:=j+1;
             bulkAcl(j):=ACLEntryIdO(
             itemCheck(idx).guid,
             Permission_Util.getPrincipalIdIUW(permissions(i).perm.acl(ac).principal,null,1),
             permissions(i).perm.acl(ac).perm
             );
           END LOOP;
         END IF; --acl not null
         -- /acls
      END IF;
   END LOOP;
  nerrorsBp:=nerrorsBp+nerrors;
   -- for perms ok, update permissions fields
  	BEGIN
		FORALL i IN 1..pids.LAST save EXCEPTIONS
			UPDATE fc
			SET
				--principal_id=pids(i), --TODO: no ownership unless superuser..
				group_id=gids(i),
				userperm=ups(i),
				groupperm=gps(i),
				otherperm=ops(i)
				WHERE
				guid=guids(i)
            ;
	EXCEPTION
		WHEN bulk_errors THEN
		nerrors := SQL%BULK_EXCEPTIONS.COUNT;
		outerrors.EXTEND(nerrors);
		FOR i IN 1..nerrors LOOP
			outerrors(i+nerrorsBp):=
				returnO
					(
					guids(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX),
					SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
					Permission_Util.generalInternalBulkErrorVal,
					SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| 'at index '||i ,1,1024)
					);
        bulkAcl.DELETE(guidMarkerAcl(permissions(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).item).a,
							   guidMarkerAcl(permissions(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).item).b);
		END LOOP;
		nerrorsBp:=nerrorsBp+nerrors;
      END;

   -- delete from acls entries not existent in input acls
   --items to delete - exist in acl but not in input
   delete from acl
    where rowid in (
    SELECT /*+ cardinality( d 100 ) */ a.rowid
     FROM
     acl a,
     TABLE(bulkAcl) d
     WHERE
       a.guid=d.guid
     MINUS
     SELECT /*+ cardinality( d 100 ) */ a.rowid
     FROM
     acl a,
     TABLE(bulkAcl) d
     WHERE
       a.guid=d.guid AND a.principal_id=d.principal_id
    )
    ;

--   -- merge into acls for the rest
   	MERGE /*+ cardinality (S 100) */ INTO acl D
      USING (SELECT guid,principal_id,perm FROM TABLE(CAST(bulkACl AS ACLIdList))) S
      ON (S.principal_id=D.principal_id AND S.guid=D.guid)
      WHEN MATCHED THEN UPDATE SET perm=S.perm
      WHEN NOT MATCHED THEN INSERT VALUES(S.guid,S.principal_id,S.perm);


	RETURN ret;
END;

/*+ public void addReplica(String lfn, SURLEntry[] newSurls) throws CatalogException, AuthorizationException, ExistsException, InvalidStatException;*/
FUNCTION addReplica(securityToken IN principalToken,
          lfn IN VARCHAR2,
          newSurls IN surlEntryT,
          outerrors OUT NOCOPY returnOT
          )
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
   rc_sRec 	rc_s;
   inSurl surl_T;

   inGuid VARCHAR2(36);
    parentCheckVal PLS_INTEGER;

   /* const */
   timings TIMESTAMP(0);


	/* errors */
	nerrors PLS_INTEGER:=0;
   nerrorsBp PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
   errSurl rcs.surl%TYPE;

   errStr VARCHAR2(300);

	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);

   -- for all lfns check perms, if ok, get guids for ops
   --   insert newSurls into rcs
   -- report errors

  DEBUGON();

  begin
	 SELECT
    TREAT (res AS resolverO).guid,
    TREAT (res AS resolverO).permission
   COLLECT INTO inGuid, parentCheckVal
	  FROM(
  SELECT
	permission_util.isOkAclX(--check perms and return error or data as usual but using non-bulk check
	  NVL(CASE
            WHEN TYPE IN ( 'f', 'd')  THEN
              resolverO(parent_id,file_id,name,guid,principal_id,group_id,userperm,groupperm,otherperm) --also do permissions
			  	   WHEN TYPE ='s' THEN
              fireman.getSymlink(DATA) --also do permissions
					ELSE
					NULL
				END
			 ,resolverO(0,2))
			 ,userPrincipal,principals,Permission_Util.setmetaOp) res
       FROM
      FC_FILE f
     WHERE LEVEL=fireman.getNoOfPathDelimitersRoot(lfn)+1
      START WITH parent_id=0 AND file_id=1
      CONNECT BY PRIOR file_id = parent_id -- check if 'd' only valid paths..
		AND NAME=fireman.getpathtoken(LEVEL-1,lfn)
		);

   EXCEPTION
      WHEN NO_DATA_FOUND THEN
         ret.result:=Permission_Util.NoSuchFileVal;
         ret.errString:=Permission_Util.NoSuchFileStr;
      RETURN ret;
   END;
    IF parentCheckVal=0 THEN
      ret.result:=Permission_Util.accessDeniedErrVal;
      ret.errString:=Permission_Util.accessDeniedErrStr;
      RETURN ret;
    elsif parentCheckVal=permission_util.loopInSymlinksDetectedVal then
      ret.result:=Permission_Util.loopInSymlinksDetectedVal;
      ret.errString:=Permission_Util.loopInSymlinksDetectedStr;
     RETURN ret;
    END IF;


   timings:=SYSTIMESTAMP();
   FOR i IN 1..newSurls.LAST LOOP
      rc_sRec.guid:=inGuid;
      rc_sRec.surl:=newSurls(i).surl;
      rc_sRec.master_replica:=newSurls(i).master_replica;
      rc_sRec.surl_lifetime:=0;
      rc_sRec.se:=NVL(newSurls(i).se,'dummy');
      rc_sRec.cTime:=NVL(newSurls(i).creationTime,timings);
      rc_sRec.mTime:=NVL(newSurls(i).modifyTime,timings);
      inSurl(i):=rc_sRec;
   END LOOP;
   BEGIN
    FORALL i IN 1..newSurls.LAST save EXCEPTIONS
         INSERT INTO rcs
            VALUES inSurl(i)
            ;

   EXCEPTION
      WHEN bulk_errors THEN
         nbulkErr := SQL%BULK_EXCEPTIONS.COUNT;
         outerrors.EXTEND(nbulkErr);
         FOR i IN 1..nbulkErr LOOP
            errSurl:=newSurls(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).surl;
            outerrors(nerrors+i):=
               returnO
                  (
                  errSurl,
                  SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
                  Permission_Util.generalInternalBulkErrorVal,
                  SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
                  );
            errStr:=SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,200);
         END LOOP;
   END; --exception


	RETURN ret;
END;

/*+ public void remove(String[] lfns) throws CatalogException, AuthorizationException, NotExistsException;*/
FUNCTION remove(securityToken IN principalToken,
          lfns IN lfnT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
   itemCheck  guidSimpleT;

  	/* errors */
  j PLS_INTEGER:=0;
	nerrors PLS_INTEGER:=0;
   nerrorsBp PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
   errGuid VARCHAR2(36);
   errSurl rcs.surl%TYPE;


	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);

   --remove all form rcs, acl, fc where lfn is permitted
      -- in the case of error
      -- triggers for acl, rcs (item err resolution:3)

   --check perms for deletion and if it's file
   -- should not be able to

 SELECT
 CASE treat(res as resolverO).permission
 	WHEN 1  THEN treat(res as resolverO).guid
 ELSE NULL END val,
 CASE treat(res as resolverO).permission
      when 1 then null
  	  WHEN 0 THEN returnO
              (treat(res as resolverO).fname,treat(res as resolverO).idx,
               Permission_Util.accessDeniedErrVal,  --no right to write
               Permission_Util.accessDeniedErrStr
              )
      WHEN 2 THEN returnO
              (treat(res as resolverO).fname,treat(res as resolverO).idx,
               Permission_Util.NoSuchFileVal, --no parent directory exists
               Permission_Util.NoSuchFileStr
              )
      WHEN permission_util.loopInSymlinksDetectedVal THEN   		
        returnO(TREAT(res AS resolverO).fname,
              TREAT(res AS resolverO).idx,
              Permission_Util.loopInSymlinksDetectedVal, --no parent directory exists
              Permission_Util.loopInSymlinksDetectedStr
              )
   		ELSE
   		returnO
              (treat(res as resolverO).fname,treat(res as resolverO).idx,
               Permission_Util.NoSuchFileVal, --no parent directory exists
               Permission_Util.NoSuchFileStr
              )
   		END ret
  bulk collect into itemCheck,outErrors
 FROM (
 SELECT
	permission_util.isOkAclX(--check perms and return error or data as usual but using non-bulk check
	  NVL((SELECT
	   		  CASE WHEN TYPE IN ( 'f', 'd')  THEN
         resolverO(idx,parent_id,file_id,
            lfn,guid,type,ctime,mtime,
            file_size,file_checksum,LFN_LIFETIME,status,
            principal_id,group_id,userperm,groupperm,otherperm,
            dangling_ok,
            data,-1)
			  	   WHEN TYPE ='s' THEN
				   	resolverO(idx,fireman.getSymlink(DATA)) --also do permissions
					ELSE
					NULL
				   END
  	   		    FROM
		            FC_FILE f
                WHERE LEVEL=case namelfn when ':' then 1 else 2 end -- always leaf or root if /
		            START WITH parent_id=z.val.parent_id AND file_id=z.val.file_id
		            CONNECT BY PRIOR file_id = parent_id -- check if 'd' only valid paths..
		              AND NAME=namelfn
			 ),resolverO(idx,2,lfn))
			 ,userPrincipal,principals,Permission_Util.removeFileOp) RES
  FROM
  (SELECT
		SUBSTR(lfn,1,INSTR(lfn,'/',-1,1)) dir,
		nvl(SUBSTR(lfn,INSTR(lfn,'/',-1,1)+1),':') nameLfn,
		lfn,
		idx
    FROM
    (
      SELECT /*+ cardinality (iz, 50) */
        COLUMN_VALUE AS lfn, ROWNUM AS idx
        FROM
        TABLE(lfns) iz
    )
	) i
  LEFT JOIN
	(
	SELECT (
    SELECT updateFCSimpleO(parent_id,file_id,NULL,dir)
    FROM
      FC_FILE f
      WHERE LEVEL=Fireman.getNoOfPathDelimiters(dir)
      START WITH parent_id=0 AND file_id=1
      CONNECT BY PRIOR file_id = parent_id AND TYPE='d' -- check if 'd' only valid paths..
        AND NAME=Fireman.getpathtoken(LEVEL-1,dir)
        ) val
    FROM
    (
    SELECT /*+ cardinality (dirz, 50) */
      DISTINCT SUBSTR(column_value,1,INSTR(column_value,'/',-1,1)) dir
    FROM
    table(lfns)dirz
    )
	) z
	 ON (i.dir=z.val.item)
	 )
	 ;
   -- remove nulls from the output, make it sparse
   j:=itemCheck.count;
    FOR i IN 1..j LOOP
      IF itemCheck(i) IS NULL THEN
        itemcheck.DELETE(i);
        nerrors:=nerrors+1;
      ELSE
        outerrors.DELETE(i);
      END IF;
    END LOOP;

   nerrorsBp:=nerrorsBp+nerrors;
   -- do this dirty job, can use pernt_id, user_id later
  	BEGIN
		FORALL i IN indices of itemCheck save EXCEPTIONS
			DELETE FROM fc
            WHERE guid=itemCheck(i)
            ;
	EXCEPTION
		WHEN bulk_errors THEN
		nerrors := SQL%BULK_EXCEPTIONS.COUNT;
		outerrors.EXTEND(nerrors);
		FOR i IN 1..nerrors LOOP
			outerrors(i+nerrorsBp):=
				returnO
					(
					lfns(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX),
					SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
					Permission_Util.generalInternalBulkErrorVal,
					SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| 'at index '||i ,1,1024)
					);
		END LOOP;
		nerrorsBp:=nerrorsBp+nerrors;
      END;
   /*
      deletes are propagated down to acl, surls..
      hence possiblity to get detailed err from common triggering logic..
   */


	RETURN ret;
END;

/*+ public void removeReplica(StringPair[] entries) throws CatalogException, AuthorizationException, NotExistsException;*/
FUNCTION removeReplica(securityToken IN principalToken,
          entries IN StringPairT,
          removeLfnOnLastReplica IN SMALLINT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
   itemCheck  StringPairT;

   guids gGuidsT ;
   surls ssurlT ;
   idxs listnum;
   j PLS_INTEGER:=0;

  	/* errors */
	nerrors PLS_INTEGER:=0;
  nerrorsBp PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
   errGuid VARCHAR2(36);
   errSurl rcs.surl%TYPE;


	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);

   --remove all form rcs, acl, fc where lfn is permitted
      -- in the case of error
      -- triggers for acl, rcs (item err resolution:1)

   --check perms for deletion
--   SELECT
--   (
--   SELECT
--   --updateFCSimpleO(parent_id,file_id,guid)
--   StringPairO(guid,i.string2)
-- 	FROM
--	FC_FILE f
--	WHERE LEVEL=getNoOfPathDelimitersRoot(i.string1)+1
--	AND Permission_Util.isOkAcl(userperm,groupperm,otherperm,
--					principal_id,userPrincipal,
--					group_id,
--			      principals,
--					guid, --input guid
--					Permission_Util.setmetaOp)=1
--	START WITH file_id=1
--	CONNECT BY PRIOR file_id = parent_id
--	AND NAME =getpathtoken(LEVEL-1,i.string1)
--	)
--   BULK COLLECT INTO itemCheck
--   FROM TABLE(entries) i; --beware of stale ids later...
--
--

 SELECT
 CASE treat(res as resolverO).permission
 	WHEN 1  THEN treat(res as resolverO).guid
 ELSE NULL END g,
 CASE treat(res as resolverO).permission
 	WHEN 1  THEN treat(res as resolverO).idx
 ELSE NULL END,
 CASE treat(res as resolverO).permission
      when 1 then null
  	  WHEN 0 THEN returnO
              (treat(res as resolverO).fname,treat(res as resolverO).idx,
               Permission_Util.accessDeniedErrVal,  --no right to write
               Permission_Util.accessDeniedErrStr
              )
      WHEN 2 THEN returnO
              (treat(res as resolverO).fname,treat(res as resolverO).idx,
               Permission_Util.NoSuchFileVal, --no parent directory exists
               Permission_Util.NoSuchFileStr
              )
      WHEN permission_util.loopInSymlinksDetectedVal THEN   		
        returnO(TREAT(res AS resolverO).fname,
              TREAT(res AS resolverO).idx,
              Permission_Util.loopInSymlinksDetectedVal, --no parent directory exists
              Permission_Util.loopInSymlinksDetectedStr
              )
   		ELSE
   		returnO
              (treat(res as resolverO).fname,treat(res as resolverO).idx,
               Permission_Util.NoSuchFileVal, --no parent directory exists
               Permission_Util.NoSuchFileStr
              )
   		END ret
  bulk collect into guids,idxs,outErrors
 FROM (
 SELECT
	permission_util.isOkAclX(--check perms and return error or data as usual but using non-bulk check
	  NVL((SELECT
	   		  CASE WHEN TYPE IN ( 'f', 'd')  THEN
         resolverO(idx,parent_id,file_id,
            lfn,guid,type,ctime,mtime,
            file_size,file_checksum,LFN_LIFETIME,status,
            principal_id,group_id,userperm,groupperm,otherperm,
            dangling_ok,
            data,-1)
			  	   WHEN TYPE ='s' THEN
				   	resolverO(idx,fireman.getSymlink(DATA)) --also do permissions
					ELSE
					NULL
				   END
  	   		    FROM
		            FC_FILE f
                WHERE LEVEL=case namelfn when ':' then 1 else 2 end -- always leaf or root if /
		            START WITH parent_id=z.val.parent_id AND file_id=z.val.file_id
		            CONNECT BY PRIOR file_id = parent_id -- check if 'd' only valid paths..
		              AND NAME=namelfn
			 ),resolverO(idx,2,lfn))
			 ,userPrincipal,principals,Permission_Util.removeFileOp) RES
  FROM
  (SELECT
		SUBSTR(lfn,1,INSTR(lfn,'/',-1,1)) dir,
		nvl(SUBSTR(lfn,INSTR(lfn,'/',-1,1)+1),':') nameLfn,
		lfn,
		idx
    FROM
    (
      SELECT /*+ cardinality (iz, 50) */
        string1 AS lfn, ROWNUM AS idx
        FROM
        TABLE(entries) iz
    )
	) i
  LEFT JOIN
	(
	SELECT (
    SELECT updateFCSimpleO(parent_id,file_id,NULL,dir)
    FROM
      FC_FILE f
      WHERE LEVEL=Fireman.getNoOfPathDelimiters(dir)
      START WITH parent_id=0 AND file_id=1
      CONNECT BY PRIOR file_id = parent_id AND TYPE='d' -- check if 'd' only valid paths..
        AND NAME=Fireman.getpathtoken(LEVEL-1,dir)
        ) val
    FROM
    (
    SELECT /*+ cardinality (dirz, 50) */
      DISTINCT SUBSTR(string1,1,INSTR(string1,'/',-1,1)) dir
    FROM
    table(entries)dirz
    )
	) z
	 ON (i.dir=z.val.item)
	 )
	 ;
   -- remove nulls from the output, make it sparse
   j:=guids.count;
    FOR i IN 1..j LOOP
      IF itemCheck(i) IS NULL THEN
        itemcheck.DELETE(i);
        guids.delete(i);
        nerrors:=nerrors+1;
      ELSE
        outerrors.DELETE(i);
        surls(i):=entries(idxs(i)).string2;
      END IF;
    END LOOP;

   nerrorsBp:=nerrorsBp+nerrors;
   -- do this dirty job, can use pernt_id, user_id later
  	BEGIN
		FORALL i IN indices of guids save EXCEPTIONS
			DELETE FROM rcs
            WHERE guid=guids(i) AND surl=surls(i)
            ;

	EXCEPTION
		WHEN bulk_errors THEN
		nerrors := SQL%BULK_EXCEPTIONS.COUNT;
		outerrors.EXTEND(nerrors);
		FOR i IN 1..nerrors LOOP
			outerrors(i+nerrorsBp):=
				returnO
					(
					entries(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).string1,
					SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
					Permission_Util.generalInternalBulkErrorVal,
					SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| 'at index '||i ,1,1024)
					);
         guids.DELETE((SQL%BULK_EXCEPTIONS(i).ERROR_INDEX));
		END LOOP;
		nerrorsBp:=nerrorsBp+nerrors;
      END;
   /*
      deletes are propagated down to acl, surls..
      hence possiblity to get detailed err from common triggering logic..
   */
	/* check if any of guids became dangling one, remove them */
	--delete from (select /* +USE_NL(i) */ g.rowid from rcg g,gg i where g.guid=i.guid and not exists (select guid from rcs s where s.guid=g.guid));

	FORALL i IN 1..surls.LAST save EXCEPTIONS
	DELETE FROM
		(SELECT
			g.ROWID
			FROM fc g WHERE g.guid=guids(i)
				AND NOT EXISTS (SELECT guid FROM rcs s WHERE s.guid=g.guid));

	/* handle errors */
 RETURN ret;
 EXCEPTION
  	WHEN bulk_errors THEN
	nerrors := SQL%BULK_EXCEPTIONS.COUNT;
	outerrors.EXTEND(nerrors);
	FOR i IN 1..nerrors LOOP
		outerrors(i+nerrorsBp):=
			returnO
				(
				guids(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX),
				SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
				Permission_Util.generalInternalBulkErrorVal,
				SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
				SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
				);
	END LOOP;
	RETURN ret;


END;

/*+ public GUIDStat[] listGuidsWithoutLfn(String filter) throws CatalogException, AuthorizationException;*/
FUNCTION listGuidsWithoutLfn(securityToken IN principalToken,
          FILTER IN VARCHAR2,
			 outEntries OUT NOCOPY guidStatT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	parentCheck lfnCheckT;


	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
	 Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);

	RETURN ret;
END;

/************************** FC derived *****************/

/*+ public void createFile(FCEntry[] entries) */
FUNCTION createFile(securityToken IN principalToken,
          entries IN FCEntryT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	permCheck FCRStatusTT;
	inFc fc_T;
	inAcl acl_T;

  rc_sRec 	rc_s;
  fc_sRec fc_R;

  /* const */
  timings TIMESTAMP(0);


  guidMarkerSurlPerms triplesT;

  aclMarkers triplesT;
  aclMarker tripleR;
	k PLS_INTEGER:=0;
  j PLS_INTEGER:=0;
  sCounter PLS_INTEGER:=1;
  aclCounter PLS_INTEGER:=0;

  masterSurlGuard PLS_INTEGER:=0;
  defaultPermFetched BOOLEAN:=FALSE;
  defaultPermission PermissionO;
  tmpPermission PermissionSimpleO;

	/* errors */
	nerrors PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
  errGuid VARCHAR2(36);
  errSurl rcs.surl%TYPE;
  nBadItems PLS_INTEGER:=0;
  debugMsg VARCHAR2(500);

	/* return */
	ret returnO:=returnO(0);
BEGIN
	outerrors:=returnOT();

   DEBUGON();
   Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);


	-- remember order - no need as we do explicit nl
	-- pivot to root dirs paths, get list of their kids
	-- we may need it if once we intercept pattern with many shared creation nodes for
	-- v. big input sets

	-- think if entries are too big for input..
	-- starts from the hidden root, so perms are evaluated for '/bla' like paths
	SELECT /*+ cardinality (i 100) */
	 (
		CAST(MULTISET
		(
		SELECT /* +use_nl(f) +cardinality (b 2) */
        -1,
         CASE connect_by_isleaf
          when 1 THEN
          Permission_Util.isOkAcl(userperm,groupperm,otherperm,
						principal_id,userPrincipal,
						group_id,
						principals,
						guid, --input guid
						Permission_Util.writeFileOp)
           else null
           end case,
            file_id,
            parent_id,
            CASE connect_by_isleaf
			  	   WHEN 1  THEN
              nvl2(i.permission,
                PermissionSimpleO(
                  userPrincipal, --principal_id
                  NULL,--Permission_Util.getprincipalId(i.permission.groupname,'g',1), --group_id
                  i.permission.userperm,i.permission.groupperm,i.permission.otherperm,
                  Permission_Util.translateAclToSimpleAcl(i.permission.acl)), --transalte acls to ids if there are any..
                  -- transalte to simple
                PermissionSimpleO(userPrincipal,group_id,userperm,groupperm,otherperm, --TODO: null if not needed (perm already in the entry)
                  CAST(MULTISET
                    (
                    SELECT p.principal_id,acl.perm
                      FROM acl,PERM_PRINCIPAL p
                        WHERE guid=f.guid AND p.principal_id=acl.principal_id
                    ) AS ACLSimpleIdList)
                  )
                ) --nvl
              ELSE NULL END
			FROM
			FC_FILE f
			WHERE LEVEL>getNoOfPathDelimiters(i.lfn)-1
			START WITH parent_id=0 AND file_id=1 --and name=getpathtoken(1,i.lfn)
			CONNECT BY PRIOR file_id = parent_id AND type='d' --!!!! check if 'd' only valid paths..
			  AND NAME=getpathtoken(LEVEL-1,i.lfn)
			) AS FCRStatusT)
		)
		BULK COLLECT INTO permCheck
		FROM TABLE(entries) i
		;


  timings:=SYSTIMESTAMP();
	FOR i IN 1..permCheck.COUNT LOOP

	 IF permCheck.EXISTS(i) THEN
	 BEGIN
		IF permCheck(i).EXISTS(2) THEN -- file already there, error
		BEGIN
			IF permCheck(i)(2).permitted=1 THEN
				-- report already existing one
				nerrors:=nerrors+1;
				outerrors.EXTEND();
				outerrors(nerrors):=returnO
               (entries(i).lfn,i,
               Permission_Util.fileOrDirAlreadyExistsVal,
               Permission_Util.fileOrDirAlreadyExistsStr);
			ELSE
				-- report no permission
        nerrors:=nerrors+1;
				outerrors.EXTEND();
				outerrors(nerrors):=returnO
               (entries(i).lfn,i,
               Permission_Util.accessDeniedErrVal,
               Permission_Util.accessDeniedErrStr);
			END IF;
		END;
		ELSIF permCheck(i).EXISTS(1) THEN --parent dir is there
		--begin
			IF permCheck(i)(1).permitted=1 THEN --permission in the parent directory ok
				--ok to go
				fc_sRec.parent_id:=permCheck(i)(1).file_id;
				SELECT fc_file_id_seq.NEXTVAL INTO fc_sRec.file_id FROM dual;
				fc_sRec.NAME:= SUBSTR(entries(i).lfn,INSTR(entries(i).lfn,'/',-1)+1);--get the last part!
				fc_sRec.type:=entries(i).lfnstat.lfntype; --has to be f
				fc_sRec.guid:=entries(i).guid;
				--if entries(i).lfnStat.creationTime=0 then
				fc_sRec.cTime:=NVL(entries(i).lfnStat.creationTime,timings);
				fc_sRec.mTime:=NVL(entries(i).lfnStat.mTime,timings);

				fc_sRec.file_size:=entries(i).lfnStat.fileSize;
				fc_sRec.lfn_lifetime:=entries(i).lfnStat.validitytTime;
				fc_sRec.file_checksum:='';
				fc_sRec.status:=0;
        fc_sRec.data:=entries(i).lfnStat.data;

        /*
        take parent permission with acls or provided permission if not empty
        if there are no permissions provided (always skip ownership),
          take them from the parent directory
        if provided use them (except userperm)
        */

        tmpPermission:=permCheck(i)(1).permission;

  			-- permissions logic
        /* always take ownership (owner, not group:3) from the token */
        fc_sRec.principal_id:=tmpPermission.principal_id;
        -- we don't get/create group or acl users
        -- create group here - important dependency on the ordering
        -- otherwise use asso array with the lfn
				--fc_sRec.group_id:=tmpPermission.group_id;
        fc_sRec.group_id:=Permission_Util.getprincipalIdIUW(entries(i).permission.groupname,null,1);
				fc_sRec.userperm:=tmpPermission.userperm;
				fc_sRec.groupperm:=tmpPermission.groupperm;
				fc_sRec.otherperm:=tmpPermission.otherperm;

				fc_sRec.dangling_ok:=0;
				--fc_sRec.DATA:=entries(i).lfnstat.DATA; alter table
        --fetch surl and acl data too

        inFc(i):=fc_sRec;  --add lfn to the story
        -----------------------
        -- handle acl's -------
        BEGIN --acls
        IF  tmpPermission.acl IS NOT NULL AND tmpPermission.acl.COUNT>0 THEN
         aclMarker.a:=aclCounter+1;
         aclMarker.b:=aclCounter+tmpPermission.acl.COUNT;
         aclMarker.c:=i;
         aclMarkers(fc_sRec.guid):=aclMarker;

         FOR ac IN 1..tmpPermission.acl.COUNT LOOP
            aclCounter:=aclCounter+1;
            inAcl(aclCounter).guid:=fc_sRec.guid;
            inAcl(aclCounter).principal_id:=tmpPermission.acl(ac).principal_id; --change it to one call per all perms
            inAcl(aclCounter).perm:=tmpPermission.acl(ac).perm;
        END LOOP;
        END IF;
        END; --acls
        masterSurlGuard:=0;
			ELSE
            BEGIN
				/*  error, parent is there but no permission to create */
            /* report no permission				*/
            nerrors:=nerrors+1;
            outerrors.EXTEND();
            outerrors(nerrors):=returnO
               (entries(i).lfn,i,
               Permission_Util.accessDeniedErrVal,
               Permission_Util.accessDeniedErrStr);
            END;
			END IF;
	   --end;
		ELSE --parent is null
			--error, nonexisting path
         BEGIN
         nerrors:=nerrors+1;
         outerrors.EXTEND;
         outerrors(nerrors):=returnO
                  (
                  entries(i).lfn,i,
                  Permission_Util.NoSuchFileVal,
                  Permission_Util.NoSuchFileStr
                  );
         END;
		END IF;
		END;
	END IF;
	END LOOP;

   IF inFc.COUNT=0 THEN
      ret.result:=Permission_Util.missingDataVal;
      ret.errString:=Permission_Util.missingDataStr;
      RETURN ret;
   END IF;
	--insert into surl,acl,fc (name inserts already during deserialization)
   ------------ fc insert -----------
   BEGIN
      FORALL i IN 1..inFc.LAST save EXCEPTIONS
         INSERT INTO fc
            VALUES inFc(i);
   EXCEPTION
      WHEN bulk_errors THEN
         nbulkErr := SQL%BULK_EXCEPTIONS.COUNT;
         outerrors.EXTEND(nbulkErr);
         FOR i IN 1..nbulkErr LOOP
            outerrors(nerrors+i):=
               returnO
                  (
                  entries(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).lfn, --entries always preserve order:1
                  SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
                  Permission_Util.generalInternalBulkErrorVal,
                  SUBSTR('Guid/LFN creation related error. Check input data'||
                  SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
                  );
            --delete from ingested surls, acl to be ingested
            errGuid:=inFc(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid;
            IF aclMarkers.EXISTS(errGuid) THEN
               inAcl.DELETE(aclMarkers(errGuid).a,aclMarkers(errGuid).b);
            END IF;
         END LOOP;
         nerrors:=nbulkErr+nerrors;
	END; --exception


   ------------ acl insert ---------
   BEGIN
      FORALL i IN 1..inACl.COUNT save EXCEPTIONS
         INSERT INTO acl
            VALUES inAcl(i);
   EXCEPTION
      WHEN bulk_errors THEN
         nbulkErr := SQL%BULK_EXCEPTIONS.COUNT;
         outerrors.EXTEND(nbulkErr);
         FOR i IN 1..nbulkErr LOOP
            errGuid:=inAcl(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid;
            outerrors(nerrors+i):=
               returnO
                  (
                  entries(aclMarkers(errGuid).c).lfn,
                  aclMarkers(errGuid).c,
                  Permission_Util.generalInternalBulkErrorVal,
                  SUBSTR('ACL related data.'||
                    SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
                  );
            -- in the future we may still keep such entry as valid (and not back-delete)
            -- nowadays:
            DELETE FROM fc WHERE guid=errGuid;
            DELETE FROM rcs WHERE guid=errGuid;
         END LOOP;
         nerrors:=nbulkErr+nerrors;
	END; --exception

   -- do escalation error check to set general resul as error o warnign eventually.
   RETURN ret;
  EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
		RETURN ret;
END;



/*+ public void mkdir(String[] directory, boolean createParents) throws CatalogException, AuthorizationException, ExistsException, NotExistsException;*/
FUNCTION mkdir(securityToken IN principalToken,
          entries IN lfnT,
          wParents IN SMALLINT,
          copyPermFromParent IN SMALLINT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
  /* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */

	 permCheck FCRStatusTT;
	 inFc fc_T;
   fc_sRec fc_R;
   inAcl acl_T;

   /* parents serializers */
   inputCheck FMRStatusT;

   aclMarkers triplesT;
   aclMarker tripleR;
   k PLS_INTEGER:=0;
   j PLS_INTEGER:=0;

   aclCounter PLS_INTEGER:=0;

   defaultPermFetched BOOLEAN:=FALSE;
   defaultPermission PermissionO;
   tmpPermission PermissionSimpleO;
   tmpDefaultPermission PermissionSimpleO;
   defaultGroupId FC_FILE.group_id%TYPE;
   /* const */
   timings TIMESTAMP(0);



	/* errors */
	nerrors PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
  errGuid VARCHAR2(36);

	/* return */
	ret returnO:=returnO(0);
BEGIN
   DEBUGON();
   outerrors:=returnOT();
	 Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);
   /* TODO: ignore wparents for now*/
   /* same as create wo rcs creation and
      default:
         perms taken from parent
         ctime,mtime set
         filesize=0
         generated guid
         type='d'
      */


    -- for non-parents do regular check of the last one.
    if wParents=0 then

      SELECT /*+ cardinality (i 100) */
     (
      CAST(MULTISET
      (
      SELECT /* +use_nl(f) +cardinality (b 2) */
          i.idx,
          --permisison-
          Permission_Util.isOkAcl(userperm,groupperm,otherperm,
              principal_id,userPrincipal,
              group_id,
              principals,
              guid, --input guid
              Permission_Util.writeFileOp),
              file_id,
              parent_id,
              case
              when copyPermFromParent=1 and connect_by_isleaf=1 THEN
                PermissionSimpleO
                (
                userprincipal,group_id,userperm,groupperm,otherperm,
                --permission_util.getacl(guid)
                CAST(MULTISET
                    (
                    SELECT p.principal_id,acl.perm
                      FROM acl,PERM_PRINCIPAL p
                    WHERE guid=f.guid AND p.principal_id=acl.principal_id
                    ) AS ACLSimpleIdList)
                )
              else null end
        FROM
        FC_FILE f
        WHERE LEVEL>getNoOfPathDelimiters(i.COLUMN_VALUE)-1 --take both but just one permisson
        START WITH parent_id=0 AND file_id=1 --and name=getpathtoken(1,i.lfn)
        CONNECT BY PRIOR file_id = parent_id AND type='d' --!!!! check if 'd' only valid paths..
          AND NAME=getpathtoken(LEVEL-1,i.COLUMN_VALUE)
        ) AS FCRStatusT)
      )
      BULK COLLECT INTO permCheck
            FROM (select /*+cardinality (z 10) */ rownum idx, column_value from TABLE(entries) z ) i
      ;


  --generic handling of inserts
  timings:=SYSTIMESTAMP;
FOR i IN 1..permCheck.COUNT LOOP
	 IF permCheck.EXISTS(i) THEN
	 BEGIN
		IF permCheck(i).EXISTS(2) THEN -- file already there, error
		BEGIN
			IF permCheck(i)(2).permitted=1 THEN
				-- report already existing one
				nerrors:=nerrors+1;
				outerrors.EXTEND();
				outerrors(nerrors):=returnO
               (entries(i),
               i,
               Permission_Util.fileOrDirAlreadyExistsVal,
               Permission_Util.fileOrDirAlreadyExistsStr);
			ELSE
				-- report no permission
        nerrors:=nerrors+1;
				outerrors.EXTEND();
				outerrors(nerrors):=returnO
               (entries(i),
               i,
               Permission_Util.accessDeniedErrVal,
               Permission_Util.accessDeniedErrStr);
			END IF;
		END;
		ELSIF permCheck(i).EXISTS(1) THEN --parent dir is there
		--begin
			IF permCheck(i)(1).permitted=1 THEN --permission in the parent directory ok
				--ok to go
				fc_sRec.parent_id:=permCheck(i)(1).file_id;
				SELECT fc_file_id_seq.NEXTVAL INTO fc_sRec.file_id FROM dual;
				--fc_sRec.file_id:=;
				fc_sRec.NAME:= SUBSTR(entries(i),INSTR(entries(i),'/',-1)+1);--get the last part!
				fc_sRec.type:='d'; -- dir always
				fc_sRec.guid:=generateGuid; --TODO: think of delegating it to java
				fc_sRec.cTime:=timings;
				fc_sRec.mTime:=timings;
				fc_sRec.file_size:=0;
				fc_sRec.lfn_lifetime:=0;
				fc_sRec.file_checksum:=' ';
				fc_sRec.status:=0;
        --perms
        --use user credentials

        IF copyPermFromParent>0 THEN
          tmpPermission:=permCheck(i)(1).permission;
          --take also acls in the acl section.. get acl from db now?
        ELSE

        -- take defaults if not taken yet
          IF defaultPermFetched=FALSE THEN
            ret:=getDefaultPermission(securityToken,
                      defaultPermission,
                      outErrors);--outerrors not really used
            if ret.result<>0 then
              return ret;
            end if;
            tmpDefaultPermission:=PermissionSimpleO(0,0,0,0,0,NULL);
            translatePermtoSimple(defaultPermission,tmpDefaultPermission);
            tmpPermission:=tmpDefaultPermission;
            -- put groupid if exists
            if principals is not null and principals.count>0
              and principals(1) is not null and principals(1).item2 is not null
            then
              tmpPermission.group_id:=principals(1).item2;
            end if;

            defaultPermFetched:=TRUE;
          ELSE

            tmpPermission:=tmpDefaultPermission;
          END IF; --defaults
        END IF; --copy from parent

        fc_sRec.principal_id:=  userPrincipal;
        fc_sRec.group_id :=  tmpPermission.group_id;
        fc_sRec.userperm :=  tmpPermission.userperm;
        fc_sRec.groupperm:=  tmpPermission.groupperm;
        fc_sRec.otherperm:=  tmpPermission.otherperm;
        --handle acls in the acl section
				fc_sRec.dangling_ok:=0;

        inFc(i):=fc_sRec;  --add lfn to the story
        -----------------------
        -- handle acl's -------
        BEGIN --acls

          IF  tmpPermission.acl IS NOT NULL AND tmpPermission.acl.COUNT>0 THEN
             aclMarker.a:=aclCounter+1;
             aclMarker.b:=aclCounter+tmpPermission.acl.COUNT;
             aclMarker.c:=i;
             aclMarkers(fc_sRec.guid):=aclMarker;
              FOR ac IN 1..tmpPermission.acl.COUNT LOOP
                aclCounter:=aclCounter+1;
                inAcl(aclCounter).guid:=fc_sRec.guid;
                inAcl(aclCounter).principal_id:=tmpPermission.acl(ac).principal_id;
                inAcl(aclCounter).perm:=tmpPermission.acl(ac).perm;
              END LOOP;
          END IF;

        END; --acls
			ELSE
          BEGIN
				--error, parent is there but no permission to create
            -- report no permission
         	nerrors:=nerrors+1;
          outerrors.EXTEND();
          outerrors(nerrors):=returnO
               (entries(i),
               i,
               Permission_Util.NoSuchFileVal,
               Permission_Util.NoSuchFileStr);
          END;
			END IF;
	   --end;
		ELSE --parent is null
         BEGIN
			--error, nonexisting path
         nerrors:=nerrors+1;
         outerrors.EXTEND;
         outerrors(nerrors):=returnO
                  (
                  entries(i),
                  i,
                  Permission_Util.InvalidParentVal,
                  Permission_Util.InvalidParentStr
                  );
         END;
		END IF;
		END;
	END IF;
	END LOOP;

   IF inFc.COUNT=0 THEN
      ret.result:=Permission_Util.missingDataVal;
      ret.errString:=Permission_Util.missingDataStr;
      RETURN ret;
   END IF;
	--insert into surl,acl,fc (name inserts already during deserialization)
   ------------ fc insert -----------
   BEGIN
      FORALL i IN 1..inFc.COUNT save EXCEPTIONS
         INSERT INTO fc
            VALUES inFc(i);
   EXCEPTION
      WHEN bulk_errors THEN
         nbulkErr := SQL%BULK_EXCEPTIONS.COUNT;
         outerrors.EXTEND(nbulkErr);
         FOR i IN 1..nbulkErr LOOP
            outerrors(nerrors+i):=
               returnO
                  (
                  entries(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX), --entries always preserve order:1
                  SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
                  Permission_Util.generalInternalBulkErrorVal,
                  SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
                  );
            --delete from acl to be ingested
            errGuid:=inFc(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid;
            IF aclMarkers.EXISTS(errGuid) THEN
               inAcl.DELETE(aclMarkers(errGuid).a,aclMarkers(errGuid).b);
            END IF;
         END LOOP;
         nerrors:=nbulkErr+nerrors;
	END; --exception

   ------------ acl insert -----------
   BEGIN
      FORALL i IN 1..inACl.COUNT save EXCEPTIONS
         INSERT INTO acl
            VALUES inAcl(i);
   EXCEPTION
      WHEN bulk_errors THEN
         nbulkErr := SQL%BULK_EXCEPTIONS.COUNT;
         outerrors.EXTEND(nbulkErr);
         FOR i IN 1..nbulkErr LOOP
            errGuid:=inAcl(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid;
            outerrors(nerrors+i):=
               returnO
                  (
                  entries(aclMarkers(errGuid).c),
                  aclMarkers(errGuid).c,
                  Permission_Util.generalInternalBulkErrorVal,
                  SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
                  );
            -- in the future we may still keep such entry as valid (and not back-delete)
            -- nowadays:
            DELETE FROM fc WHERE guid=errGuid;
         END LOOP;
         nerrors:=nbulkErr+nerrors;
	END; --exception
  else
  ret:=mkdirParents(securityToken,userPrincipal,principals,
    entries,
    copyPermFromParent,
    outerrors);

  end if;
	RETURN ret;
  EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR('General error in mkdir:'||SQLERRM, 1, 1024);
		RETURN ret;
END;


/*+ public void rmdir(String[] directory, boolean recursive) throws CatalogException, AuthorizationException, ExistsException, NotExistsException; */
FUNCTION rmdir(securityToken IN principalToken,
          entries IN lfnT,
          recursive IN SMALLINT, -- FIXME: remove, not needed anymore
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	dirCheck RmDirStatusT;
  files bulkPrincipalidT;
  parents bulkPrincipalidT;

		/* errors */
	nerrors PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
  errGuid VARCHAR2(36);
  errSurl rcs.surl%TYPE;

	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
   DEBUGON();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);
  --check if anything in the dir
  -- if sth is there error
  -- else delete
  -- for all dirs in input check permission


  --  check in the dir:
      -- for dirpath - check perms and if kids count = 0
        --returns
          --  0 on no rights
          --  1 ok
          --  2 non existence
          --  3 kids exist
      -- simliarly to creation except count check should stop with 1 kid entry

  SELECT /* + cardinality (i 100) */ NVL((
    SELECT /* +use_nl(f) */
            --permisison-
            RmDirStatusO(
            COLUMN_VALUE,
            parent_id,
            file_id,
            DECODE(Permission_Util.isOkAcl(userperm,groupperm,otherperm,
                principal_id,userPrincipal,
                group_id,
                principals,
                guid,
                Permission_Util.removeFileOp),
                0,0,
                1,
                DECODE
                  (
                    (SELECT COUNT(*) FROM dual WHERE EXISTS
                      (SELECT /*+FIRST_ROWS */ NULL
                        FROM fc WHERE parent_id=f.file_id
                      )
                    ),
                    0,1, -- no kids, ok
                    1,3  -- there are kids, err
                  )
                )
              )
    FROM
    FC_FILE f
    WHERE LEVEL=Fireman.getNoOfPathDelimitersRoot(COLUMN_VALUE)+1
    START WITH parent_id=0 AND file_id=1 --and name=getpathtoken(1,i.lfn)
    CONNECT BY PRIOR file_id = parent_id AND type='d' --!!!! check if 'd' only valid paths..
    AND NAME=Fireman.getpathtoken(LEVEL-1,COLUMN_VALUE)
    ),
    RmDirStatusO(
    COLUMN_VALUE,
    0,0,2)) v
  BULK COLLECT INTO dirCheck
  FROM
  TABLE(entries) i;

  -- handle erroneus items
  FOR i IN 1..dirCheck.COUNT LOOP
      IF dirCheck(i).permitted=1 THEN
        files(i):=dirCheck(i).file_id;
        parents(i):=dirCheck(i).parent_id;
      ELSIF dirCheck(i).permitted=0 THEN
        files(i):=NULL;
        parents(i):=NULL;
        nerrors:=nerrors+1;
				outerrors.EXTEND();
				outerrors(nerrors):=returnO
               (dirCheck(i).lfn,i,
               Permission_Util.accessDeniedErrVal,
               Permission_Util.accessDeniedErrStr);
      ELSIF dirCheck(i).permitted=2 THEN
        files(i):=NULL;
        parents(i):=NULL;
        nerrors:=nerrors+1;
				outerrors.EXTEND();
				outerrors(nerrors):=returnO
               (dirCheck(i).lfn,i,
               Permission_Util.NoSuchFileVal,
               Permission_Util.NoSuchFileStr);
      ELSIF dirCheck(i).permitted=3 THEN
        files(i):=NULL;
        parents(i):=NULL;
        nerrors:=nerrors+1;
				outerrors.EXTEND();
				outerrors(nerrors):=returnO
               (dirCheck(i).lfn,i,
               Permission_Util.directoryNotEmptyVal,
               Permission_Util.directoryNotEmptyStr);

      END IF;
  END LOOP;

  -- delete the rest, nulls are skipped internally
  -- delete is propagated via trigger
  BEGIN
  FORALL i IN 1..files.COUNT save EXCEPTIONS
    DELETE FROM fc
      WHERE parent_id=parents(i) AND file_id=files(i);
  EXCEPTION
	  /* handle errors thru */
		WHEN bulk_errors THEN
		nerrors := SQL%BULK_EXCEPTIONS.COUNT;
		outerrors.EXTEND(nerrors);
		FOR i IN 1..nerrors LOOP
			outerrors(i):=
				returnO
					(
					dirCheck(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).lfn,
					SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
					Permission_Util.generalInternalBulkErrorVal,
					SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||i ,1,1024)
					);
		END LOOP;
	END;		--exception

	RETURN ret;
END;



-- TODO: remove as obsolete
-- TODO: change error handling
/*+ public FCEntry[] getLfnStat(String[] lfn) throws CatalogException, AuthorizationException, NotExistsException;*/
function getLFNStat(securityToken in principalToken,
          lfns in lfnT,
          outEntries OUT NOCOPY lfnStatT,
          outerrors out nocopy returnOT)
  return returnO
  AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	parentCheck lfnCheckT;

	/* errors */
	nerrors PLS_INTEGER:=0;

	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);

   /* gets stat based on lfns, based on their type:
         for dirs return lfnstat
         for files surstat:1:2
   */
   SELECT
   (SELECT
--   FCEntryO(
--      SUBSTR(SYS_CONNECT_BY_PATH(NAME,'/'),3),
--      guid,
      lfnstatO(file_size,ctime,mTime,type,null,lfn_lifetime) --later table
--      ,
--      permissionO
--      (
--         Permission_Util.getPrincipalNameStr(principal_id),
--         Permission_Util.getPrincipalNameStr(group_id),
--         userperm,groupperm,otherperm,
--         --permission_util.getacl(guid)
--         CAST(MULTISET
--         	(
--         	SELECT
--            p.NAME,acl.perm
--            FROM acl,PERM_PRINCIPAL p
--            WHERE guid=f.guid AND p.principal_id=acl.principal_id
--          ) AS ACLList)
--      )
--   )
 	FROM
	FC_FILE f
	WHERE LEVEL=getNoOfPathDelimitersRoot(i.COLUMN_VALUE)+1
	AND Permission_Util.isOkAcl(userperm,groupperm,otherperm,
					principal_id,userPrincipal,
					group_id,
			    principals,
					guid, --input guid
					Permission_Util.listFileOp)=1
	START WITH file_id=1
	CONNECT BY PRIOR file_id = parent_id
	AND NAME =getpathtoken(LEVEL-1,i.COLUMN_VALUE)
	)
   BULK COLLECT INTO outEntries
   FROM TABLE(lfns) i;

   FOR i IN 1..outEntries.LAST LOOP
      IF outEntries(i) IS NULL THEN
         nerrors:=nerrors+1;
         outErrors.EXTEND();
         outErrors(nerrors):=
            returnO(lfns(i),i,Permission_Util.accessDeniedErrVal,Permission_Util.accessDeniedErrStr);
      END IF;
   END LOOP;
	RETURN ret;
END;

/* public FCEntry[] getFileCatalogEntry(String[] lfns) */
FUNCTION getFileCatalogEntry(securityToken IN principalToken,
       lfns IN lfnT,
       outEntries OUT NOCOPY FCEntryT,
       outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
  j pls_integer:=0;

	/* errors */


	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
--   outEntries:=FCEntryT();
   debugon();
	 Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);

   /* gets stat based on lfns, based on their type:
         for dirs return lfnstat
         for files surstat:1:2
   */


  SELECT
 CASE TREAT(res AS resolverO).permission
 	WHEN 1 THEN
	 FCEntryO(
      TREAT(res AS resolverO).fname,
      TREAT(res AS resolverO).guid,
      lfnstatO(TREAT(res AS resolverO).file_size,TREAT(res AS resolverO).ctime,TREAT(res AS resolverO).mTime,TREAT(res AS resolverO).ftype,TREAT(res AS resolverO).DATA,TREAT(res AS resolverO).lfn_lifetime), --alter table
      permissionO
      (
	     (SELECT p.NAME FROM PERM_PRINCIPAL p WHERE	principal_id=TREAT(res AS resolverO).principal_id), --35sec
       (SELECT p.NAME FROM PERM_PRINCIPAL p WHERE	principal_id=TREAT(res AS resolverO).group_id),
         TREAT(res AS resolverO).userperm,TREAT(res AS resolverO).groupperm,TREAT(res AS resolverO).otherperm,
         CAST(MULTISET
         	(
         	SELECT
            p.NAME,acl.perm
            FROM acl,PERM_PRINCIPAL p
            WHERE guid=TREAT(res AS resolverO).guid AND p.principal_id=acl.principal_id
          ) AS ACLList)
      ))
  ELSE
      NULL
  END val,
	CASE TREAT(res AS resolverO).permission
      when 1 then null
  	  WHEN 0 THEN
	  returnO(
        TREAT(res AS resolverO).fname,
        TREAT(res AS resolverO).idx,
        Permission_Util.accessDeniedErrVal,  --no right to write
        Permission_Util.accessDeniedErrStr
      )
        WHEN 2 THEN
		returnO(TREAT(res AS resolverO).fname,
        TREAT(res AS resolverO).idx,
        Permission_Util.NoSuchFileVal, --no parent directory exists
        Permission_Util.NoSuchFileStr
      )
      WHEN permission_util.loopInSymlinksDetectedVal THEN   		
   		returnO(TREAT(res AS resolverO).fname,
        TREAT(res AS resolverO).idx,
        Permission_Util.loopInSymlinksDetectedVal, --no parent directory exists
        Permission_Util.loopInSymlinksDetectedStr
      )
      else
      	returnO(TREAT(res AS resolverO).fname,
        TREAT(res AS resolverO).idx,
        Permission_Util.NoSuchFileVal, --no parent directory exists
        Permission_Util.NoSuchFileStr
      )      
   		END rets
 BULK COLLECT INTO	outEntries,outErrors
 FROM (
 SELECT
	permission_util.isOkAclX(--check perms and return error or data as usual but using non-bulk check
	  NVL((SELECT
	   		  CASE WHEN TYPE IN ('f','d')  THEN
          resolverO(idx,parent_id,file_id,
            lfn,guid,type,ctime,mtime,
            file_size,file_checksum,LFN_LIFETIME,status,
            principal_id,group_id,userperm,groupperm,otherperm,
            dangling_ok,
            data,-1)
			  	   WHEN TYPE ='s' THEN
				   	resolverO(idx,fireman.getSymlink(DATA)) --also do permissions
					ELSE
					NULL
				   END
  	   		    FROM
		            FC_FILE f
                WHERE LEVEL=case namelfn when ':' then 1 else 2 end -- always leaf or root if /
		            START WITH parent_id=z.val.parent_id AND file_id=z.val.file_id
		            CONNECT BY PRIOR file_id = parent_id -- check if 'd' only valid paths..
		              AND NAME=namelfn
			 ),resolverO(idx,2,lfn))
			 ,userPrincipal,principals,Permission_Util.listFileOp) RES
  FROM
  (SELECT
		SUBSTR(lfn,1,INSTR(lfn,'/',-1,1)) dir,
		nvl(SUBSTR(lfn,INSTR(lfn,'/',-1,1)+1),':') nameLfn,
		lfn,
		idx
    FROM
    (
      SELECT /*+ cardinality (iz, 50) */
        COLUMN_VALUE AS lfn,ROWNUM AS idx
        FROM
        TABLE(lfns) iz
    )
	) i
  LEFT JOIN
	(
	SELECT (
    SELECT updateFCSimpleO(parent_id,file_id,NULL,dir)
    FROM
      FC_FILE f
      WHERE LEVEL=Fireman.getNoOfPathDelimiters(dir)
      START WITH parent_id=0 AND file_id=1
      CONNECT BY PRIOR file_id = parent_id AND TYPE='d' -- check if 'd' only valid paths..
        AND NAME=Fireman.getpathtoken(LEVEL-1,dir)
        ) val
    FROM
    (
    SELECT /*+ cardinality (dirz, 50) */
      DISTINCT SUBSTR(column_value,1,INSTR(column_value,'/',-1,1)) dir
    FROM
    table(lfns) dirz
    )
	) z
	 ON (i.dir=z.val.item)
	 )
	 ;
   -- remove nulls from the output, make it sparse
   j:=outEntries.count;
    FOR i IN 1..j LOOP
      IF outEntries(i) IS NULL THEN
        outEntries.DELETE(i);
      else
        outErrors.DELETE(i);
      END IF;
    END LOOP;

	RETURN ret;

END;




/* public StringPair[] getLfnForGuid(String[] guids) */
FUNCTION getLfnForGuid(securityToken IN principalToken,
          guids IN lfnT,
          outEntries OUT NOCOPY StringPairT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS

	j PLS_INTEGER;

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;
  mask pls_integer;

	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;

	ret returnO:=returnO(0);
BEGIN
	outerrors:=returnOT();
  DEBUGON();

	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);
  
  mask:=permission_util.getOpMask(Permission_Util.permissionOp);

  select
	case permitted
	WHEN 1 THEN StringPairO(
      s.val,
      (SELECT          
          fireman.reversePath(SYS_CONNECT_BY_PATH(F.NAME,'/'))
         FROM FC_FILE f
         WHERE NAME=':'
         START WITH f.guid = s.val
         CONNECT BY PRIOR parent_id = file_id)
         )
	  ELSE NULL END,
	  case permitted
	  	  WHEN 0 THEN returnO
              (s.val,idx,
               Permission_Util.accessDeniedErrVal,  --no right to write
               Permission_Util.accessDeniedErrStr
              )
        WHEN 2 THEN returnO
              (s.val,s.idx,
               Permission_Util.NoSuchFileVal, --no parent directory exists
               Permission_Util.NoSuchFileStr
              )
   		else
   		null
   		end case
      	BULK COLLECT INTO outEntries,outErrors
   		from (
   SELECT /*+ cardinality( i 5 ) cardinality(inP 10) */
         case
          when ((sign(bitand(userperm,mask))=1 and principal_id=userPrincipal)
            or (sign(bitand(groupperm,mask))=1 and group_id is not null)
            or sign(bitand(otherperm,mask))=1)
            then 1
          when userperm is null then 2
          when userperm is not null then-- check acl
            (SELECT COUNT(*)
          --    into permitted
              FROM dual WHERE
              EXISTS
              (
              SELECT /*+ cardinality (inP2 10) */  NULL
                FROM
                  acl a,
                  TABLE(principals) inP2
                  WHERE
                  (a.guid=guid
                  AND (inP2.item2=a.principal_id OR a.principal_id=userprincipal)
                  AND SIGN(BITAND(perm,mask))=1))
              or exists
                (
                select /*+ cardinality (inP2 5) cardinality (su 5) */ null
                from
                TABLE(permission_util.superUsers) su,
                TABLE(principals) inP2
                    where su.column_value=inP2.item2 or su.column_value=userPrincipal
                )
          )
		else 2
		end as permitted,
        COLUMN_VALUE val,
        idx
    FROM     
     (((SELECT COLUMN_VALUE, ROWNUM idx from  TABLE(guids)) i
        LEFT OUTER JOIN fc s ON s.guid=i.COLUMN_VALUE)
      ) 
      )s
   ;

   -- remove nulls from the output, make it sparse
   j:=outEntries.count;
    FOR i IN 1..j LOOP
      IF outEntries(i) IS NULL THEN
        outEntries.DELETE(i);
      ELSE
        outerrors.DELETE(i);
      END IF;
    END LOOP;

	RETURN ret;
END;


/* public void symlink(StringPair[] symlinks, boolean copyPermFromParent) */
FUNCTION symlink(securityToken IN principalToken,
          entries IN StringPairT,
          copyPermFromParent IN SMALLINT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
  AS

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	permCheck FCCreateT;
	inFc fc_T;
  fc_sRec fc_R;
--  inAcl acl_T;
--  aclMarkers triplesT;

  /* perms */
 defaultPermFetched BOOLEAN:=FALSE;
 defaultPermission PermissionO;
 tmpPermission PermissionSimpleO;


  j PLS_INTEGER:=0;
  idxPos pls_INTEGER;

  timings TIMESTAMP(0);

	/* errors */
	nerrors PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
  errGuid VARCHAR2(36);



	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
   debugon();
   -- check permissions on parent and for existence
   -- do create with no perms or other columns set, fill up data with no check on target
   Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);


  --semi-repetition with as w create
  -- permissions are derived from parent dir and are not mutable - always permitted if readdir is permitted

	SELECT /*+cardinality(ins 10) use_nl(q)*/
	 CASE NVL(d.val.permitted,2)
	   WHEN 1 THEN
	   --ok, can create if file does not exist
	   CASE nvl2(q.parent_id,1,0)
	   	WHEN 0 THEN FCCreateO
      (
			   r,
	   		 d.val.file_id, --parent
			   fc_file_id_seq.NEXTVAL, --file_id
				 d.val.permission
       )
			    --ok, get the data - parent id, new file_id and parental permissions
        ELSE NULL END
	 ELSE NULL END AS val,
     CASE NVL(d.val.permitted,2)  -- fetch permission values checked in th
	   WHEN 1 THEN
	   CASE nvl2(q.parent_id,1,0)   -- check for parents existence
	   	WHEN 1 THEN returnO
               (i.lfn,r,
               Permission_Util.fileOrDirAlreadyExistsVal,
               Permission_Util.fileOrDirAlreadyExistsStr --err, file already there
               )
        ELSE NULL  END
	   WHEN 0 THEN returnO
               (i.lfn,r,
               Permission_Util.accessDeniedErrVal,  --no right to write
               Permission_Util.accessDeniedErrStr
                )
	   WHEN 2 THEN 	returnO(i.lfn,r,
               Permission_Util.InvalidParentVal, --no parent directory exists
               Permission_Util.InvalidParentStr
                )
	   END AS err		--err, parent not there
     BULK COLLECT INTO permCheck,outerrors
	    FROM
      (((SELECT SUBSTR(string1,1,INSTR(string1,'/',-1,1)) dir, string1 lfn, ROWNUM AS r FROM TABLE(entries) ins)
      i LEFT OUTER JOIN (
      SELECT  (SELECT
            FCRStatusExtO
            (
              lfn,
              --permisisons check-
              Permission_Util.isOkAcl(
                      userperm,groupperm,otherperm,
                      principal_id,userPrincipal,
                      group_id,
                      principals,
                      guid, --input guid
                      Permission_Util.writeFileOp),
              file_id,
              parent_id,
              PermissionSimpleO
                (userPrincipal,group_id,userperm,groupperm,otherperm,
                    null
--                   CAST(MULTISET
--                        (
--                          SELECT p.principal_id,acl.perm
--                            FROM acl,PERM_PRINCIPAL p
--                              WHERE guid=f.guid AND p.principal_id=acl.principal_id
--                        ) AS ACLSimpleIdList
--                       )
                )
            )
            FROM
            FC_FILE f
            WHERE LEVEL=Fireman.getNoOfPathDelimiters(lfn)
            START WITH parent_id=0 AND file_id=1 --and name=getpathtoken(1,i.lfn)
            CONNECT BY PRIOR file_id = parent_id AND type='d' -- check if 'd' only valid paths..
              AND NAME=Fireman.getpathtoken(LEVEL-1,lfn)) AS val
              FROM
              (
                SELECT DISTINCT SUBSTR(string1,1,INSTR(string1,'/',-1,1)) lfn
                  FROM TABLE(entries) i
              )
            )
            d ON i.dir=d.val.lfn)
          LEFT OUTER JOIN fc q
            ON parent_id=d.val.file_id AND NAME=SUBSTR(lfn,INSTR(i.lfn,'/',-1,1)+1)
         )
        ;
      -- make it sparse
    j:=permCheck.COUNT;
    FOR i IN 1..j LOOP
      IF permCheck(i) IS NULL THEN
        permCheck.DELETE(i);
      ELSE
        outerrors.DELETE(i);
      END IF;
    END LOOP;

   timings:=SYSTIMESTAMP();
    FOR i IN 1..permCheck.COUNT LOOP
				--ok to go
        idxPos:=permCheck(i).idx;
				fc_sRec.parent_id:=permCheck(i).parent_id;
				fc_sRec.file_id:=permCheck(i).file_id;

				fc_sRec.NAME:= SUBSTR(entries(idxPos).string1,INSTR(entries(idxPos).string1,'/',-1)+1);--get the last part!
				fc_sRec.type:='s'; --symlink

        --symlink stuff
        fc_sRec.data:=entries(idxPos).string2;

				fc_sRec.guid:=generateGuid;
				fc_sRec.cTime:=timings;
				fc_sRec.mTime:=timings;
				fc_sRec.file_size:=0;
				fc_sRec.lfn_lifetime:=0;
				fc_sRec.file_checksum:= ' '; --entries(i).guidstat.checkSum;
				fc_sRec.status:=0;            --entries(i).guidstat.status; -- TODO: proper init ? defaults?
--
--
--        /*
--        take parent permission with acls or provided permission if not empty
--        if there are no permissions provided (always skip ownership),
--          take them from the parent directory
--        if provided use them (except userperm)
--        */
--        -- permisson already fetched with the qry
        tmpPermission:=permCheck(i).permission;

  			-- permissions logic
        /* always take ownership (owner, not group?) from the token */
        fc_sRec.principal_id:=userPrincipal;

        --TODO: or change it to zeroes.
				fc_sRec.group_id:=tmpPermission.group_id;
				fc_sRec.userperm:=tmpPermission.userperm;
				fc_sRec.groupperm:=tmpPermission.groupperm;
				fc_sRec.otherperm:=tmpPermission.otherperm;

				fc_sRec.dangling_ok:=0;

        inFc(i):=fc_sRec;  --add lfn to the story

	END LOOP;

   IF inFc.COUNT=0 THEN
      ret.result:=Permission_Util.missingDataVal;
      ret.errString:=Permission_Util.missingDataStr;
      RETURN ret;
   END IF;
--	--insert into surl,acl,fc (name inserts already during deserialization)
--   ------------ fc insert -----------
   BEGIN
      forall i in 1..inFc.last save exceptions
         insert into fc
            values inFc(i);

   EXCEPTION
      WHEN bulk_errors THEN
         nbulkErr := SQL%BULK_EXCEPTIONS.COUNT;
         outerrors.EXTEND(nbulkErr);
         FOR i IN 1..nbulkErr LOOP
            outerrors(nerrors+i):=
               returnO
                  (
                  entries(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).string1, --entries always preserve order:1
                  SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
                  Permission_Util.generalInternalBulkErrorVal,
                  SUBSTR('Guid/LFN related data. '||
                  SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
                  );
            --delete from ingested surls, acl to be ingested
            errGuid:=inFc(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid;
--            IF aclMarkers.EXISTS(errGuid) THEN
--               inAcl.DELETE(aclMarkers(errGuid).a,aclMarkers(errGuid).b);
--            END IF;
         END LOOP;
         nerrors:=nbulkErr+nerrors;
	END; --exception
  RETURN ret;
  EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
  	RETURN ret;
END;


/* Delete an LFN, symlink or virtual directory. Use rmdir for normal directories.
    public void unlink(String[] lfns)  */
FUNCTION unlink(securityToken IN principalToken,
          lfns IN lfnT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
  AS

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	parentCheck lfnCheckT;

	/* serializer */
	dirCheck RmDirStatusT;
  files bulkPrincipalidT;
  parents bulkPrincipalidT;


	/* errors */
	nerrors PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
  errGuid VARCHAR2(36);
  errSurl rcs.surl%TYPE;


	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();

 DEBUGON();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);

  SELECT /* + cardinality (i 100) */ NVL((
    SELECT /* +use_nl(f) */
            --permisison-
            RmDirStatusO(
            COLUMN_VALUE,
            parent_id,
            file_id,
            Permission_Util.isOkAcl(userperm,groupperm,otherperm,
                principal_id,userPrincipal,
                group_id,
                principals,
                guid,
                Permission_Util.removeFileOp)
              )
    FROM
    FC_FILE f
    WHERE LEVEL=Fireman.getNoOfPathDelimitersRoot(COLUMN_VALUE)+1
    START WITH parent_id=0 AND file_id=1 --and name=getpathtoken(1,i.lfn)
    CONNECT BY PRIOR file_id = parent_id
    AND NAME=Fireman.getpathtoken(LEVEL-1,COLUMN_VALUE)
    ),
    RmDirStatusO(
    COLUMN_VALUE,
    0,0,2)) v
  BULK COLLECT INTO dirCheck
  FROM
  TABLE(lfns) i;

  -- handle erroneus items
  FOR i IN 1..dirCheck.COUNT LOOP
      IF dirCheck(i).permitted=1 THEN
        files(i):=dirCheck(i).file_id;
        parents(i):=dirCheck(i).parent_id;
      ELSIF dirCheck(i).permitted=0 THEN
        files(i):=NULL;
        parents(i):=NULL;
        nerrors:=nerrors+1;
				outerrors.EXTEND();
				outerrors(nerrors):=returnO
               (dirCheck(i).lfn,i,
               Permission_Util.accessDeniedErrVal,
               Permission_Util.accessDeniedErrStr);
      ELSIF dirCheck(i).permitted=2 THEN
        files(i):=NULL;
        parents(i):=NULL;
        nerrors:=nerrors+1;
				outerrors.EXTEND();
				outerrors(nerrors):=returnO
               (dirCheck(i).lfn,i,
               Permission_Util.NoSuchFileVal,
               Permission_Util.NoSuchFileStr);
      END IF;
  END LOOP;

  -- delete the rest, nulls are skipped internally
  -- delete is propagated via trigger
  BEGIN
  FORALL i IN 1..files.COUNT save EXCEPTIONS
    DELETE FROM fc
      WHERE parent_id=parents(i) AND file_id=files(i);
  EXCEPTION
	  /* handle errors thru */
		WHEN bulk_errors THEN
		nerrors := SQL%BULK_EXCEPTIONS.COUNT;
		outerrors.EXTEND(nerrors);
		FOR i IN 1..nerrors LOOP
			outerrors(i):=
				returnO
					(
					dirCheck(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).lfn,
					SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
					Permission_Util.generalInternalBulkErrorVal,
					SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||i ,1,1024)
					);
		END LOOP;
	END;		--exception

	RETURN ret;
END unlink;


/* public void updateModifyTime(String[] lfns, long time) */
FUNCTION updateModifyTime(securityToken IN principalToken,
       lfns IN lfnT,
       inTime IN PLS_INTEGER,
       outerrors OUT NOCOPY returnOT)
  RETURN returnO
  AS

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
  parentItem listNum;
	fileItem listNum;

  timeVal fc.CTIME%type;

	/* errors */
	nerrors PLS_INTEGER:=0;
  j PLS_INTEGER;

  nerrorsBp PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
  errGuid VARCHAR2(36);

	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
   -- check permissions on the items and update mtime
   Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);

  -- check and get handles, then update and report errors
   SELECT
 CASE treat(res as resolverO).permission
  WHEN 1  THEN treat(res as resolverO).parent_id
 ELSE NULL END,
 CASE treat(res as resolverO).permission
  WHEN 1  THEN treat(res as resolverO).file_id
 ELSE NULL END,
 treat(res as resolverO).getReturn() ret
  bulk collect into parentItem,fileItem,outErrors
 FROM (
 SELECT
	permission_util.isOkAclX(--check perms and return error or data as usual but using non-bulk check
	  NVL((SELECT
	   		  CASE WHEN TYPE IN ( 'f', 'd')  THEN
         resolverO(idx,parent_id,file_id,
            lfn,guid,type,ctime,mtime,
            file_size,file_checksum,LFN_LIFETIME,status,
            principal_id,group_id,userperm,groupperm,otherperm,
            dangling_ok,
            data,-1)
			  	   WHEN TYPE ='s' THEN
				   	resolverO(idx,fireman.getSymlink(DATA)) --also do permissions
					ELSE
					NULL
				   END
  	   		    FROM
		            FC_FILE f
                WHERE LEVEL=case namelfn when ':' then 1 else 2 end -- always leaf or root if /
		            START WITH parent_id=z.val.parent_id AND file_id=z.val.file_id
		            CONNECT BY PRIOR file_id = parent_id -- check if 'd' only valid paths..
		              AND NAME=namelfn
			 ),resolverO(idx,2,lfn))
			 ,userPrincipal,principals,Permission_Util.setmetaOp) RES
  FROM
  (SELECT
		SUBSTR(lfn,1,INSTR(lfn,'/',-1,1)) dir,
		nvl(SUBSTR(lfn,INSTR(lfn,'/',-1,1)+1),':') nameLfn,
		lfn,
		idx
    FROM
    (
      SELECT /*+ cardinality (iz, 50) */
        COLUMN_VALUE AS lfn, ROWNUM AS idx
        FROM
        TABLE(lfns) iz
    )
	) i
  LEFT JOIN
	(
	SELECT (
    SELECT updateFCSimpleO(parent_id,file_id,NULL,dir)
    FROM
      FC_FILE f
      WHERE LEVEL=Fireman.getNoOfPathDelimiters(dir)
      START WITH parent_id=0 AND file_id=1
      CONNECT BY PRIOR file_id = parent_id AND TYPE='d' -- check if 'd' only valid paths..
        AND NAME=Fireman.getpathtoken(LEVEL-1,dir)
        ) val
    FROM
    (
    SELECT /*+ cardinality (dirz, 50) */
      DISTINCT SUBSTR(column_value,1,INSTR(column_value,'/',-1,1)) dir
    FROM
    table(lfns)dirz
    )
	) z
	 ON (i.dir=z.val.item)
	 )
	 ;
   -- remove nulls from the output, make it sparse
   j:=parentItem.count;
    FOR i IN 1..j LOOP
      IF parentItem(i) IS NULL THEN
        parentItem.DELETE(i);
        fileItem.DELETE(i);
        nerrors:=nerrors+1;
      ELSE
        outerrors.DELETE(i);
      END IF;
    END LOOP;

   nerrorsBp:=nerrorsBp+nerrors;
   timeVal := to_timestamp('1970-01-01', 'YYYY-MM-DD TZD')+(inTime/86400000) ;

 	BEGIN
		FORALL i IN indices of parentItem save EXCEPTIONS
			update fc
        set mtime=timeVal
            WHERE
            parent_id=parentItem(i) and file_id=fileItem(i)
            ;
	EXCEPTION
   --FIXME: chnage indexing to take it from the input not from the forall input
		WHEN bulk_errors THEN
		nerrors := SQL%BULK_EXCEPTIONS.COUNT;
		outerrors.EXTEND(nerrors);
		FOR i IN 1..nerrors LOOP
			outerrors(i+nerrorsBp):=
				returnO
					(
					lfns(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX),
					SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
					Permission_Util.generalInternalBulkErrorVal,
					SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| 'at index '||i ,1,1024)
					);
		END LOOP;
		nerrorsBp:=nerrorsBp+nerrors;
    END;

	RETURN ret;
END;


/* public void updateValidityTime(String[] lfns, long time) */
FUNCTION updateValidityTime(securityToken IN principalToken,
       lfns IN lfnT,
       intime IN PLS_INTEGER,
       outerrors OUT NOCOPY returnOT)
  RETURN returnO
  AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
  parentItem listNum;
	fileItem listNum;

  timeVal fc.CTIME%type;

	/* errors */
	nerrors PLS_INTEGER:=0;
  j PLS_INTEGER;

  nerrorsBp PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
  errGuid VARCHAR2(36);

	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
   -- check permissions on the items and update mtime
   Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);

  -- check and get handles, then update and report errors
   SELECT
 CASE treat(res as resolverO).permission
  WHEN 1  THEN treat(res as resolverO).parent_id
 ELSE NULL END,
 CASE treat(res as resolverO).permission
  WHEN 1  THEN treat(res as resolverO).file_id
 ELSE NULL END,
 treat(res as resolverO).getReturn() ret
  bulk collect into parentItem,fileItem,outErrors
 FROM (
 SELECT
	permission_util.isOkAclX(--check perms and return error or data as usual but using non-bulk check
	  NVL((SELECT
	   		  CASE WHEN TYPE IN ( 'f', 'd')  THEN
         resolverO(idx,parent_id,file_id,
            lfn,guid,type,ctime,mtime,
            file_size,file_checksum,LFN_LIFETIME,status,
            principal_id,group_id,userperm,groupperm,otherperm,
            dangling_ok,
            data,-1)
			  	   WHEN TYPE ='s' THEN
				   	resolverO(idx,fireman.getSymlink(DATA)) --also do permissions
					ELSE
					NULL
				   END
  	   		    FROM
		            FC_FILE f
                WHERE LEVEL=case namelfn when ':' then 1 else 2 end -- always leaf or root if /
		            START WITH parent_id=z.val.parent_id AND file_id=z.val.file_id
		            CONNECT BY PRIOR file_id = parent_id -- check if 'd' only valid paths..
		              AND NAME=namelfn
			 ),resolverO(idx,2,lfn))
			 ,userPrincipal,principals,Permission_Util.removeFileOp) RES
  FROM
  (SELECT
		SUBSTR(lfn,1,INSTR(lfn,'/',-1,1)) dir,
		nvl(SUBSTR(lfn,INSTR(lfn,'/',-1,1)+1),':') nameLfn,
		lfn,
		idx
    FROM
    (
      SELECT /*+ cardinality (iz, 50) */
        COLUMN_VALUE AS lfn, ROWNUM AS idx
        FROM
        TABLE(lfns) iz
    )
	) i
  LEFT JOIN
	(
	SELECT (
    SELECT updateFCSimpleO(parent_id,file_id,NULL,dir)
    FROM
      FC_FILE f
      WHERE LEVEL=Fireman.getNoOfPathDelimiters(dir)
      START WITH parent_id=0 AND file_id=1
      CONNECT BY PRIOR file_id = parent_id AND TYPE='d' -- check if 'd' only valid paths..
        AND NAME=Fireman.getpathtoken(LEVEL-1,dir)
        ) val
    FROM
    (
    SELECT /*+ cardinality (dirz, 50) */
      DISTINCT SUBSTR(column_value,1,INSTR(column_value,'/',-1,1)) dir
    FROM
    table(lfns)dirz
    )
	) z
	 ON (i.dir=z.val.item)
	 )
	 ;
   -- remove nulls from the output, make it sparse
   j:=parentItem.count;
    FOR i IN 1..j LOOP
      IF parentItem(i) IS NULL THEN
        parentItem.DELETE(i);
        fileItem.DELETE(i);
        nerrors:=nerrors+1;
      ELSE
        outerrors.DELETE(i);
      END IF;
    END LOOP;

   nerrorsBp:=nerrorsBp+nerrors;
   --TODO: we may change to guard validity for the past
   timeVal := to_timestamp('1970-01-01', 'YYYY-MM-DD')+(inTime/86400000) ;

 	BEGIN
		FORALL i IN indices of parentItem save EXCEPTIONS
			update fc
        set mtime=timeVal
            WHERE
            parent_id=parentItem(i) and file_id=fileItem(i)
            ;
	EXCEPTION
		WHEN bulk_errors THEN
		nerrors := SQL%BULK_EXCEPTIONS.COUNT;
		outerrors.EXTEND(nerrors);
		FOR i IN 1..nerrors LOOP
			outerrors(i+nerrorsBp):=
				returnO
					(
					lfns(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX),
					SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
					Permission_Util.generalInternalBulkErrorVal,
					SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| 'at index '||i ,1,1024)
					);
		END LOOP;
		nerrorsBp:=nerrorsBp+nerrors;
    END;


	RETURN ret;
END;

/*  Checks if the current user has the required permission bits on the specified items. */


FUNCTION checkPermission(securityToken IN principalToken,
          entries IN StringBooleanT,
          perm IN SMALLINT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
  AS
  /* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */


	/* errors */
	nerrors PLS_INTEGER:=0;

	/* return */
	ret returnO:=returnO(0);
BEGIN
  outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);
  DEBUGON();
  -- TODO: replace input filter with pipe func:4

  SELECT val
  BULK COLLECT INTO outerrors
  FROM
  (
   --check guid permissions
    SELECT /*+  cardinality( i 100 ) cardinality (inP 10) use_nl(inP) */
        case
          Permission_Util.isOkAclMask(userperm,groupperm,otherperm,
                 principal_id,userPrincipal,
                 group_id,
                 principals,
                 i.string, --input guid
                 perm)
        when 1 then NULL
        when 0 then returnO(i.string,i.idx,Permission_Util.accessDeniedErrVal,Permission_Util.accessDeniedErrStr,0)
        when 2 then returnO(i.string,i.idx,Permission_Util.NoSuchFileVal,Permission_Util.NoSuchFileStr,0)
        WHEN permission_util.loopInSymlinksDetectedVal THEN   		
          returnO(i.string,i.idx,Permission_Util.loopInSymlinksDetectedVal,Permission_Util.loopInSymlinksDetectedStr)
        else null end as val
      FROM
      (
      select /*+cardinality (i 20) */ boolval,string,idx from
      TABLE(entries)
      ) i
      LEFT OUTER JOIN fc bp ON (i.string=bp.guid) WHERE boolval=1
      UNION ALL
      --lfn part
--      SELECT
  --       (TREAT(res AS resolverO)).getReturn()
    --     FROM (
         SELECT
          TREAT(permission_util.isOkAclXMask(--check perms and return error or data as usual but using non-bulk check
            NVL((SELECT
                  CASE WHEN TYPE IN ( 'f', 'd')  THEN
                 resolverO(idx,parent_id,file_id,
                    lfn,guid,type,ctime,mtime,
                    file_size,file_checksum,LFN_LIFETIME,status,
                    principal_id,group_id,userperm,groupperm,otherperm,
                    dangling_ok,
                    data,-1)
                     WHEN TYPE ='s' THEN
                    resolverO(idx,fireman.getSymlink(DATA)) --also do permissions
                  ELSE
                  NULL
                   END
                      FROM
                        FC_FILE f
                        WHERE LEVEL=case namelfn when ':' then 1 else 2 end -- always leaf or root if /
                        START WITH parent_id=z.val.parent_id AND file_id=z.val.file_id
                        CONNECT BY PRIOR file_id = parent_id -- check if 'd' only valid paths..
                          AND NAME=namelfn
               ),resolverO(idx,2,lfn))
               ,userPrincipal,principals,perm) as resolverO).getReturn() RES
          FROM
          (SELECT
            SUBSTR(lfn,1,INSTR(lfn,'/',-1,1)) dir,
            nvl(SUBSTR(lfn,INSTR(lfn,'/',-1,1)+1),':') nameLfn,
            lfn,
            idx
            FROM
            (
              SELECT /*+ cardinality (iz, 50) */
                string AS lfn,boolval, ROWNUM AS idx
                FROM
                TABLE(entries) iz
            ) where boolval=0
          ) i
          LEFT JOIN
          (
          SELECT (
            SELECT updateFCSimpleO(parent_id,file_id,NULL,dir)
            FROM
              FC_FILE f
              WHERE LEVEL=Fireman.getNoOfPathDelimiters(dir)
              START WITH parent_id=0 AND file_id=1
              CONNECT BY PRIOR file_id = parent_id AND TYPE='d' -- check if 'd' only valid paths..
                AND NAME=Fireman.getpathtoken(LEVEL-1,dir)
                ) val
            FROM
            (
            SELECT /*+ cardinality (dirz, 50) */
              DISTINCT SUBSTR(string,1,INSTR(string,'/',-1,1)) dir,boolval
            FROM
            table(entries)dirz
            ) where boolval=0
          ) z
           ON (i.dir=z.val.item)
--        )
      )
      WHERE val IS NOT NULL
        ;

	RETURN ret;
END;







/*  + public void mv(StringPair[] entry) throws CatalogException, AuthorizationException, ExistsException, NotExistsException;*/
FUNCTION mv(securityToken IN principalToken,
          entries IN StringPairT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	parentCheck lfnCheckT;


	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
   Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);



























	RETURN ret;
END;

/**
  TODO: change documentation, this fucntion never throws exception
        if kids are inaccessible.
        LIMIT impl
*/
/*+ public String[] locate(String startDir, String pattern) throws CatalogException, AuthorizationException, NotExistsException;*/
FUNCTION locate(securityToken IN principalToken,
      startDir IN VARCHAR2,
			pattern IN VARCHAR2,
      LIMIT IN SMALLINT,
			outEntries OUT NOCOPY lfnT,
      outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	parentCheck FC_FILE.file_id%TYPE;
  parentCheckParent FC_FILE.file_id%TYPE;
  parentCheckVal PLS_INTEGER;

  /* path ops */
  newPath pls_integer;

	/* return */
	ret returnO:=returnO(0);
BEGIN
  outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);
  debugon();

  -- decode startup dir and start traversal with given pattern until reaching the limit
  BEGIN

	 SELECT
    TREAT (res AS resolverO).file_id,
    TREAT (res AS resolverO).parent_id,
    TREAT (res AS resolverO).permission
   COLLECT INTO parentCheck, parentCheckParent, parentCheckVal
	  FROM(
  SELECT
	permission_util.isOkAclX(--check perms and return error or data as usual but using non-bulk check
	  NVL(CASE
            WHEN TYPE IN ( 'f', 'd')  THEN
              resolverO(parent_id,file_id,NAME,guid,principal_id,group_id,userperm,groupperm,otherperm) --also do permissions
			  	   WHEN TYPE ='s' THEN
              fireman.getSymlink(DATA) --also do permissions
					ELSE
					NULL
				END
			 ,resolverO(0,2))
			 ,userPrincipal,principals,Permission_Util.readDirOp) res
       FROM
      FC_FILE f
     WHERE LEVEL=fireman.getNoOfPathDelimitersRoot(startDir)+1
      START WITH parent_id=0 AND file_id=1
      CONNECT BY PRIOR file_id = parent_id -- check if 'd' only valid paths..
		AND NAME=fireman.getpathtoken(LEVEL-1,startDir) and type='d'
		);

  EXCEPTION
    WHEN NO_DATA_FOUND THEN
       ret.result:=Permission_Util.NoSuchFileVal;
       ret.errString:=Permission_Util.NoSuchFileStr;
    RETURN ret;
  END;
  IF parentCheckVal=0 THEN
     ret.result:=Permission_Util.accessDeniedErrVal;
     ret.errString:=Permission_Util.accessDeniedErrStr;
     RETURN ret;
  elsif parentCheck=permission_util.loopInSymlinksDetectedVal then
      ret.result:=Permission_Util.loopInSymlinksDetectedVal;
      ret.errString:=Permission_Util.loopInSymlinksDetectedStr;
   RETURN ret;     
  END IF;

  IF startDir='/' THEN newPath:=3; ELSE newPath:=0; END IF;


  select n
  BULK COLLECT INTO outEntries
  from
  (
  SELECT rownum i,substr(SYS_CONNECT_BY_PATH(NAME,'/'),newPath) n
  FROM fc f
  where NAME LIKE pattern ESCAPE '\'
  START WITH file_id=parentCheck and parent_id=parentCheckParent
  CONNECT BY PRIOR file_id = parent_id
  AND (Permission_Util.isOkAcl(userperm,groupperm,otherperm,
                principal_id,userPrincipal,
                group_id,
                principals,
                guid, --input guid
                Permission_Util.readFileOp)=1)
  )
  where i<=LIMIT
  ;



	RETURN ret;
END;

/*+ public StringPair[] getGuidForLfn(String[] lfn) throws CatalogException, AuthorizationException, NotExistsException;*/
FUNCTION getGuidForLfn(securityToken IN principalToken,
          lfns IN lfnT,
          outEntries OUT NOCOPY StringPairT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
  j pls_integer;

	/* return */
	ret returnO:=returnO(0);
BEGIN

	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);
   DEBUGON();

SELECT
  (TREAT(res AS resolverO)).getStringPairO() val,
  (TREAT(res AS resolverO)).getReturn() ret
  bulk collect into outEntries,outErrors
 FROM (
 SELECT
	permission_util.isOkAclX(--check perms and return error or data as usual but using non-bulk check
	  NVL((SELECT
	   		  CASE WHEN TYPE IN ( 'f', 'd')  THEN
         resolverO(idx,parent_id,file_id,
            lfn,guid,type,ctime,mtime,
            file_size,file_checksum,LFN_LIFETIME,status,
            principal_id,group_id,userperm,groupperm,otherperm,
            dangling_ok,
            data,-1)
			  	   WHEN TYPE ='s' THEN
				   	resolverO(idx,fireman.getSymlink(DATA)) --also do permissions
					ELSE
					NULL
				   END
  	   		    FROM
		            FC_FILE f
                WHERE LEVEL=case namelfn when ':' then 1 else 2 end -- always leaf or root if /
		            START WITH parent_id=z.val.parent_id AND file_id=z.val.file_id
		            CONNECT BY PRIOR file_id = parent_id -- check if 'd' only valid paths..
		              AND NAME=namelfn
			 ),resolverO(idx,2,lfn))
			 ,userPrincipal,principals,Permission_Util.listDirOp) RES
  FROM
  (SELECT
		SUBSTR(lfn,1,INSTR(lfn,'/',-1,1)) dir,
		nvl(SUBSTR(lfn,INSTR(lfn,'/',-1,1)+1),':') nameLfn,
		lfn,
		idx
    FROM
    (
      SELECT /*+ cardinality (iz, 50) */
        COLUMN_VALUE AS lfn,ROWNUM AS idx
        FROM
        TABLE(lfns) iz
    )
	) i
  LEFT JOIN
	(
	SELECT (
    SELECT updateFCSimpleO(parent_id,file_id,NULL,dir)
    FROM
      FC_FILE f
      WHERE LEVEL=Fireman.getNoOfPathDelimiters(dir)
      START WITH parent_id=0 AND file_id=1
      CONNECT BY PRIOR file_id = parent_id AND TYPE='d' -- check if 'd' only valid paths..
        AND NAME=Fireman.getpathtoken(LEVEL-1,dir)
        ) val
    FROM
    (
    SELECT /*+ cardinality (dirz, 50) */
      DISTINCT SUBSTR(column_value,1,INSTR(column_value,'/',-1,1)) dir
    FROM
    table(lfns)dirz
    )
	) z
	 ON (i.dir=z.val.item)
	 )
	 ;
   -- remove nulls from the output, make it sparse
   j:=outEntries.count;
    FOR i IN 1..j LOOP
      IF outEntries(i) IS NULL THEN
        outEntries.DELETE(i);
      ELSE
        outerrors.DELETE(i);
      END IF;
    END LOOP;

	RETURN ret;
END;

-- TODO: remove as obsolete
/*+ public StringPair[] getLfn(String[] guid) throws CatalogException, AuthorizationException, NotExistsException;*/
FUNCTION getLfn(securityToken IN principalToken,
          guids IN guidSimpleT,
			 outEntries OUT NOCOPY StringPairT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	parentCheck lfnCheckT;


	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);

	RETURN ret;
END;









/***************************************************************
SAme as fcrcreate with default dir as parent
****************************************************************/
FUNCTION createGuid(
	securityToken IN principalToken,
	entries IN RCEntryT,
	outerrors OUT NOCOPY returnOT
	)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;
   permDefault INTEGER;

   /* defaults */
   def DefaultsO;

	/* serializer */
	permCheck FCRStatusTT;
	inFc fc_T;
	inAcl acl_T;
  inSurl surl_T;

  rc_sRec 	rc_s;
  fc_sRec fc_R;


  guidMarkerSurlPerms triplesT;

  aclMarkers triplesT;
  aclMarker tripleR;
	k PLS_INTEGER:=0;
  j PLS_INTEGER:=0;
  sCounter PLS_INTEGER:=1;
  aclCounter PLS_INTEGER:=0;

   masterSurlGuard PLS_INTEGER:=0;

   /* consts*/
   timing TIMESTAMP(0);

	/* errors */
	nerrors PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
   errGuid VARCHAR2(36);
   errSurl rcs.surl%TYPE;
   nBadItems PLS_INTEGER:=0;


	/* return */
	ret returnO:=returnO(0);
BEGIN
	outerrors:=returnOT();

   DEBUGON();

   Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);
   -- get default parent, no perms check otherwise
   ret:=getDefaultSettings(securityToken,def);
   if ret.result<>0 then
    return ret;
   end if;
    SELECT
    Permission_Util.isOkAcl(userperm,groupperm,otherperm,
					principal_id,userPrincipal,
					group_id,
			    principals,
					guid, --input guid
					Permission_Util.getMetaOp)
      INTO permDefault
	FROM FC_FILE f
	WHERE file_id=def.defaultParent
	;
	IF permDefault <> 1 THEN
           ret.result:=Permission_Util.fileOrDirAlreadyExistsVal;
           ret.errstring:=Permission_Util.fileOrDirAlreadyExistsStr;
           RETURN ret;
   END IF;
   timing:=SYSTIMESTAMP;
	FOR i IN 1..entries.COUNT LOOP

				--ok to go
				fc_sRec.parent_id:=def.defaultParent;
				SELECT fc_file_id_seq.NEXTVAL INTO fc_sRec.file_id FROM dual;

				fc_sRec.NAME:= entries(i).guid;--substr(entries(i).lfn,instr(entries(i).lfn,'/',-1)+1);--get the last part!
				fc_sRec.type:='f'; --either d or f (for now)
				fc_sRec.guid:=entries(i).guid;
				fc_sRec.cTime:=NVL(entries(i).guidstat.creationTime,timing);
        fc_sRec.mTime:=NVL(entries(i).guidstat.modifyTime,timing);
				fc_sRec.file_checksum:=entries(i).guidstat.checkSum;
        fc_sRec.file_size:=entries(i).guidstat.fileSize;
        fc_sRec.lfn_lifetime:=0;
				fc_sRec.status:=entries(i).guidstat.status;



				fc_sRec.principal_id:=userPrincipal;
				fc_sRec.group_id:=Permission_Util.getPrincipalIdIUW(entries(i).permission.groupname,null,1); -- TODO: is it always provided one:5 creation of the users...

				fc_sRec.userperm:=entries(i).permission.userperm;
				fc_sRec.groupperm:=entries(i).permission.groupperm;
				fc_sRec.otherperm:=entries(i).permission.otherperm;
				fc_sRec.dangling_ok:=0;

            /*surls*/
        		guidMarkerSurlPerms(entries(i).guid).a:=sCounter;
            guidMarkerSurlPerms(entries(i).guid).b:=sCounter+entries(i).surlstats.COUNT;
            guidMarkerSurlPerms(entries(i).guid).c:=i;
            FOR j IN 1..entries(i).surlstats.COUNT LOOP --must have at least one element
               rc_sRec.guid:=entries(i).guid;
               rc_sRec.surl:=entries(i).surlstats(j).surl;
               rc_sRec.mTime:=NVL(entries(i).surlstats(j).modifyTime,timing);
               rc_sRec.cTime:=NVL(entries(i).surlstats(j).creationTime,timing);
               IF entries(i).surlstats(j).master_replica=1 AND masterSurlGuard=0 THEN
                  rc_sRec.master_replica:=1;
                  masterSurlGuard:=1;
               ELSE
                  rc_sRec.master_replica:=0;
               END IF;
               rc_sRec.se:=NVL(entries(i).surlstats(j).se,'dummy');
               rc_sRec.surl_lifetime:=0;	--not used yet
               inSurl(sCounter):=rc_sRec;
               sCounter:=sCounter+1;
            END LOOP;
            IF masterSurlGuard=0 THEN
         -- raise error no master_surl has been specified
            nerrors:=nerrors+1;
            outerrors.EXTEND;
            outerrors(nerrors):=returnO
                     (
                     entries(i).guid,
                     i,
                     Permission_Util.noMasterReplicaGivenVal,
                     Permission_Util.noMasterReplicaGivenStr
                     );
            inSurl.DELETE(
            guidMarkerSurlPerms(fc_sRec.guid).a,
            guidMarkerSurlPerms(fc_sRec.guid).b
            );
            ELSE
            inFc(i):=fc_sRec;  --add lfn to the story
            -----------------------
            -- handle acl's -------
            BEGIN --acls
        		IF  entries(i).permission.acl IS NOT NULL AND entries(i).permission.acl.COUNT>0 THEN
             aclMarker.a:=aclCounter+1;
             aclMarker.b:=aclCounter+entries(i).permission.acl.COUNT;
             aclMarker.c:=i;
             aclMarkers(fc_sRec.guid):=aclMarker;

             FOR ac IN 1..entries(i).permission.acl.COUNT LOOP
                aclCounter:=aclCounter+1;
                inAcl(aclCounter).guid:=fc_sRec.guid;
                inAcl(aclCounter).principal_id:=Permission_Util.getPrincipalIdIUW(entries(i).permission.acl(ac).principal,null,1); --change it to one call per all perms
                inAcl(aclCounter).perm:=entries(i).permission.acl(ac).perm;
            END LOOP;
            END IF;
            END; --acls
            END IF;
           masterSurlGuard:=0;
	END LOOP;

   IF inFc.COUNT=0 OR inSurl.COUNT=0 THEN
      ret.result:=Permission_Util.missingDataVal;
      ret.errString:=Permission_Util.missingDataStr;
      RETURN ret;
   END IF;
	--insert into surl,acl,fc (name inserts already during deserialization)
   ------------ fc insert -----------
   BEGIN
      FORALL i IN 1..inFc.LAST save EXCEPTIONS
         INSERT INTO fc
            VALUES inFc(i);
   EXCEPTION
      WHEN bulk_errors THEN
         nbulkErr := SQL%BULK_EXCEPTIONS.COUNT;
         outerrors.EXTEND(nbulkErr);
         FOR i IN 1..nbulkErr LOOP
            outerrors(nerrors+i):=
               returnO
                  (
                  entries(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid, --entries always preserve order:1
                  SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
                  Permission_Util.generalInternalBulkErrorVal,
                  SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
                  );
            --delete from ingested surls, acl to be ingested
            errGuid:=inFc(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid;
            inSurl.DELETE(guidMarkerSurlPerms(errGuid).a,guidMarkerSurlPerms(errGuid).b);
            IF aclMarkers.EXISTS(errGuid) THEN
               inAcl.DELETE(aclMarkers(errGuid).a,aclMarkers(errGuid).b);
            END IF;
         END LOOP;
         nerrors:=nbulkErr+nerrors;
	END; --exception
   ------------ surl insert ---------
   BEGIN
   IF inSurl.COUNT=0 THEN
      ret.result:=Permission_Util.missingDataVal;
      ret.errString:=Permission_Util.missingDataStr;
      RETURN ret;
   END IF;
    FORALL i IN 1..inSurl.LAST save EXCEPTIONS
         INSERT INTO rcs
            VALUES inSurl(i);

   EXCEPTION
      WHEN bulk_errors THEN
         nbulkErr := SQL%BULK_EXCEPTIONS.COUNT;
         outerrors.EXTEND(nbulkErr);
         FOR i IN 1..nbulkErr LOOP
            errSurl:=inSurl(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).surl;
            errGuid:=inSurl(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid;
            outerrors(nerrors+i):=
               returnO
                  (
                  errSurl,
                  guidMarkerSurlPerms(errGuid).c,
                  Permission_Util.generalInternalBulkErrorVal,
                  SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
                  );
            --in the future we may have optional success... then you'd:
--            guidMarkerSurlPerms(errGuid).c:=guidMarkerSurlPerms(errGuid).c+1;
--            if guidMarkerSurlPerms(errGuid).c >= guidMarkerSurlPerms(errGuid).b-
--                  guidMarkerSurlPerms(errGuid).a
--               then
--               delete from fc where guid = errGuid;
--            end if;
            --but today we just:
            DELETE FROM fc WHERE guid = errGuid;
            --acl is always optional so need for back-delete
            IF inAcl.COUNT>0 AND aclMarkers.EXISTS(inSurl(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid) THEN
            inAcl.DELETE(
               aclMarkers(inSurl(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid).a,
               aclMarkers(inSurl(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid).b
               );
            END IF;
            --delete from input surl,acl
            inSurl.DELETE(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX);

         END LOOP;
         nerrors:=nbulkErr+nerrors;
	END; --exception

   ------------ acl insert ---------
   BEGIN
      FORALL i IN 1..inACl.LAST save EXCEPTIONS
         INSERT INTO acl
            VALUES inAcl(i);
   EXCEPTION
      WHEN bulk_errors THEN
         nbulkErr := SQL%BULK_EXCEPTIONS.COUNT;
         outerrors.EXTEND(nbulkErr);
         FOR i IN 1..nbulkErr LOOP
            errGuid:=inAcl(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid;
            outerrors(nerrors+i):=
               returnO
                  (
                  entries(aclMarkers(errGuid).c).guid,
                  aclMarkers(errGuid).c,
                  Permission_Util.generalInternalBulkErrorVal,
                  SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
                  );
            -- in the future we may still keep such entry as valid (and not back-delete)
            -- nowadays:
            DELETE FROM fc WHERE guid=errGuid;
            DELETE FROM rcs WHERE guid=errGuid;
         END LOOP;
         nerrors:=nbulkErr+nerrors;
	END; --exception

   -- do escalation error check to set general resul as error o warnign eventually.
   RETURN ret;
END;


/***************************************************************
	Adds new surl(replica) mapping for an existing guid (only one)
	to the set of existing ones, adds new entry.


	Error	If guid does not exist
				is entry already there
****************************************************************/

FUNCTION addGuidReplica(
	securityToken IN principalToken,
	inguid IN VARCHAR2,
	newSurls IN SURLEntryT,
   outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	DATA dataAndStatusOT:=dataAndStatusOT();

	/*serialisation*/
	guidIn guidOT:=guidOT(NEW guidO(inguid));
	rc_gRecs rc_gT;
	rc_gRec 	rc_g;
	rc_sRecs rc_sT;
	rc_sRec 	rc_s;

   masterSurls ssurlT;

	guidMarkerPerms pairsT;
	guidMarkerSurlPerms pairsT;

	j PLS_INTEGER:=0;
	k PLS_INTEGER:=0;
	masterSurlGuard PLS_INTEGER:=0;

	perms PermissionEntryT;
	changeTime TIMESTAMP(0);

	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
	starterr PLS_INTEGER:=1;

	/* return */
	ret returnO:=returnO(0);
BEGIN
  DEBUGON();
	changeTime:=SYSTIMESTAMP;
	outerrors:=returnOT();
   /*   this user must be there already - passive check*/
   Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);
	/* serialize and
	   check if guid has permission/exist.
		Currenty permisison_util.getGuidPermission does not
		return reason code - non-exiting/not allowed.*/
		DATA.EXTEND();
    DATA(1):=dataAndStatusO(inguid,0);
		checkPrincipalPermissions(
				userPrincipal,
            principals,
            guidOT(guidO(inguid)),
				DATA,
				Permission_Util.setmetaOp);

	IF DATA(1).permitted=1 THEN	--always only one from perms
	FOR y IN 1..newSurls.COUNT LOOP
		rc_sRec.guid:=inguid;
		rc_sRec.surl:=newSurls(y).surl;
		--change!!
      IF newSurls(y).master_replica=1 THEN
         masterSurls(y):=rc_sRec.surl;
      END IF;
		--rc_sRec.master_replica:=newSurls(y).master_replica; --check if another is master now, remember item
		--change!!
      --reset mrep to avoid null exception
      rc_sRec.master_replica:=0;
		rc_sRec.surl_lifetime:=0;
		rc_sRec.se:=NVL(newSurls(y).se,'dummy');--TODO:
		rc_sRec.cTime:=changeTime;
		rc_sRec.mTime:=changeTime;
		rc_sRecs(y):=rc_sRec;
	END LOOP;
		/* update rcg, insert into rcs */

	UPDATE fc
			SET mTime=changeTime
				WHERE guid=inguid
				;
	BEGIN
   IF rc_sRecs.COUNT>0 THEN
		FORALL i IN 1..rc_sRecs.LAST save EXCEPTIONS
			INSERT INTO rcs
				VALUES rc_sRecs(i);
   END IF;
	EXCEPTION
	  /* handle errors thru */
		WHEN bulk_errors THEN
		nerrors := SQL%BULK_EXCEPTIONS.COUNT;
		outerrors.EXTEND(nerrors);
		FOR i IN 1..nerrors LOOP
			outerrors(i):=
				returnO
					(
					rc_sRecs(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).surl,
					SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
					Permission_Util.generalInternalBulkErrorVal,
					SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||i ,1,1024)
					);
               rc_sRecs.DELETE(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX);
		END LOOP;
	END;		--exception
	ELSE --if not permitted
		--some permission errors occured
		ret.result:=Permission_Util.accessDeniedErrVal;
		ret.errString:=Permission_Util.accessDeniedErrStr;
	END IF;
   IF rc_sRecs.COUNT<=0 OR masterSurls.COUNT=0 THEN RETURN ret; END IF;
   FORALL i IN 1..rc_sRecs.COUNT save EXCEPTIONS
		UPDATE rcs a
			SET master_replica=(
				SELECT
				nvl2(NULLIF(a.surl,masterSurls(i)),0,1)
					FROM rcs s
				WHERE a.guid=s.guid AND a.surl=s.surl
				)
			WHERE
				guid=inguid
				;


  RETURN ret;
EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
		RETURN ret;

END;

/***************************************************************
	For existing guid change Guid stat arguments only, which are
	checksum,status, modifytime, size, creationtime
	/param guid guid to change
	/param stat stats for change

	Error if no such guid or surl exist.
****************************************************************/
FUNCTION updateGuidStat(securityToken IN principalToken,
          guids IN guidOT,
          stats IN GuidStatT,
          outerrors OUT NOCOPY returnOT)
  RETURN returnO
AS
	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
	starterr PLS_INTEGER:=1;

	/*serialisation*/
	rc_gRecs rc_gT;
	rc_gRec 	rc_g;
	rc_sRecs rc_sT;
	rc_sRec 	rc_s;
	rc_sRecsFinal rc_sT;
	rc_sRecFinal 	rc_s;

	guidsRecs gGuidsT;
	file_checksumRecs file_checksumT;
	statusRecs  statusT;
	mTimeRecs  mTimeT;
	file_sizeRecs  file_sizeT;


	k PLS_INTEGER;
	guidMarkerPerms pairsT;

   /* const */
   timings TIMESTAMP(0);


	/* permissions */
	DATA dataAndStatusOT;


	/* return */
	ret returnO:=returnO(0);
BEGIN
	outerrors:=returnOT();

-- 	changeTime:=systimestamp;
	/* check perm and existence of the guid */
	/* serialize and
	   check if guid has permission/exist.
		Currenty permisison_util.getGuidPermission does not
		return reason code - non-exiting/not allowed.*/
		FOR i IN 1..guids.COUNT LOOP
			--data(i).guid:=guids(i).guid;
			-- remember where's guid as we may loose ordering
			guidMarkerPerms(guids(i).guid).a:=i;
		END LOOP;

		checkPrincipalPermissions(
				securityToken=>securityToken,
            guids=>guids,
				DATA=>DATA,
				op=>Permission_Util.setMetaOp);

	/* for these which are ok update stats,
	the rest already have errors set so remove them from input
	Note - change is unconditional, i.e. resets times regardless
	of values provided
	*/
	--get rid of errorenous items from stats

   DEBUGON();

	timings:=SYSTIMESTAMP();
	FOR i IN 1..stats.COUNT LOOP
		IF DATA(i).permitted = 1 THEN
			k:=guidMarkerPerms(DATA(i).guid).a;
			guidsRecs(i):=DATA(i).guid;
			file_checksumRecs(i):=stats(k).checkSum;
			statusRecs(i):=stats(k).status;
--         if stats(k).modifyTime=0 then
      mTimeRecs(i):=NVL(stats(k).modifyTime,timings);
    --        else
      --      mTimeRecs(i):=stats(k).modifyTime;
--         end if;
			file_sizeRecs(i):=stats(k).fileSize;

		ELSE
		--add to errorenous
		outerrors.EXTEND;
		nerrorsBp:=nerrorsBp+1;
		outerrors(nerrorsBp):=
			returnO
					(
						DATA(i).guid,
						guidMarkerPerms(DATA(i).guid).a,
						Permission_Util.accessDeniedErrVal,
						Permission_Util.accessDeniedErrStr
					);
		END IF;
	END LOOP;

	--now update what has been given...
	FORALL i IN 1..guidsRecs.COUNT save EXCEPTIONS
		UPDATE fc g SET
			file_checkSum=file_checksumRecs(i),
			status=statusRecs(i),
			mTime=mTimeRecs(i),
			file_Size=file_sizeRecs(i)
--			cTime=cTimeRecs(i)
		WHERE guid=guidsRecs(i);

  RETURN ret;

EXCEPTION
	WHEN bulk_errors THEN
	nerrors := SQL%BULK_EXCEPTIONS.COUNT;
	outerrors.EXTEND(nerrors);
	FOR i IN 1..nerrors LOOP
		outerrors(i+nerrorsBp):=
			returnO
				(
				guidsRecs(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX),
				guidMarkerPerms(guidsRecs(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX)).a,
				Permission_Util.generalInternalBulkErrorVal,
				SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
				guidMarkerPerms(guidsRecs(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX)).a,1,1024)
				);
	END LOOP;

	RETURN ret;
END;

/***************************************************************
	Update replica stats for an existing replica for a single
	given guid:
	modifytime, (creationtime), se, probably also master_replica


	TODO: CHANGE TO DO UPDATE ONLY - SURL MUST EXIST
****************************************************************/
FUNCTION updateSurlStat(securityToken IN principalToken,
          inguid IN VARCHAR2,
          surlStat IN SURLEntryT,
          outerrors OUT NOCOPY returnOT)
RETURN returnO
AS
	/*serialisation*/
	rc_gRecs rc_gT;
	rc_gRec 	rc_g;
	rc_sRecs rc_sT;
	rc_sRec 	rc_s;

	k PLS_INTEGER;
	guidMarkerPerms pairsT;

   /* const */
   timings TIMESTAMP(0);

	/* permissions */
	DATA dataAndStatusOT:=dataAndStatusOT();

	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
	starterr PLS_INTEGER:=1;

	ret returnO:=returnO(0);
BEGIN
	outerrors:=returnOT();

   DEBUGON();
   /* check write permissions for single guid */
	/* handle nonauth, non existing guids */

	/* serialize input */
   DATA.EXTEND();
   DATA(1):=dataAndStatusO(inguid,0);

	-- remember where's guid as we may loose ordering

	checkPrincipalPermissions(
				securityToken=>securityToken,
				DATA=>DATA,
				op=>Permission_Util.setMetaOp);

	IF DATA(1).permitted=1 THEN
		/*
		delete all surls for guid
		insert new surls...
		or delete non existing in input, update existing ones and insert
		non-existing ones.. - not worthy as number of replicas should be realtively small (<10^3:3:4:5)
		*/
	timings:=SYSTIMESTAMP();
	FOR i IN 1..surlStat.LAST LOOP
		rc_sRec.guid:=inguid;
		rc_sRec.surl:=surlStat(i).surl;
		rc_sRec.master_replica:=surlStat(i).master_replica;
		rc_sRec.surl_lifetime:=0;	--not used yet
		rc_sRec.se:=NVL(surlStat(i).se,'dummy'); -- TODO:
		rc_sRec.cTime:=NVL(surlStat(i).creationTime,timings);
		rc_sRec.mTime:=NVL(surlStat(i).modifyTime,timings);
		rc_sRecs(i):=rc_sRec;
	END LOOP;

	--- can be changed later to delete few/merge
	-- or even maybe to regular, non bulk insert
	-- as we delete all pre-exsiting replicas

	DELETE FROM rcs WHERE guid=inguid;
	FORALL i IN 1..rc_sRecs.LAST save EXCEPTIONS
		INSERT INTO rcs
			VALUES rc_sRecs(i);
	ELSE
		ret.result:=Permission_Util.accessDeniedErrVal;
		ret.errString:=Permission_Util.accessDeniedErrStr;
	END IF;

 RETURN ret;

 EXCEPTION
 	WHEN bulk_errors THEN
	nerrors := SQL%BULK_EXCEPTIONS.COUNT;
	outerrors.EXTEND(nerrors);
	FOR i IN 1..nerrors LOOP
		outerrors(i):=
			returnO
				(
				SUBSTR(rc_sRecs(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).surl,1,36),
				SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
				Permission_Util.generalInternalBulkErrorVal,
				SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||i ,1,1024)
				);
	END LOOP;

	RETURN ret;
--	when others then
--		ret.result:=SQLCODE;
--		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
----		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
--		return ret;
END;
/****************************************************************
	Set master bit on surl for guid,surl pair
	Unset bit on previous master surl

	Error. we should check if guids are distinct, if no, throw error
		or take the first one (currently).
****************************************************************/
FUNCTION setMasterReplica(securityToken IN principalToken,
          guidSurl IN StringPairT,
          outerrors OUT NOCOPY returnOT)
RETURN returnO
AS
	/*serialisation*/
	rc_gRecs rc_gT;
	rc_gRec 	rc_g;
	rc_sRecs rc_sT;
	rc_sRec 	rc_s;

	surls ssurlT;
	guids gGuidsT;

	k PLS_INTEGER;
	guidMarkerPerms pairsT;

	/* permissions */
	DATA dataAndStatusOT:=dataAndStatusOT();

	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);

	ret returnO:=returnO(0);
BEGIN
	/* check no duplicates, if there are any, last one would win,
		unless there's problem with permissions...*/

  DEBUGON();

	/* check write permissions for guids */
   DATA.EXTEND(guidSurl.COUNT);
	FOR i IN 1..guidSurl.LAST LOOP
			DATA(i):=dataAndStatusO(guidSurl(i).string1,0);
			-- remember where's guid,surl as we may loose ordering
			guidMarkerPerms(DATA(i).guid).a:=i;
	END LOOP;

	checkPrincipalPermissions(
				securityToken=>securityToken,
				DATA=>DATA,
				op=>Permission_Util.setmetaOp);

	/* handle errors and noaccess items */
	FOR i IN 1..DATA.LAST LOOP
		k:=guidMarkerPerms(DATA(i).guid).a;
		IF DATA(i).permitted=1 THEN
			guids(i):=guidSurl(k).string1;
			surls(i):=guidSurl(k).string2;
		ELSE
			--add to errorenous
      outerrors:=returnOT();
			outerrors.EXTEND;
			nerrorsBp:=nerrorsBp+1;
			outerrors(nerrorsBp):=
				returnO
						(
							DATA(i).guid,
							k,
							Permission_Util.accessDeniedErrVal,
							Permission_Util.accessDeniedErrStr
						);
		END IF;
	END LOOP;


	--probably more efficent would be passing in the whole collection	-- we couldnt get item granularity though

	/* for a given guid - reset master bit on existing one
	- set surl with master bit set to 0 and update new one
			(if exists) to 1*/
	FORALL i IN 1..surls.LAST save EXCEPTIONS
		UPDATE rcs a
			SET master_replica=(
				SELECT
				nvl2(NULLIF(a.surl,surls(i)),0,1)
					FROM rcs s
				WHERE a.guid=s.guid AND a.surl=s.surl
				)
			WHERE
				guid=guids(i)
				;
/* handle errors */
 RETURN ret;
 EXCEPTION
 	WHEN bulk_errors THEN
	nerrors := SQL%BULK_EXCEPTIONS.COUNT;
	outerrors.EXTEND(nerrors);
	FOR i IN 1..nerrors LOOP
		outerrors(i+nerrorsBp):=
			returnO
				(
				guids(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX),
				guidMarkerPerms(guids(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX)).a,
				Permission_Util.generalInternalBulkErrorVal,
				SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
				guidMarkerPerms(guids(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX)).a,1,1024)
				);
	END LOOP;
	RETURN ret;

	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
		RETURN ret;
END;

/****************************************************************
	Update status for guid, based on the mask provided
		ro,rw, final,..:3

****************************************************************/
FUNCTION updateStatus(securityToken IN principalToken,
          guids IN guidOT,
          status IN INTEGER,
          mask IN INTEGER,
          outerrors OUT NOCOPY returnOT)
RETURN returnO
AS
ret returnO:=returnO(0);
BEGIN
	--later
--check setmeta bit
--calculate the value using the mask (take the old value and apply the mask)




 RETURN ret;
 EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
		RETURN ret;
END;


/****************************************************************
	Remove  mappings for guid,surl pair.
	guids deleted if no at least single surl exists for it.


****************************************************************/
-- TODO: check proper permissions (set_metadata)
FUNCTION removeGuidReplica(securityToken IN principalToken,
          guidSurl IN StringPairT,
          removeGuidOnLastReplica IN SMALLINT,
          outerrors OUT NOCOPY returnOT)
RETURN returnO
AS
	/*serialisation*/
	surls ssurlT;
	guids gGuidsT;
	delGuids guidOT:=guidOT();

	k PLS_INTEGER;
	guidMarkerPerms pairsT;


	/* permissions */
	DATA dataAndStatusOT:=dataAndStatusOT();

	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);

	ret returnO:=returnO(0);
BEGIN
	/* remember positioning */
	/* check permissions for removal or write:6*/
	/* check write permissions for guids */
   DATA.EXTEND(guidSurl.COUNT);
	FOR i IN 1..guidSurl.LAST LOOP
			DATA(i):=dataAndStatusO(guidSurl(i).string1,0);
			-- remember where's guid,surl as we may loose ordering
			guidMarkerPerms(DATA(i).guid).a:=i;
	END LOOP;

	checkPrincipalPermissions(
				securityToken=>securityToken,
				DATA=>DATA,
				op=>Permission_Util.setmetaOp);

	/* handle errors and noaccess items */
	/* serialize - guid,surl */
	FOR i IN 1..DATA.LAST LOOP
		k:=guidMarkerPerms(DATA(i).guid).a;
		IF DATA(i).permitted=1 THEN
			guids(i):=guidSurl(k).string1;
			surls(i):=guidSurl(k).string2;
		ELSE
		/* handle noaccess errors, remove them from input */
			--add to errorenous
			outerrors.EXTEND;
			nerrorsBp:=nerrorsBp+1;
			outerrors(nerrorsBp):=
				returnO
						(
							DATA(i).guid,
							k,
							Permission_Util.accessDeniedErrVal,
							Permission_Util.accessDeniedErrStr
						);
		END IF;
	END LOOP;

	BEGIN
	/* delete from rcs where guid=guid(i) and surl=surl(i); */
	FORALL i IN 1..surls.LAST save EXCEPTIONS
		DELETE /* +RULE */ FROM rcs
			WHERE guid=guids(i) AND surl=surls(i)
			;
	EXCEPTION
		WHEN bulk_errors THEN
		nerrors := SQL%BULK_EXCEPTIONS.COUNT;
		outerrors.EXTEND(nerrors);
		FOR i IN 1..nerrors LOOP
			k:=guidMarkerPerms(guids(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX)).a;
			outerrors(i+nerrorsBp):=
				returnO
					(
					guids(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX),
					k,
					Permission_Util.generalInternalBulkErrorVal,
					SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
					k,1,1024)
					);
			guids.DELETE(k);
			surls.DELETE(k);
		END LOOP;
		nerrorsBp:=nerrorsBp+nerrors;
	END;

	/* check if any of guids became dangling one, remove them */
	--delete from (select /* +USE_NL(i) */ g.rowid from rcg g,gg i where g.guid=i.guid and not exists (select guid from rcs s where s.guid=g.guid));
	IF removeGuidOnLastReplica<>0 THEN
	FORALL i IN 1..surls.LAST save EXCEPTIONS
	DELETE FROM
		(SELECT /*+ cardinality( guids 100 )  */
			g.ROWID
			FROM fc g WHERE g.guid=guids(i)
				AND NOT EXISTS (SELECT guid FROM rcs s WHERE s.guid=g.guid));
  END IF;

	/* handle errors */

 RETURN ret;
 EXCEPTION
  	WHEN bulk_errors THEN
	nerrors := SQL%BULK_EXCEPTIONS.COUNT;
	outerrors.EXTEND(nerrors);
	FOR i IN 1..nerrors LOOP
		k:=guidMarkerPerms(guids(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX)).a;
		outerrors(i+nerrorsBp):=
			returnO
				(
				guids(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX),
				k,
				Permission_Util.generalInternalBulkErrorVal,
				SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
				k,1,1024)
				);
	END LOOP;
	RETURN ret;
--
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
		RETURN ret;
END;



/****************************************************************
	Remove guid along with all mapped surls.
	Deletes also any permission info.
****************************************************************/
FUNCTION removeGuid(securityToken IN principalToken,
          guids IN guidOT,
          outerrors OUT NOCOPY returnOT)
RETURN returnO
AS
	/*serialisation*/
	delguids gGuidsT;

	k PLS_INTEGER;
	guidMarkerPerms pairsT;

	/* permissions */
	DATA dataAndStatusOT;

	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);

	ret returnO:=returnO(0);
BEGIN
	outerrors:=returnOT();
	/* check permissions for removal */
	/* handle noaccess errors, remove them from input */
	/* serialize - guid,surl */
   DEBUGON();

	FOR i IN 1..guids.LAST LOOP
			-- remember where's guid,surl as we may loose ordering
			guidMarkerPerms(guids(i).guid).a:=i;
	END LOOP;

	checkPrincipalPermissions(
				securityToken=>securityToken,
				guids=>guids,
				DATA=>DATA,
				op=>Permission_Util.removeFileOp);

	/* handle errors and noaccess items */
	/* serialize - guid,surl */
	FOR i IN 1..DATA.LAST LOOP
		k:=guidMarkerPerms(DATA(i).guid).a;
		IF DATA(i).permitted=1 THEN
			delguids(i):=guids(k).guid;
		ELSE
		/* handle noaccess errors, remove them from input */
			--add to errorenous
			outerrors.EXTEND;
			nerrorsBp:=nerrorsBp+1;
			outerrors(nerrorsBp):=
				returnO
						(
							DATA(i).guid,
							k,
							Permission_Util.accessDeniedErrVal,
							Permission_Util.accessDeniedErrStr
						);
		END IF;
	END LOOP;

	/* delete from rcg and rcs where guid=guid(i); */
	/* handle errors */


	BEGIN
	FORALL i IN 1..delguids.COUNT save EXCEPTIONS
		DELETE FROM fc
			WHERE guid=delguids(i);
	EXCEPTION
		WHEN bulk_errors THEN
		nerrors := SQL%BULK_EXCEPTIONS.COUNT;
		outerrors.EXTEND(nerrors);
		FOR i IN 1..nerrors LOOP
			k:=guidMarkerPerms(delguids(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX)).a;
			outerrors(i+nerrorsBp):=
				returnO
					(
					delguids(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX),
					k,
					Permission_Util.generalInternalBulkErrorVal,
					SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
					k,1,1024)
					);
		END LOOP;

	END; --begin

	/* acl,rcs deleted via trigger */

 RETURN ret;
 EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
		RETURN ret;
END;



/****************************************************************
	Sets permissions for the guid (if user have P right)

****************************************************************/
FUNCTION setGuidPermission(securityToken IN principalToken,
          permissions IN PermissionEntryT,
          outerrors OUT NOCOPY returnOT)
RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
   itemCheck updateFCSimpleT;

   pids bulkPrincipalidT;
   gids bulkGroupIdT;
   ups bulkUserPT;
   gps bulkGroupPT;
   ops bulkOtherPT;
   guids bulkGuidT;


   j PLS_INTEGER:=0;

   bulkAcl ACLIdList:=ACLIdList();
   guidMarker pairR;
   guidMarkerAcl pairsT;


  	/* errors */
	nerrors PLS_INTEGER:=0;
   nerrorsBp PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
   errGuid VARCHAR2(36);
   errSurl rcs.surl%TYPE;
   nBadItems PLS_INTEGER:=0;

	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);

   /* consider updating using full bulk wo forall, just check returning for errors */
   /* what could be possible error there:6 - locks:7*/

   -- check if perms ok, report errors
   --get data for sel update, get data for
   DEBUGON();

   SELECT
   (
   SELECT
   updateFCSimpleO(parent_id,file_id,guid,NULL)
 	FROM
	FC_FILE f
	WHERE
   guid = i.item
   AND Permission_Util.isOkAcl(userperm,groupperm,otherperm,
					principal_id,userPrincipal,
					group_id,
			      principals,
					guid, --input guid
					Permission_Util.permissionOp)=1
	)
   BULK COLLECT INTO itemCheck
   FROM TABLE(permissions) i;

   FOR i IN 1..itemCheck.LAST LOOP
      IF itemCheck(i) IS NULL THEN
         nerrors:=nerrors+1;
         outErrors.EXTEND();
         outErrors(nerrors):=
            returnO(permissions(i).item,i,Permission_Util.accessDeniedErrVal,Permission_Util.accessDeniedErrStr);
      ELSE
         pids(i):=Permission_Util.getPrincipalIdIUW(permissions(i).perm.userName,null,1);
         gids(i):=Permission_Util.getPrincipalIdIUW(permissions(i).perm.groupName,null,1);
         ups(i):=permissions(i).perm.userperm;
         gps(i):=permissions(i).perm.groupperm;
         ops(i):=permissions(i).perm.otherperm;
         guids(i):=itemCheck(i).guid;
         -- acls
         IF permissions(i).perm.acl IS NOT NULL THEN
         guidMarker.a:=j+1;
   		guidMarker.b:=j+permissions(i).perm.acl.COUNT;
      	guidMarkerAcl(itemCheck(i).guid):=guidMarker;
  			bulkAcl.EXTEND(permissions(i).perm.acl.COUNT);
			FOR ac IN 1..permissions(i).perm.acl.COUNT LOOP
			 j:=j+1;
			 bulkAcl(j):=ACLEntryIdO(
			 itemCheck(i).guid,
			 Permission_Util.getPrincipalIdIUW(permissions(i).perm.acl(ac).principal,null,1),
			 permissions(i).perm.acl(ac).perm
			 );
			END LOOP;
         END IF; --acl not null
         -- /acls
      END IF;
   END LOOP;

   -- for perms ok, update permissions fields
  	BEGIN
		FORALL i IN 1..pids.COUNT save EXCEPTIONS
			UPDATE fc
			SET
				principal_id=pids(i),
				group_id=gids(i),
				userperm=ups(i),
				groupperm=gps(i),
				otherperm=ops(i)
				WHERE
				guid=guids(i)
            ;
	EXCEPTION
		WHEN bulk_errors THEN
		nerrors := SQL%BULK_EXCEPTIONS.COUNT;
		outerrors.EXTEND(nerrors);
		FOR i IN 1..nerrors LOOP
			outerrors(i+nerrorsBp):=
				returnO
					(
					permissions(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).item,
					SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
					Permission_Util.generalInternalBulkErrorVal,
					SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| 'at index '||i ,1,1024)
					);
			bulkAcl.DELETE(guidMarkerAcl(permissions(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).item).a,
							   guidMarkerAcl(permissions(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).item).b);
		END LOOP;
		nerrorsBp:=nerrorsBp+nerrors;
      END;

   -- delete from acls entries not existent in input acls
   --items to delete - exist in acl but not in input
   DELETE FROM acl
   WHERE
   (guid, principal_id) IN
   (
     SELECT /*+ cardinality( d 100 ) */ DISTINCT a.guid, a.principal_id
     FROM
     acl a,
     TABLE(CAST(bulkAcl AS ACLIdList)) d
     WHERE
       a.guid=d.guid
     MINUS
     SELECT /*+ cardinality( d 100 ) */ a.guid, a.principal_id
     FROM
     acl a,
     TABLE(CAST(bulkAcl AS ACLIdList)) d
     WHERE
       a.guid=d.guid AND a.principal_id=d.principal_id
    )
    ;
   -- merge into acls for the rest
   -- when sth wrong here throw general exception
   	MERGE /*+ cardinality (S 100) */ INTO acl D
      USING (SELECT guid,principal_id,perm FROM TABLE(CAST(bulkACl AS ACLIdList))) S
      ON (S.principal_id=D.principal_id AND S.guid=D.guid)
      WHEN MATCHED THEN UPDATE SET perm=S.perm
      WHEN NOT MATCHED THEN INSERT VALUES(S.guid,S.principal_id,S.perm);


	RETURN ret;
END;


/****************************************************************
	Optional
	Call to set deafult permissions for newly created guid when
	no permission is passed in during creation.

	This is users' metadata so we need setter for permission
	as only chosen ones should be able to update this one.
	Currently onl user can change it.
****************************************************************/
--TODO: change to virtual permission tree using fc storage and not FAS.
FUNCTION setDefaultPrincipalPermission(securityToken IN principalToken,
          principal IN VARCHAR2,
          defaultPermission IN PermissionO,
          outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS

	/* uses fas ability to store permissions for bare (no-rc) guids
			cheap but could be changed when fas is not there anymore...
			Also guid generation from within oracle could be risky...
	*/
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializers */
	dataO dataAndStatusO;
	permOut PermissionO;
	permGuid VARCHAR2(36);
	perm PermissionEntryT;

	ret returnO:=returnO(0);

BEGIN
	--later
	/* check if permission to change ok/or does not exist and
	user itself is setting or admin*/
   DEBUGON();

	Permission_Util.getPrincipalIdsO(securityToken, userPrincipal, principals,1);

	/* subqry could be wo pd, but this way we make it inependent
		from initial globaldefault if maliciously changed */
	/* merge into FAS if permitted */

	/* get other, non-perm fields from globaldafaults,
		may cause diffrent settings for users defined
		in diff point in time
	*/

	/* merge into pd if permitted */


	BEGIN
	SELECT /*+ cardinality (s 10) cardinality (inP 2)*/
			dataAndStatusO(s.guid,
			Permission_Util.isOkAclJoin(
			bp.userperm,bp.groupperm,bp.otherperm,
				bp.principal_id,userPrincipal,
				inP.item2,
				principals,
				s.guid,
				Permission_Util.writeFileOp
				))
	COLLECT INTO dataO
	FROM
			pd s , bp ,
			TABLE(principals) inP
			WHERE  s.PRINCIPAL_ID = userPrincipal AND s.guid=bp.guid
						AND bp.group_id=inP.item2(+)
		;
	EXCEPTION --should not happen...
		WHEN NO_DATA_FOUND THEN
    permGuid:=generateGuid;
		INSERT INTO pd(principal_id,guid,perm,defaultParent)
			VALUES(userPrincipal,permGuid,defaultPermission,2);		-- TODO: create default home same time...
                                                            -- create it in the /VO/userPrincipal, dont really check for permissions
                                                            -- use it
                                                            -- defaultParent, hack for now

      -- This should come from global defaults
		perm:=
		PermissionEntryT
		(
		PermissionEntryO
			 (
			 permGuid,
				PermissionO(
					securityToken.user_name,
					securityToken.group_names(1).NAME,
					255,--bin_to_num(1,1,1,1,1,1,1,1),
					80,--bin_to_num(0,1,0,1,0,0,0,0),
					0,--bin_to_num(0,0,0,0,0,0,0,0),
					NULL)
				)
		);

		ret:=Permission_Util.addGuid(
				 securityToken,
				 perm,
				 outerrors);
		RETURN ret;
	END;

	IF dataO.permitted=1 THEN	--can update
		UPDATE pd SET perm=defaultPermission
			WHERE principal_id=userPrincipal;
	ELSE --not permitted
		ret.result:=Permission_Util.accessDeniedErrVal ;
		ret.errString:=Permission_Util.accessDeniedErrStr;
	END IF;

--	merge /*+ cardinality (S 100) */ into pd D
--	using (
--		select /*+ cardinality (s 100) cardinality (inP 10)*/
--			s.guid,
--			(case(
--			permission_util.isOkAcl(
--			bp.userperm,bp.groupperm,bp.otherperm,
--				bp.principal_id,userPrincipal,
--				inP.item2,
--				principals,
--				s.guid,
--				permission_util.WriteFileOp
--				)
--				)
--			when 0 then s.perm
--			else defaultPermission end) pe
--				from
--					pd s , bp ,
--					table(principals) inP
--					where  s.PRINCIPAL_ID = userPrincipal and s.guid=bp.guid
--								and bp.group_id=inP.item2(+)
--			) S
--	on (d.guid=s.guid)
--	when matched then update set perm=s.pe
--	when not matched then
--		insert (guid,principal_id,perm)
--			values(permGuid,userPrincipal,defaultPermission)
--		returning perm into permOut
--		;


 RETURN ret;
 EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
		RETURN ret;
END;



/****************************************************************
	Optional
	If no default permission exists for a user, use global one.
	It has to be defined, preconfigured so we don't have to check
	for its existence.
****************************************************************/
FUNCTION setDefaultGlobalPermission(securityToken IN principalToken,
          defaultPermission IN PermissionO,
          outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializers */
	principalId pd.principal_id%type;

	ret returnO:=returnO(0);

BEGIN
	--later
	/* check if permission to change ok/or does not exist and
	user itself is setting or admin*/

	Permission_Util.getPrincipalIdsO(securityToken, userPrincipal, principals,1);


	UPDATE pd SET perm=defaultPermission
	WHERE guid =
	(	SELECT /*+ cardinality (s 100) cardinality (inP 10)*/
 		(CASE(
 		Permission_Util.isOkAclJoin(
		bp.userperm,bp.groupperm,bp.otherperm,
			bp.principal_id,userPrincipal,
			inP.item2,
	      principals,
			s.guid,
			Permission_Util.writeFileOp)
			) WHEN 1 THEN s.guid ELSE NULL END)
	FROM
		pd s , bp ,
		TABLE(CAST(principals AS PermStringIntPairOT)) inP
		WHERE  bp.guid = globalDefaultGuid AND s.guid=bp.guid
				   AND bp.group_id=inP.item2(+)
		)
		RETURNING principal_id INTO principalId;


	IF principalId IS NULL THEN
		ret.result:=Permission_Util.accessDeniedErrVal ;
		ret.errString:=Permission_Util.accessDeniedErrStr;
		RETURN ret;
	END IF;


 RETURN ret;
 EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
		RETURN ret;
END;


/* getters */

/****************************************************************
	Get guid permissions if principal has rights.
****************************************************************/
FUNCTION getGuidPermission(securityToken IN principalToken,
        guids IN guidOT,
        outEntries OUT NOCOPY PermissionEntryT,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS
/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	parentCheck lfnCheckT;

	/* errors */
	nerrors PLS_INTEGER:=0;

	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);

   SELECT
   (SELECT PermissionEntryO(
      i.guid,
      permissionO(
         Permission_Util.getPrincipalNameStr(principal_id),
         Permission_Util.getPrincipalNameStr(group_id),
         userperm,groupperm,otherperm,
         --permission_util.getacl(guid)
         CAST(MULTISET
         	(
         	SELECT
            p.NAME,acl.perm
				FROM acl ,PERM_PRINCIPAL p
					WHERE acl.guid=f.guid AND p.principal_id=acl.principal_id
			) AS ACLList)
         )
      )
	FROM
	FC_FILE f
	WHERE
   guid = i.guid
   AND Permission_Util.isOkAcl(userperm,groupperm,otherperm,
					principal_id,userPrincipal,
					group_id,
			      principals,
					guid, --input guid
					Permission_Util.listFileOp)=1
   )
   BULK COLLECT INTO outEntries
   FROM TABLE(guids) i
   ;

   FOR i IN 1..outEntries.LAST LOOP
      IF outEntries(i) IS NULL THEN
         nerrors:=nerrors+1;
         outErrors.EXTEND();
         outErrors(nerrors):=
            returnO(guids(i).guid,i,Permission_Util.accessDeniedErrVal,Permission_Util.accessDeniedErrStr);
      END IF;
   END LOOP;
	RETURN ret;END;

/****************************************************************
	Reversed lookup - return guid for given surl

	For denied/nonnexisting-return null + error

	todo: check if order is ok.

			!! change it to return StringPairT !!
			for now change so it returns
****************************************************************/


FUNCTION getGuidForSurl(securityToken IN principalToken,
        surls IN surlT,
        outEntries OUT NOCOPY StringPairT,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS
	/*serialisation*/
	tmpStat StringPairT;
	permSimple booleanT;

	k PLS_INTEGER;
	guidMarkerPerms bigpairsT;

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;

	ret returnO:=returnO(0);
BEGIN
	outerrors:=returnOT();
   DEBUGON();

	FOR i IN 1..surls.LAST LOOP
			-- remember where's guid,surl as we may loose ordering
			guidMarkerPerms(surls(i)).a:=i;
	END LOOP;
	outEntries:=StringPairT();

	Permission_Util.getPrincipalIdsO(securityToken,
		userPrincipal,
		principals,
		0);


SELECT /*+ cardinality( i 5 ) cardinality(inP 10) use_nl(inP) */
		StringPairO(i.COLUMN_VALUE,s.guid),
		Permission_Util.isOkAclJoin(bp.userperm,bp.groupperm,bp.otherperm,
			bp.principal_id,userPrincipal,
			bp.group_id,
	      principals,
			s.guid, Permission_Util.listFileOp
		)
	BULK COLLECT INTO tmpStat,permSimple
	FROM
	 ((( TABLE(surls) i LEFT OUTER JOIN rcs s ON s.surl=i.COLUMN_VALUE)
		LEFT OUTER JOIN fc bp ON s.guid=bp.guid)
				  LEFT OUTER JOIN TABLE(principals) inP
				  ON bp.group_id=inP.item2)
	;
	/* handle errors */
	outEntries:=StringPairT();
	outEntries.EXTEND(tmpstat.COUNT);
	FOR i IN 1..tmpstat.LAST LOOP
--	DBMS_OUTPUT.PUT_LINE(guidSimple(i));
		IF permSimple(i)=1 THEN
			outEntries(guidMarkerPerms(tmpStat(i).string1).a):=tmpStat(i);
		ELSE
			outEntries(guidMarkerPerms(tmpStat(i).string1).a):=NULL;
			outerrors.EXTEND;
			nerrorsBp:=nerrorsBp+1;
			outerrors(nerrorsBp):=
				returnO
						(
						tmpStat(i).string1,
						guidMarkerPerms(tmpStat(i).string1).a,
						Permission_Util.accessDeniedErrVal,
						Permission_Util.accessDeniedErrStr
						);
		END IF;

	END LOOP;
 RETURN ret;
END;



/****************************************************************
	Get master replica string for a guid
****************************************************************/
FUNCTION getMasterReplica(securityToken IN principalToken,
        guids IN guidOT,
        outEntries OUT NOCOPY StringPairT,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS

	/*serialisation*/

	tmpStat StringPairT;
	permSimple booleanT;

	guidMarkerPerms pairsT;

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	DATA Permission_Util.dataAndStatusT;


	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;


	ret returnO:=returnO(0);
BEGIN
	outErrors:=returnOT();
	outEntries:=StringPairT();

	Permission_Util.getPrincipalIdsO(securityToken,
		userPrincipal,
		principals,
		0);

	/* serialize */
	/* check guids for read (list:1) permissions */
	/* handle no access errors */

	/* transform guid into data */
	FOR i IN 1..guids.LAST LOOP
		guidMarkerPerms(guids(i).guid).a:=i;
	END LOOP;



SELECT /*+ cardinality( i 100 ) cardinality (inP 5)  */
		StringPairO(i.guid,g.surl),
		Permission_Util.isOkAclJoin(userperm,groupperm,otherperm,
			bp.principal_id,userPrincipal,item2,
			principals,
			i.guid,Permission_Util.listFileOp
		)
--		)
 BULK COLLECT INTO tmpStat,permSimple
 FROM
 ((( TABLE(guids) i LEFT OUTER JOIN rcs g ON i.guid=g.guid AND master_replica=1)
 		LEFT OUTER JOIN fc bp ON i.guid=bp.guid)
			  LEFT OUTER JOIN TABLE(principals) inP
 ON
 (
   bp.group_id=inP.item2
  )
)
;

	outEntries.EXTEND(tmpstat.COUNT);
	FOR i IN 1..tmpstat.LAST LOOP
--	DBMS_OUTPUT.PUT_LINE(guidSimple(i));
		IF permSimple(i)=1 THEN
			outEntries(guidMarkerPerms(tmpStat(i).string1).a):=tmpStat(i);
		ELSE
			outEntries(guidMarkerPerms(tmpStat(i).string1).a):=NULL;
			outerrors.EXTEND;
			nerrorsBp:=nerrorsBp+1;
			outerrors(nerrorsBp):=
				returnO
						(
						tmpStat(i).string1,
						guidMarkerPerms(tmpStat(i).string1).a,
						Permission_Util.accessDeniedErrVal,
						Permission_Util.accessDeniedErrStr
						);
		END IF;

	END LOOP;

	RETURN ret;
 EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
		RETURN ret;
END;



/****************************************************************
	Get list of all surls for a guid.

	Almost same logic as in getMasterReplica (above)

****************************************************************/
FUNCTION listSurlsByGuid(securityToken IN principalToken,
        guids IN guidOT,
        outEntries OUT NOCOPY StringPairT,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS
	/*serialisation*/

	/* permissions*/
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;


	k PLS_INTEGER;
	guidMarkerPerms pairsT;

	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;

	ret returnO:=returnO(0);
BEGIN
	outerrors:=returnOT();
	/* serialize */
	/* check guids for read (list:2) permissions */
	/* handle no access errors */

	/* transform guid into data */
	FOR i IN 1..guids.LAST LOOP
		guidMarkerPerms(guids(i).guid).a:=i;
	END LOOP;

--	/* check permissions, handle errors */
	Permission_Util.getPrincipalIdsO(securityToken,
		userPrincipal,
		principals,
		0);


	SELECT StringPairO(guid,s)
	 BULK COLLECT INTO outEntries
	FROM
	(SELECT /*+  cardinality( i 100 ) cardinality (inP 10)  */
		DISTINCT i.guid,
		(CASE(Permission_Util.isOkAclJoin(userperm,groupperm,otherperm,
			g.principal_id,userPrincipal,item2,
			principals,
			i.guid,Permission_Util.listFileOp
			)) WHEN 1 THEN surl ELSE NULL END) s
	 FROM
	((( ( TABLE(guids) i LEFT OUTER JOIN fc g ON i.guid=g.guid
			AND i.guid=g.guid)
				LEFT OUTER JOIN rcs s ON s.guid=i.guid)
				  LEFT OUTER JOIN TABLE(principals) inP ON g.group_id=inP.item2)

	)
	)
	;

	/* fill up outEntries for pemitted items, handle errors */
	/* assumption outerr have all guids */
	FOR i IN 1..outEntries.LAST LOOP
		IF outEntries(i).string2 IS NULL AND outEntries(i).string1 IS NOT NULL THEN
			outerrors.EXTEND;
			nerrorsBp:=nerrorsBp+1;
			k:=guidMarkerPerms(outEntries(i).string1).a;
			outerrors(nerrorsBp):=
				returnO
						(
						outEntries(i).string1,
						k,
						Permission_Util.accessDeniedErrVal,
						Permission_Util.accessDeniedErrStr
						);

		END IF;
	END LOOP;
	RETURN ret;
 EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
		RETURN ret;
END;



/****************************************************************
	List full surl info (with optional permissions) for guid.

	Same as previous but gets all data from rcg, rcs and optionally
	from permission module.
****************************************************************/
FUNCTION listReplicasByGuid(securityToken IN principalToken,
        guids IN guidOT,
        withPermissions IN INTEGER,
        outEntries OUT NOCOPY RCEntryT,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS

	/*serialisation*/
	tmpguidStat GuidStatExtT;
	tmpSurlEntries SurlEntryExtRT;

	k PLS_INTEGER;
	guidMarkerPerms pairsT;

	tmpGuidStr RC_REPLICA.guid%TYPE;

   permissions PermissionEntryT;
   rcss surlEntryTablesT;
   rcgs rc_gT;
	/* permissions*/
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

   g PLS_INTEGER;
   j PLS_INTEGER;
	nErrors PLS_INTEGER:=0;

   ret returnO:=returnO(0);

	outGuidStats       outGuidStatsT;
	outSurlStats       outSurlStatsT;
	outPermissions    outPermissionsT;
	gGuids            gGuidsT;

	 outEntriesR RCEntryRT;
BEGIN
	outerrors:=returnOT();
	/* check permissions for guids*/
	/* handle errors */

	/* get all ok surls info for given guids in [guid,surl] pairs
	with attached permissions if flag wants so */


  /* if permissions needed, use permission_util.getpermission,
      then derive security errors from it, use guids as keys as well.
    else
      permission_util.checkpermission
    end if;
  */

  /* first get all guidstat from rcg (where permission ok), serialize */

  /* get all surlsstat items from rcs where guid permitted, serialize */
  	/* transform guid into data */
	FOR i IN 1..guids.LAST LOOP
		guidMarkerPerms(guids(i).guid).a:=i;
	END LOOP;

	Permission_Util.getPrincipalIdsO(securityToken,
		userPrincipal,
		principals,
		0);
-- TODO: always checks for permissions here but not always returns acls, change it.
-- change to real bulk ops
  -- and remove isOkAclPermission dependency
/* got null if not permitted/not existing */

	SELECT /*+ cardinality( i 100 ) cardinality (inP 5)  */
			RCEntryO(
			i.guid,
			guidStatO(g.file_size,g.cTime,g.file_checksum,g.status,g.mTime),
			CAST(MULTISET
			(
			SELECT s.surl,s.mTime, s.cTime,s.se,s.master_replica
				FROM RC_SURL s
				WHERE s.guid=i.guid
			) AS surlEntryT)
			,
			Permission_Util.isOkAclPermission(g.userperm,g.groupperm,g.otherperm,
			g.principal_id,userPrincipal,
			g.group_id,item2,
	        principals,
			i.guid,
			Permission_Util.listFileOp
		)
		)
	 BULK COLLECT INTO outEntries
	 FROM
	 (( (TABLE(guids) i LEFT OUTER JOIN fc g ON i.guid=g.guid
			AND i.guid=g.guid)
				  LEFT OUTER JOIN TABLE(principals) inP
	 ON
	 (
		g.group_id=inP.item2
	  )
	)
	)
	;
	FOR i IN 1..outEntries.LAST LOOP
		IF outEntries(i).permission IS NULL THEN --not permitted
			outEntries(i).guidstat:=NULL;
			outEntries(i).surlstats:=NULL;
			--add error
			outerrors.EXTEND;
			nErrors:=nerrors+1;
			outerrors(nErrors):=
				returnO
						(
						outEntries(i).guid,
						guidMarkerPerms(outEntries(i).guid).a,
						Permission_Util.accessDeniedErrVal,
						Permission_Util.accessDeniedErrStr
						);
		ELSE
		  IF withPermissions=0 THEN --use getpermissions to get both perm values (same readFileOp) and perform access control check
			outEntries(i).permission:=NULL;
		   END IF;
		END IF;
	END LOOP;

 RETURN ret;
 EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
		RETURN ret;
END;



/****************************************************************
	check for existence of the guid in the catalog
	never returns any non severe errors
****************************************************************/
-- TODO: decide if ordering needed...
FUNCTION hasGuid(securityToken IN principalToken,
        guids IN guidOT,
        outEntries OUT NOCOPY StringBooleanT,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS

	/* serialization */
	k PLS_INTEGER;
	guidMarkerPerms pairsT;

	DATA Permission_Util.dataAndStatusT;

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;


	/* return */
	ret returnO:=returnO(0);

BEGIN
--	outEntries:=booleanT();
	outerrors:=returnOT();
	/* permissions of guid, if permitted=1 then outEntries=true  */

	/* remember order */
--	for i in 1..guids.last loop
--		-- remember where's guid,surl as we may loose ordering
--		guidMarkerPerms(guids(i).guid).a:=i;
--	end loop;

	Permission_Util.getPrincipalIdsO(securityToken,
		userPrincipal,
		principals,
		0);

	SELECT /*+cardinality(i 100) cardinality(inP 1)  */
		StringBooleanO(i.guid,
		(CASE(Permission_Util.isOkAclJoin(userperm,groupperm,otherperm,
			bp.principal_id,userPrincipal,group_id,
			principals,
			i.guid,Permission_Util.listFileOp
			)) WHEN 1 THEN 1 ELSE 0 END),0) --either denied or nonexist..
		BULK COLLECT INTO outEntries
			 FROM
	(( (TABLE(guids) i LEFT OUTER JOIN fc bp ON i.guid=bp.guid)
		LEFT JOIN TABLE(principals) inP ON bp.group_id=inP.item2)
	)
	;

	/* ugly but necessary as we have to order output */
--	outEntries.extend(data.count);
--	for i in 1..data.last loop
--		outEntries(guidMarkerPerms(data(i).guid).a):=data(i).permitted;
--	end loop;

 RETURN ret;
 EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
		RETURN ret;
END;



/****************************************************************
		Get stats only for guids.

		Errors:
		Erroneous entries at idx i hold null for respective guid
		in input guid set.
****************************************************************/
FUNCTION getGuidStat(securityToken IN principalToken,
        guids IN guidOT,
        outEntries OUT NOCOPY GuidStatT,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS
	/*serialisation*/
	tmpStat GuidStatT:=GuidStatT();

	guidSimple guidSimpleT;
	permSimple booleanT;

	k PLS_INTEGER;
	guidMarkerPerms pairsT;

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;
	DATA Permission_Util.dataAndStatusT;

	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;

	ret returnO:=returnO(0);
BEGIN
	outEntries:=GuidStatT();
	outErrors:=returnOT();

	Permission_Util.getPrincipalIdsO(securityToken,
		userPrincipal,
		principals,
		0);

	/* transform guid into data */
	FOR i IN 1..guids.LAST LOOP
--		DBMS_OUTPUT.PUT_LINE(guids(i).guid);
		guidMarkerPerms(guids(i).guid).a:=i;
	END LOOP;

SELECT /*+ cardinality( i 100 ) cardinality (inP 5)  */
			i.guid,
			Permission_Util.isOkAclJoin(g.userperm,g.groupperm,g.otherperm,
			g.principal_id,userPrincipal,
			g.group_id,
	      principals,
			i.guid,
			Permission_Util.listFileOp
		)
		,guidStatO(g.file_size,g.cTime,g.file_checksum,g.status,g.mTime)
	 BULK COLLECT INTO
	 guidSimple,permSimple
	 ,tmpStat
 FROM
 (( TABLE(guids) i LEFT OUTER JOIN fc g ON i.guid=g.guid
 		AND i.guid=g.guid)
			  LEFT OUTER JOIN TABLE(principals) inP
 ON
 (
   g.group_id=inP.item2
  )
)
;
	outEntries.EXTEND(tmpstat.COUNT);
	FOR i IN 1..tmpstat.LAST LOOP
--	DBMS_OUTPUT.PUT_LINE(guidSimple(i));
		IF permSimple(i)=1 THEN
			outEntries(guidMarkerPerms(guidSimple(i)).a):=tmpStat(i);
		ELSE
			outEntries(guidMarkerPerms(guidSimple(i)).a):=NULL;
			outerrors.EXTEND;
			nerrorsBp:=nerrorsBp+1;
			outerrors(nerrorsBp):=
				returnO
						(
						guidSimple(i),
						guidMarkerPerms(guidSimple(i)).a,
						Permission_Util.accessDeniedErrVal,
						Permission_Util.accessDeniedErrStr
						);
		END IF;

	END LOOP;
	RETURN ret;
 EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
		RETURN ret;

END;

/****************************************************************
	Get stats only for specific guid,surl pair.

	same as for rcg but on surl.
****************************************************************/
FUNCTION getSurlStat(securityToken IN principalToken,
        guidSurls IN StringPairT,
        outEntries OUT NOCOPY SURLEntryT,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS
	/* serialization */
	KEYS bigStringKeyT;
	permSimple booleanT;

	tmpStat StringPairT;

	k PLS_INTEGER;
	guidMarkerPerms bigpairsT;

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;

	ret returnO:=returnO(0);
BEGIN
	outEntries:=SURLEntryT();
	outErrors:=returnOT();

  DEBUGON();

	Permission_Util.getPrincipalIdsO(securityToken,
		userPrincipal,
		principals,
		0);

	/* check permissions for read (list:7) */
	/* for [guid,surl] pairs select from rcs */
	/*serialisation*/

	/* transform guid into data */
	FOR i IN 1..guidSurls.LAST LOOP
		guidMarkerPerms(guidSurls(i).string1).a:=i;
	END LOOP;

--
--	select /*+ cardinality( i 10 ) cardinality (inP 5)  */
--				i.string1||i.string2,
--				permission_util.isOkAclJoin(bp.userperm,bp.groupperm,bp.otherperm,
--				bp.principal_id,userPrincipal,
--				bp.group_id,
--				principals,
--				i.string1,
--				permission_util.readFileOp),
--				SurlEntryO(g.surl,g.mTime, g.cTime,g.se,g.master_replica)
--	 bulk collect into keys,permSimple,tmpStat
--	 from
--	 ((( table(guidSurls) i left outer join rcs g on i.string1=g.guid and i.string2=g.surl)
--			left outer join fc bp on i.string1=bp.guid)
--				  left outer join table(principals) inP
--	 on
--	 (
--		bp.group_id=inP.item2
--	  )
--	)
--	;

  SELECT /*+cardinality( i 5 ) cardinality(inP 10) */
      StringPairO(i.string1,s.guid),
      Permission_Util.isOkAclJoin(bp.userperm,bp.groupperm,bp.otherperm,
        bp.principal_id,userPrincipal,
        bp.group_id,
        principals,
        s.guid, Permission_Util.listFileOp
      )
      ,
    SurlEntryO(s.surl,s.mTime, s.cTime,s.se,s.master_replica)
    BULK COLLECT INTO tmpStat,permSimple,outEntries
    FROM
     ((( TABLE(guidSurls) i LEFT OUTER JOIN rcs s ON s.surl=i.string1)
      LEFT OUTER JOIN fc bp ON s.guid=bp.guid)
            LEFT OUTER JOIN TABLE(principals) inP
            ON bp.group_id=inP.item2)
  ;

	FOR i IN 1..tmpstat.LAST LOOP
--	DBMS_OUTPUT.PUT_LINE(guidSimple(i));
		IF permSimple(i)<>1 THEN
--			outEntries(guidMarkerPerms(tmpStat(i).string1).a):=tmpStat(i);
--		else
      --outEntries(guidMarkerPerms(tmpStat(i).string1).a):=null; -- TODO: check if delete is ok
      outEntries.DELETE(guidMarkerPerms(tmpStat(i).string1).a);
			outerrors.EXTEND;
			nerrorsBp:=nerrorsBp+1;
			IF permSimple(i)=0 THEN
      outerrors(nerrorsBp):=
				returnO
						(
						tmpStat(i).string1,
						guidMarkerPerms(tmpStat(i).string1).a,
						Permission_Util.accessDeniedErrVal,
						Permission_Util.accessDeniedErrStr
						);
       ELSIF permSimple(i)=2 THEN
      outerrors(nerrorsBp):=
				returnO
						(
						tmpStat(i).string1,
						guidMarkerPerms(tmpStat(i).string1).a,
						Permission_Util.NoSuchFileVal,
						Permission_Util.NoSuchFileStr
						);
       ELSE
      outerrors(nerrorsBp):=
				returnO
						(
						tmpStat(i).string1,
						guidMarkerPerms(tmpStat(i).string1).a,
						Permission_Util.generalInternalErrorVal,
						Permission_Util.generalInternalErrorStr
						);

       END IF;

		END IF;

	END LOOP;

	RETURN ret;
 EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
		RETURN ret;
END;



/****************************************************************
	Get optional defaults for user
****************************************************************/
FUNCTION getDefaultPrincipalPermission(securityToken IN principalToken,
        principal IN VARCHAR2,
        outEntry OUT NOCOPY PermissionO,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;

	ret returnO:=returnO(0);

BEGIN
	Permission_Util.getPrincipalIdsO(securityToken,
		userPrincipal,
		principals,
		0);

SELECT /*+ cardinality (inP 10)*/
 		CASE(
 		Permission_Util.isOkAclJoin(
		bp.userperm,bp.groupperm,bp.otherperm,
			bp.principal_id,userPrincipal,
			inP.item2,
	      principals,
			s.guid,
			Permission_Util.readFileOp)
			) WHEN 1 THEN s.perm ELSE NULL END
			COLLECT INTO outEntry
	FROM
		pd s , bp ,
		TABLE(principals) inP
		WHERE  s.principal_id=userPrincipal AND s.guid=bp.guid
				   AND bp.group_id=inP.item2(+)
	;
	IF outEntry IS NULL THEN
		ret.result:=Permission_Util.accessDeniedErrVal ;
		ret.errString:=Permission_Util.accessDeniedErrStr;
	END IF;
	RETURN ret;


 EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
		RETURN ret;
END;



/****************************************************************
	Get default permission for catalog.
	security check for "admin" and all other that have permissions
	on it set up by "admin".
	Permission is stored in FAS as usual for guid.
****************************************************************/
FUNCTION getDefaultGlobalPermission(securityToken IN principalToken,
        outEntry OUT NOCOPY PermissionO,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	ret returnO:=returnO(0);
BEGIN
	outerrors:=returnOT();

	Permission_Util.getPrincipalIdsO(securityToken,
		userPrincipal,
		principals,
		0);

SELECT /*+ cardinality (inP 5)*/
 		CASE(
 		Permission_Util.isOkAclJoin(
		bp.userperm,bp.groupperm,bp.otherperm,
			bp.principal_id,userPrincipal,
			inP.item2,
	    principals,
			s.guid,
			Permission_Util.readFileOp)
			) WHEN 1 THEN s.perm ELSE NULL END
			COLLECT INTO outEntry
	FROM
		pd s , bp ,
		TABLE(principals) inP
		WHERE  bp.guid = globalDefaultGuid AND s.guid=bp.guid
				   AND bp.group_id=inP.item2(+)
    ;


	IF outEntry IS NULL THEN
		ret.result:=Permission_Util.cannotGetDeafultPermissionVal ;
		ret.errString:=Permission_Util.cannotGetDeafultPermissionStr;
	END IF;
	RETURN ret;
END;


/***************************************
      get user settings
****************************************/
FUNCTION getDefaultUserSettings(securityToken IN principalToken,
        outEntry OUT NOCOPY DefaultsO
        )
	RETURN returnO
AS

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;

	ret returnO:=returnO(0);

BEGIN
	Permission_Util.getPrincipalIdsO(securityToken,
		userPrincipal,
		principals,
		0);

SELECT /*+ cardinality (inP 10)*/
 		CASE(
 		Permission_Util.isOkAclJoin(
		bp.userperm,bp.groupperm,bp.otherperm,
			bp.principal_id,userPrincipal,
			inP.item2,
	      principals,
			s.guid,
			Permission_Util.getMetaOp)
			) WHEN 1 THEN
         DefaultsO
            (
            defaultCreateLFN,
            defaultCreateLFNPattern,
            defaultParent
            )
         ELSE NULL END
			COLLECT INTO outEntry
	FROM
		pd s , bp ,
		TABLE(principals) inP
		WHERE  s.principal_id=userPrincipal AND s.guid=bp.guid
				   AND bp.group_id=inP.item2(+)
	;
	IF outEntry IS NULL THEN
		ret.result:=Permission_Util.accessDeniedErrVal ;
		ret.errString:=Permission_Util.accessDeniedErrStr;
	END IF;
	RETURN ret;

 EXCEPTION
	WHEN OTHERS THEN
		ret.result:=SQLCODE;
		ret.errString:=SUBSTR(SQLERRM, 1, 1024);
--		DBMS_OUTPUT.PUT_LINE('final exception:' || ret.errString);
		RETURN ret;
END;




/* internal get of defaults */
FUNCTION getDefaultGlobalSettings(securityToken IN principalToken,
        outEntry OUT NOCOPY DefaultsO
        )
	RETURN returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;

	ret returnO:=returnO(0);
BEGIN

	Permission_Util.getPrincipalIdsO(securityToken,
		userPrincipal,
		principals,
		0);

SELECT /*+ cardinality (inP 10)*/
 		CASE(
 		Permission_Util.isOkAclJoin(
		bp.userperm,bp.groupperm,bp.otherperm,
			bp.principal_id,userPrincipal,
			inP.item2,
	      principals,
			s.guid,
			Permission_Util.getMetaOp)
			) WHEN 1 THEN
         DefaultsO
            (
            defaultCreateLFN,
            defaultCreateLFNPattern,
            defaultParent
            )
          ELSE NULL END
			COLLECT INTO outEntry
	FROM
		pd s , bp ,
		TABLE(principals) inP
		WHERE  bp.guid = globalDefaultGuid AND s.guid=bp.guid
				   AND bp.group_id=inP.item2(+)
	;


	IF outEntry IS NULL THEN
		ret.result:=Permission_Util.cannotGetDeafultPermissionVal ;
		ret.errString:=Permission_Util.cannotGetDeafultPermissionStr;
	END IF;
	RETURN ret;
END;


/***********************************/
FUNCTION getDefaultSettings(securityToken IN principalToken,
        outEntry OUT NOCOPY DefaultsO
        )
RETURN returnO
AS

	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* errors */
	nerrors PLS_INTEGER;
	nerrorsBp PLS_INTEGER:=0;

	ret returnO:=returnO(0);

BEGIN
	ret:=getDefaultUserSettings(securityToken,outEntry);
	IF ret.result=0 AND outEntry IS NOT NULL THEN
		RETURN ret;
	ELSE
		ret:=getDefaultGlobalSettings(securityToken,outEntry);
		IF ret.result=0 AND outEntry IS NOT NULL THEN RETURN ret; END IF;
	END IF;

	RETURN ret;

END;



/************************** RC END ***************************************/

/************************* METADATA **************************************/


function getSingleIdForMeta(
  userPrincipal in PERM_PRINCIPAL.principal_id%TYPE,
	principals in out nocopy PermStringIntPairOT,
  op in pls_integer,
  item in StringBooleanO,
  fileid out nocopy mval.file_id%type,
  parentid out nocopy mval.parent_id%type)
return returnO
as

	/* return */
	ret returnO;
begin
  if item.boolval = 1 then
    SELECT
    	case
             Permission_Util.isOkAclMask(userperm,groupperm,otherperm,
                 principal_id,userPrincipal,
                 group_id,
                 principals,
                 item.string, --input guid
                 op)
        WHEN 1 THEN returnO(0)
        WHEN 0 THEN returnO(item.string,1,Permission_Util.accessDeniedErrVal,Permission_Util.accessDeniedErrStr,0)
        ELSE NULL END AS val,
        parent_id,file_id
      collect into ret,parentid, fileid
	      FROM
      	fc bp where bp.guid=item.string;
  else
	SELECT 	case  Permission_Util.isOkAclMask(userperm,groupperm,otherperm,
                 principal_id,userPrincipal,
                 group_id,
                 principals,
                 item.string, --input guid
                 op)
        WHEN 1 THEN returnO(0)
        WHEN 0 THEN returnO(item.string,1,Permission_Util.accessDeniedErrVal,Permission_Util.accessDeniedErrStr,0)
        ELSE NULL END AS val,parent_id,file_id
        collect into ret,parentid, fileid
        FROM
        FC_FILE f
        WHERE LEVEL=fireman.getNoOfPathDelimitersRoot(item.string)+1
        START WITH parent_id=0 AND file_id=1
        CONNECT BY PRIOR file_id = parent_id
          AND NAME=fireman.getpathtoken(LEVEL-1,item.string)
  ;
  end if;

  return ret;
  EXCEPTION
      WHEN NO_DATA_FOUND THEN
         ret:=returnO(0);
         ret.result:=Permission_Util.NoSuchFileVal;
         ret.errString:=Permission_Util.NoSuchFileStr;
      RETURN ret;
END;

 /* Sets the values for a group of attributes of an item/entry.
    public void setAttributes(String item, Attribute[] attributes)    */
FUNCTION setAttributes(securityToken IN principalToken,
        item in StringBooleanO,
        attributeNames IN AttributeOT,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS
/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */

  /* new attributes */

  fileid mval.file_id%type;
  parentid mval.parent_id%type;
  metainput metaObjectIntT;


	/* return */
	ret returnO;
BEGIN
   outerrors:=returnOT();
	 Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);

  -- check for perms on the item, could be either guid or lfn
  -- if permitted go, produce error if not
    -- insert into nested array input attribs
      -- check if name/type exists in the schema (global one)
      -- if not create it

--check/get all att_ids
 ret:=getSingleIdForMeta(
  userPrincipal,principals,Permission_Util.setmetaOp,
  item,fileid, parentid);
 if ret.result<>0 then
   return ret;
 end if;


select /*+ cardinality(i 10) */
  metaObjectIntO(schema_id,att_id,name ,
  decode(att_id, null,1,0),
  atttype,value)
  bulk collect into metainput
from
 	(table (attributeNames) i left outer join matt on att_name=name)
  ;

  --create attributes if dont exist, got all ids..
--insert into both
 insert first
	when toInsert=1 then
		 into matt(att_id,att_name,type_id)
			values(meta_att_seq_id.nextval,attname,attype)
	   into mval(att_id,parent_id,file_id,val)
			values(meta_att_seq_id.currval,parentid,fileid,val)
  select *
  from table(metainput)  where toInsert=1
	;


  merge into mval D
    using
    (select attid,val,attype, parentid parent_id, fileid file_id
      from table(metainput) where toInsert=0) S
    on(D.parent_id=S.parent_id and D.file_id=S.file_id and D.att_id=s.attid)
    when matched then
      update set val=S.val
    when not matched then
      insert (att_id,parent_id,file_id,val)
        values(s.attid,parentid,fileid,s.val)
  ;
--
--delete from mval where (schema_id,att_id,parent_id,file_id) in
--  (select schema_id,att_id,parent_id,file_id from mval where schema_id=0
--  and  parent_id=parentid and file_id=fileid
--  minus
--  select schema_id,att_id,parent_id,file_id from mval
--  join
--  table(metaInput) s
--  on schema_id=schemaid
--  and att_id=attid  and parent_id=parentid and file_id=fileid
--)
--;
return ret;

END;


/* Clears the values for a group of attributes.
    public void clearAttributes(String item, String[] attributeNames) */
FUNCTION clearAttributes(securityToken IN principalToken,
        item in StringBooleanO,
        attributeNames IN stringT,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS
/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */

  /* new attributes */

  fileid mval.file_id%type;
  parentid mval.parent_id%type;
  metainput metaObjectIntT;


	/* return */
	ret returnO;
BEGIN
   outerrors:=returnOT();
	 Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,1);

  -- check for perms on the item, could be either guid or lfn
  -- if permitted go, produce error if not
    -- insert into nested array input attribs
      -- check if name/type exists in the schema (global one)
      -- if not create it

--check/get all att_ids
 ret:=getSingleIdForMeta(
  userPrincipal,principals,Permission_Util.setmetaOp,
  item,fileid, parentid);
 if ret.result<>0 then
   return ret;
 end if;

  -- delete all values for an attribute

  delete from mval where (schema_id,att_id,parent_id,file_id ) in (
  select schema_id,att_id,parent_id,file_id from mval join table
  (metaObjectIntT(metaObjectIntO(0,1,null,0,'string','value33'),metaObjectIntO(0,4,null,0,'string','value44'))) s
  on schema_id=schemaid and att_id=attid  and parent_id=0 and file_id=1
  )
  ;

	RETURN ret;
END;


/* Gets a group of attributes of a given item/entry.
    public Attribute[] getAttributes(String item, String[] attributeNames) */
FUNCTION getAttributes(securityToken IN principalToken,
        item in StringBooleanO,
        attributeNames IN stringT,
        outEntries OUT NOCOPY AttributeOT,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS
/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */

  /* new attributes */

  fileid mval.file_id%type;
  parentid mval.parent_id%type;
  metainput metaObjectIntT;


	/* return */
	ret returnO;
BEGIN
   outerrors:=returnOT();
	 Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);

--check/get all att_ids
 ret:=getSingleIdForMeta(
  userPrincipal,principals,Permission_Util.getmetaOp,
  item,fileid, parentid);
 if ret.result<>0 then
   return ret;
 end if;

  select AttributeO(att_name,type_id,val)
  bulk collect into outEntries
  from
    (msch natural join matt)  natural join mval
	where schema_name = 'default schema'
	and att_name
	in
	(select /*+ cardinality (ats 10) */ column_value from
		table(attributeNames) ats)
	and (parent_id =parentId and file_id=fileid)
  ;




	RETURN ret;
END;


/* List all attributes associated with an entry.
    public Attribute[] listAttributes(String item)     */
FUNCTION listAttributes(securityToken IN principalToken,
        item in StringBooleanO,
        outEntries OUT NOCOPY AttributeOT,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS
/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */

  /* new attributes */

  fileid mval.file_id%type;
  parentid mval.parent_id%type;
  metainput metaObjectIntT;


	/* return */
	ret returnO;
BEGIN
   outerrors:=returnOT();
	 Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);

--check/get all att_ids
 ret:=getSingleIdForMeta(
  userPrincipal,principals,Permission_Util.getmetaOp,
  item,fileid, parentid);
 if ret.result<>0 then
   return ret;
 end if;

  select AttributeO(att_name,type_id,val)
  bulk collect into outEntries
  from
    (msch natural join matt)  natural join mval
	where schema_name = 'default schema'
	and (parent_id =parentId and file_id=fileid)
  ;

	RETURN ret;
END;

/* Performs a generic query on the catalog, returning the corresponding item's identifiers.
    public String[] query(String query, String type, int limit, int offset)     */

FUNCTION QUERY(securityToken IN principalToken,
        QUERY IN VARCHAR2,
        LIMIT IN SMALLINT,
        offset IN SMALLINT,
        outEntries OUT NOCOPY stringT,
        outerrors OUT NOCOPY returnOT)
	RETURN returnO
AS
/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	parentCheck lfnCheckT;


	/* return */
	ret returnO:=returnO(0);
BEGIN
   outerrors:=returnOT();
	 Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);


	RETURN ret;
END;


/************************* METADATA END************************/


/* intternal helpers*/



/* testing helper */
FUNCTION genLfns
	RETURN lfnT
	deterministic
AS
BEGIN
  DEBUGON();
	RETURN lfnt(
	'/a','/a/c','/a/a','/sdfsd','/a/a/a','/dir_1/dir_2/dir_3',
--'/dir_1/dir_2/dir_3a','/dird_1/dird_2/dird_3/dird_4/file_7','/dir_1/dir_2/dir_3b','/dird_1/dird_2/dird_3/dird_4/file_7',
--'/dir_1/dir_2/dir_3c','/dird_1/dird_2/dird_3/dird_4/file_7d',
--'/dir_1/dir_2/dir_3e','/dird_1/dird_2/dird_3/dird_4/file_7f',
--'/a','/a/c','/a/a','/sdfsd','/a/a/a','/dir_1/dir_2/dir_3',
--'/dir_1/dir_2/dir_3a','/dird_1/dird_2/dird_3/dird_4/file_7','/dir_1/dir_2/dir_3b','/dird_1/dird_2/dird_3/dird_4/file_7',
--'/dir_1/dir_2/dir_3c','/dird_1/dird_2/dird_3/dird_4/file_2',
--'/dir_1/dir_2/dir_3e','/dird_1/dird_2/dird_3/dird_4/file_4f',
--'/a','/a/c','/a/a','/sdfsd','/a/a/a','/dir_1/dir_2/dir_3',
--'/dir_1/dir_2/dir_3a','/dird_1/dird_2/dird_3/dird_4/file_7','/dir_1/dir_2/dir_3b','/dird_1/dird_2/dird_3/dird_4/file_7',
--'/dir_1/dir_2/dir_3c','/dird_1/dird_2/dird_3/dird_4/file_7d',
--'/dir_1/dir_2/dir_3e','/dird_1/dird_2/dird_3/dird_4/file_7f',
--'/a','/a/c','/a/a','/sdfsd','/a/a/a','/dir_1/dir_2/dir_3',
--'/dir_1/dir_2/dir_3a','/dird_1/dird_2/dird_3/dird_4/file_7','/dir_1/dir_2/dir_3b','/dird_1/dird_2/dird_3/dird_4/file_7',
--'/dir_1/dir_2/dir_3c','/dird_1/dird_2/dird_3/dird_4/file_7d',
--'/dir_1/dir_2/dir_3e','/dird_1/dird_2/dird_3/dird_4/file_7f',
--'/a','/a/c','/a/a','/sdfsd','/a/a/a','/dir_1/dir_2/dir_3',
--'/dir_1/dir_2/dir_3a','/dird_1/dird_2/dird_3/dird_4/file_7','/dir_1/dir_2/dir_3b','/dird_1/dird_2/dird_3/dird_4/file_7',
--'/dir_1/dir_2/dir_3c','/dird_1/dird_2/dird_3/dird_4/file_7d',
--'/dir_1/dir_2/dir_3e','/dird_1/dird_2/dird_3/dird_4/file_7f',
'/dir_1/dir_2/dir_2','/dird_1/dird_2/dird_3/dird_4/file_7g',
'/dir_1/dir_2/dir_3','/dird_1/dird_2/dird_3/dird_4/file_7');
END;


FUNCTION genFRCEntryT(n PLS_INTEGER )
	RETURN FRCEntryT
	deterministic
AS
	gstat guidStatO:=guidStatO(133,SYSTIMESTAMP,'checksum',1,SYSTIMESTAMP);
	surlStats surlEntryT:=surlEntryT();
	permission permissionO;
	lfnstat lfnstatO:=lfnstatO(133,SYSTIMESTAMP,SYSTIMESTAMP,'d',' ',10000);
	lfn VARCHAR2(50);
	ret FRCEntryT:=FRCEntryT();
BEGIN

  DEBUGON();
	permission:=
	PermissionO('egee','egee-groupsss',
					63,--bin_to_num(1,1,1,1,1,1),
					0,--bin_to_num(0,1,1,1,0,0),
					0,--bin_to_num(0,0,0,0,0,0),
					ACLList(
							ACLEntryO('nobody',0)
							 )
			 		);
	surlStats.EXTEND(2);
	FOR i IN 1..2 LOOP
		surlStats(i) :=
			surlEntryO('new_surl_i'||i,SYSTIMESTAMP,SYSTIMESTAMP,'newSe_'||i,0);
	END LOOP;

	lfn:=sys_guid;

--  select *
--  bulk collect into ret
--  from (
--  select a.*, rownum rnum from
--  (
	SELECT
	FRCEntryO(
	f.guid,
	guidstatO(f.file_size,f.ctime,f.file_checksum,f.status,f.mtime),
	NULL,
	permissionO(
      Permission_Util.getPrincipalNameStr(f.principal_id),
      Permission_Util.getPrincipalNameStr(f.group_id),
      userperm,groupperm,otherperm,	ACLList(
							ACLEntryO('nobodyX',100)
							 )),
      SUBSTR(SYS_CONNECT_BY_PATH(NAME,'/'),3),
      lfnstatO(f.file_size,f.ctime,f.mTime, f.type,null,f.lfn_lifetime)
    )
    BULK COLLECT INTO ret
 	FROM
	FC_FILE f  --where guid in (select distinct trim(guid) from rcs r where rownum<n*10)
	START WITH file_id=1 --and name=getpathtoken(1,'/big1')
	CONNECT BY PRIOR file_id = parent_id AND file_id<=n
--  ) a
--
--   where rownum <= 10000+n )
--   where rnum >= 10000
;


	ret(1):=	(
		FRCEntryO
			(
				lfn, --guid
				gstat,
				surlStats ,
				permission ,
				'/entry-'||lfn,
				lfnStat
			)
		);
	ret(2):=	(
		FRCEntryO
			( 'lala',
				NULL,
				NULL ,
				NULL ,
				'/big1/dir_1/dir_2/dir_3/dir_4/dir_5/file_3',
				NULL
			)
		);
	ret(3):=	(
		FRCEntryO
			( 'lala',
				NULL,
				NULL ,
				NULL ,
				'/dupadupa/dir_1/dir_2/dir_3/dir_4/dir_5/file_3',
				NULL
			)
		);

   ret(4):=	(
		FRCEntryO
			(
				lfn,
				gstat,
				surlStats ,
				permission ,
				'/',
				lfnStat
			)
		);

	RETURN ret;
END;


FUNCTION genSurls(n PLS_INTEGER)
   RETURN surlT
AS
ret surlT;
BEGIN
  DEBUGON();
   SELECT DISTINCT surl
   BULK COLLECT INTO ret
   FROM rcs s WHERE guid IN
   (SELECT a.* FROM (SELECT DISTINCT guid FROM rcs r) a WHERE ROWNUM<n) GROUP BY surl;
--     select  column_value
--   bulk collect into ret
--   from ww where rownum<n;
--   ret(1):='dupa dupa dupa';
--   null;null;
   RETURN ret;
END;


FUNCTION	getPathToken(lev IN INTEGER,path IN VARCHAR2)
	RETURN FC_FILE.NAME%TYPE
	DETERMINISTIC
AS
ending INTEGER;
starting INTEGER;
BEGIN
	 IF lev = 0 THEN
	 	RETURN ':';
	END IF;
	ending:=INSTR(path,'/',1,lev+1);
	starting:=INSTR(path,'/',1,lev);
	IF ending=0 AND starting<>0 THEN
	ending:=LENGTH(path)+1;
	END IF;

RETURN	SUBSTR(path,starting+1,ending-starting-1);
END;

------------------------------------------------------
-- pseudo regexp for asterisk
FUNCTION	getPathTokenEx(lev IN INTEGER,path IN VARCHAR2)
	RETURN FC_FILE.NAME%TYPE
	DETERMINISTIC
AS
ending INTEGER;
starting INTEGER;
ret FC_FILE.NAME%TYPE;
BEGIN

	ending:=INSTR(path,'/',1,lev+1);
	starting:=INSTR(path,'/',1,lev);
	IF ending=0 AND starting<>0 THEN
	ending:=LENGTH(path)+1;
	END IF;
  ret := SUBSTR(path,starting+1,ending-starting-1) ;
  ret:=REPLACE(ret,'*','%');
RETURN ret;
END;


FUNCTION getpathtokenEx2(lev IN INTEGER,path IN VARCHAR2, NAME IN VARCHAR2)
	RETURN FC_FILE.NAME%TYPE
	DETERMINISTIC
AS
ending INTEGER;
starting INTEGER;
ret FC_FILE.NAME%TYPE;
BEGIN

	ending:=INSTR(path,'/',1,lev+1);
	starting:=INSTR(path,'/',1,lev);
	IF ending=0 AND starting<>0 THEN
	ending:=LENGTH(path)+1;
	END IF;
  ret := SUBSTR(path,starting+1,ending-starting-1) ;
  ret:=REPLACE(ret,'*','%');
  DBMS_OUTPUT.PUT_LINE('n:'||NAME || ',  l: ' || lev ||', r:'||ret);
RETURN ret;
END;


FUNCTION reversePath(path IN VARCHAR2)
	RETURN VARCHAR2
	DETERMINISTIC
AS
ending PLS_INTEGER;
starting PLS_INTEGER;
k PLS_INTEGER:=1;
ret VARCHAR2(32000);
BEGIN

	starting:=INSTR(path,'/',-1,k+1);
   ending:=INSTR(path,'/',-1,k);
   LOOP

      k:=k+1;
      ret:=ret||SUBSTR(path,starting,ending-starting);
      starting:=INSTR(path,'/',-1,k+1);
      ending:=INSTR(path,'/',-1,k);
      EXIT WHEN starting=0;
   END LOOP;

RETURN ret;
END;


procedure transformFcDataDir(
  inFc in out fc_T,
  fc_sRec in out fc_R,
  inAcl in out acl_T,
  tmpPermission in out PermissionSimpleO,
  aclMarkers in out triplesT,
  aclMarker in out tripleR,
  aclCounter in out PLS_INTEGER,
  inputidx in pls_integer,
  inFCcounter in pls_integer,
  name in fc.name%type,
  fileId in fc.file_id%type,
  parentId in fc.parent_id%type,
  principalId in fc.principal_id%TYPE,
  timings in timestamp
)
as


begin

				fc_sRec.parent_id:=parentId;
				fc_sRec.file_id:=fileId;
				--fc_sRec.file_id:=;
				fc_sRec.NAME:= name;
				fc_sRec.type:='d'; -- dir always
				fc_sRec.guid:=generateGuid; --TODO: think of delegating it to java
				fc_sRec.cTime:=timings;
				fc_sRec.mTime:=timings;
				fc_sRec.file_size:=0;
				fc_sRec.lfn_lifetime:=0;
				fc_sRec.file_checksum:=' ';
				fc_sRec.status:=0;
        --perms
        --use user credentials

        fc_sRec.principal_id:=  principalId;
        fc_sRec.group_id :=  tmpPermission.group_id;
        fc_sRec.userperm :=  tmpPermission.userperm;
        fc_sRec.groupperm:=  tmpPermission.groupperm;
        fc_sRec.otherperm:=  tmpPermission.otherperm;
        --handle acls in the acl section
				fc_sRec.dangling_ok:=0;

        inFc(inFCcounter):=fc_sRec;  --add lfn to the story

        -----------------------
        -- handle acl's -------
        BEGIN --acls

          IF  tmpPermission.acl IS NOT NULL AND tmpPermission.acl.COUNT>0 THEN
             aclMarker.a:=aclCounter+1;
             aclMarker.b:=aclCounter+tmpPermission.acl.COUNT;
             aclMarker.c:=inputIdx;
             aclMarkers(fc_sRec.guid):=aclMarker;
              FOR ac IN 1..tmpPermission.acl.COUNT LOOP
                aclCounter:=aclCounter+1;
                inAcl(aclCounter).guid:=fc_sRec.guid;
                inAcl(aclCounter).principal_id:=tmpPermission.acl(ac).principal_id;
                inAcl(aclCounter).perm:=tmpPermission.acl(ac).perm;
              END LOOP;
          END IF;

        END; --acls


end;

/* can set chosen perms if copyPermFromParent=2 */
function mkdirParents (
    securityToken IN principalToken,
    userPrincipal PERM_PRINCIPAL.principal_id%TYPE,
    principals PermStringIntPairOT,
    entries IN lfnT,
    copyPermFromParent IN SMALLINT,
    outerrors IN OUT NOCOPY returnOT,
    ignoreExistingLeaf in boolean
          )
RETURN returnO
as

  type tokenLevelData is record (startingLev integer, parentId integer, perm PermissionSimpleO);
  type tokenMemorized is record (checked boolean, idxlist listnum);
  type tokenLevelDataTbl is table of tokenLevelData index by pls_integer;

	TYPE pairsAT IS TABLE OF tokenMemorized INDEX BY VARCHAR2(1024);

	TYPE tokenIndx IS RECORD (a varchar2(250),b INTEGER);
	TYPE tokenIndxT IS TABLE OF tokenIndx INDEX BY pls_integer;

	retToken tokenIndxT;
	retReal  pairsAT;
  memorizedIndex tokenLevelDataTbl;
  tokenTmp tokenLevelData;

  tmpToken tokenMemorized;
	counter varchar2(250);
	--perms check
	permCheck FCRStatusTT;

	permCheck2 FCRStatusT;

	k integer:=1;


	lev integer;
	path varchar2(32000);

  ending INTEGER;
  endingOld integer:=-1;

  -- serializer

  inFc fireman.fc_T;
  fc_sRec fireman.fc_R;
  inAcl fireman.acl_T;

  fcCounter pls_integer:=1;
  fcMarkers fireman.triplesT;
  fcMarker fireman.tripleR;

  --
  timings TIMESTAMP(0);

  --existing stuff, principal
  defaultPermFetched BOOLEAN:=FALSE;
  defaultPermission PermissionO;
  tmpDefaultPermission PermissionSimpleO;

  aclMarkers fireman.triplesT;
  aclMarker fireman.tripleR;
  aclCounter pls_integer:=0;

	/* errors */
	nerrors PLS_INTEGER:=0;
	nbulkErr PLS_INTEGER:=0;
	bulk_errors EXCEPTION;
	PRAGMA EXCEPTION_INIT(bulk_errors, -24381);
  errGuid VARCHAR2(36);

  /* heuristics */
  averageLength number;

	/* return */
	ret returnO:=returnO(0);

begin
  debugOn();

  SELECT token,idx
  bulk collect into  retToken
  FROM TABLE(pipeTokenizedLfnsO2(entries)) i;

--
  tmpToken.idxList:=listNum();
  for i in 1..retToken.count loop

    if retReal.exists(retToken(i).a) then tmpToken:=retReal(retToken(i).a); end if;
    tmpToken.idxList.extend();
    tmpToken.idxList(tmpToken.idxList.last()):=retToken(i).b;
    retReal(retToken(i).a):=tmpToken;
    tmpToken.idxList.delete();
  end loop;


  -- user heuristics to skip repeated subpaths
  -- heuristics is used only when deep path combined with multiple entries are passed
  -- easy way for now:
    -- if more then 5 and when avg length(entry(i)) > 10
--  select
--  avg(fireman.getNoOfPathDelimiters(column_value))
--  collect into averageLength
--  from table(entries) i
--  ;

	SELECT /*+ cardinality (i 10) */
     (
      CAST(MULTISET
      (
      SELECT /* +use_nl(f) +cardinality (i 2) */
          --permisison-
          i.idx,
          CASE connect_by_isleaf
          when 1 THEN
          Permission_Util.isOkAcl(userperm,groupperm,otherperm,
              principal_id,userPrincipal,
              group_id,
              principals,
              guid, --input guid
              permission_util.writeFileOp)
              else null -- no need at all on other levels
              end case,
              file_id,
              parent_id,
              case
              when connect_by_isleaf=1 and copyPermFromParent=1 then
              PermissionSimpleO
                (
                1,group_id,userperm,groupperm,otherperm,
                --permission_util.getacl(guid)
                CAST(MULTISET
                    (
                    SELECT p.principal_id,acl.perm
                      FROM acl,PERM_PRINCIPAL p
                    WHERE guid=f.guid AND p.principal_id=acl.principal_id
                    ) AS ACLSimpleIdList)
                )
                ELSE NULL END
        FROM
        FC_FILE f
--        WHERE LEVEL>fireman.getNoOfPathDelimiters(i.COLUMN_VALUE)-1 --take both but just one permisson
        START WITH parent_id=0 AND file_id=1 --and name=getpathtoken(1,i.lfn)
        CONNECT BY PRIOR file_id = parent_id --!!!! check if 'd' only valid paths..
          AND NAME=fireman.getpathtoken(LEVEL-1,i.COLUMN_VALUE)
        ) AS FCRStatusT)
      )
      BULK COLLECT INTO permCheck
      FROM (select /*+cardinality (z 10) */ rownum idx, column_value from TABLE(entries) z ) i
      ;

	-- for all
	-- compare idx string with checked string, length of arr should be equal to
  timings:=SYSTIMESTAMP;
   <<pathsloopin>>
   for i in 1..permCheck.count loop --loop over all paths
 	 -- for each path, check where is the starting point and deconstruct starting from the last exisiting token
      -- to the end SUBSTR(path,INSTR(path,'/',1,lev)+1);
		-- check if anything is there using permCheck
    fcMarker.a:=fcCounter;
		path:=entries(i);

    -- just a check for the repeating paths in the input with no need for distinct in the in, qry..
    if retReal(path).checked = true then goto endpathsloopin; end if;
    retReal(path).checked:=true;

		-- check what's the starting level for this path and it's parent_id, permission pair using retReal
    if not memorizedIndex.exists(i) then --nobody set token values for this path yet
    -- start browsing from scratch if no other value set
    -- get permission from the last existing father, throw exception if not permitted
    -- remember permission of last existing node for the specific token
      lev:=permCheck(i).count;
      --check existence and rights
      if lev=fireman.getNoOfPathDelimiters(path)+1 then
        --error: already exists of the egiven length
        if ignoreExistingLeaf=true then goto endpathsloopin; end if;
 				nerrors:=nerrors+1;
				outerrors.EXTEND();
				outerrors(nerrors):=returnO
               (entries(i),
               i,
               Permission_Util.fileOrDirAlreadyExistsVal,
               Permission_Util.fileOrDirAlreadyExistsStr);
        goto endpathsloopin;
      end if;
      if permCheck(i)(lev).permitted=0 then
				-- report no permission
        nerrors:=nerrors+1;
				outerrors.EXTEND();
				outerrors(nerrors):=returnO
               (substr(path,1,INSTR(path,'/',1,lev)),
               i,
               Permission_Util.accessDeniedErrVal,
               Permission_Util.accessDeniedErrStr);
        goto endpathsloopin;
      end if;

      tokenTmp.startingLev:=lev;
      tokenTmp.parentId:=permCheck(i)(lev).file_id;

      IF copyPermFromParent=1 THEN
          tokenTmp.perm:=permCheck(i)(lev).Permission;
        ELSE
        --take defaults
          IF defaultPermFetched=FALSE THEN
            ret:=getDefaultPermission(securityToken,
                      defaultPermission,
                      outErrors);--outerrors not really used
            if ret.result<>0 then
              return ret;
            end if;
            tmpDefaultPermission:=PermissionSimpleO(0,0,0,0,0,NULL);
            translatePermtoSimple(defaultPermission,tmpDefaultPermission);
            tokenTmp.perm:=tmpDefaultPermission;
            if principals is not null and principals.count>0
              and principals(1) is not null and principals(1).item2 is not null
            then
              tokenTmp.perm.group_id:=principals(1).item2;
            end if;

            defaultPermFetched:=TRUE;
          ELSE
            tokenTmp.perm:=tmpDefaultPermission;
          END IF; --defaults
        END IF; --copy from parent
      memorizedIndex(i):=tokenTmp;
    else
      --somebody set the starting level already, get the data, starting level, parentId, permission
      lev:=memorizedIndex(i).startingLev;
      tokenTmp.startingLev:=memorizedIndex(i).startingLev;
      -- permission
      tokenTmp.perm:=memorizedIndex(i).perm;
      -- parent_id
      tokenTmp.parentId:=memorizedIndex(i).parentId;
    end if;
    -- for all involved input indices (paths) for a given token remember starting point = level, file_id, permissionO for this level
    -- and add token, file_id, permission to insert array

  <<tokenlooping>>
	for z in (select /*+ cardinality (tbl 10) */ token, subtoken, lev,
    fc_file_id_seq.NEXTVAL newfileId
    FROM TABLE(pipeTokenizedLfns(path,lev)) tbl)
    loop
      -- for the given tokens for a given level to the last one, check what are involved indices
      -- and for all indices set the starting level to the lev, and parent_id, permission pair for the token
      tokenTmp.startingLev:=z.lev+1;

      -- generate directory data
      fireman.transformFcDataDir(
        inFc,
        fc_sRec,
        inAcl,
        tokenTmp.perm,
        aclMarkers,
        aclMarker,
        aclCounter,
        i,
        fcCounter,
        z.subtoken,
        z.newFileId,
        tokenTmp.parentId,
        userPrincipal,
        timings
      );


      -- \generate directory data

      tokenTmp.parentId:=z.newfileId;


      --mark all involved entries with the current level data
      for k in 1..retReal(z.token).idxList.count loop
        --there could be check here if we dont set token data for the 2nd time for a current index
        --if retReal(z.token)(k) <> i then
          memorizedIndex(retReal(z.token).idxList(k)):=tokenTmp;
--          insert into tlog  values(path,z.token,k,retReal(z.token).idxList(k));
        --end if;
      end loop;
  --    commit;
  fcCounter:=fcCounter+1;
	end loop tokenloopin;
  fcMarker.b:=FCcounter; --increased after deconstruction
  fcMarker.c:=i;
  fcMarkers(fc_sRec.guid):=fcMarker;


  <<endpathsloopin>>  --for 'continue' simulation
  null;
  end loop pathsloopin;


  --do inserts, change to sp
    ------------ fc insert -----------
   BEGIN
      FORALL i IN 1..inFc.COUNT save EXCEPTIONS
         INSERT INTO fc
            VALUES inFc(i);
   EXCEPTION
      WHEN bulk_errors THEN
         nbulkErr := SQL%BULK_EXCEPTIONS.COUNT;
         outerrors.EXTEND(nbulkErr);
         FOR i IN 1..nbulkErr LOOP
         errGuid:=inFc(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid;
            outerrors(nerrors+i):=
               returnO
                  (
                  entries(fcMarkers(errGuid).c), --entries always preserve order:1 --wrong!!
                  SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,
                  Permission_Util.generalInternalBulkErrorVal,
                  SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
                  );
            --delete from acl to be ingested
            IF aclMarkers.EXISTS(errGuid) THEN
               inAcl.DELETE(aclMarkers(errGuid).a,aclMarkers(errGuid).b);
            END IF;
         END LOOP;
         nerrors:=nbulkErr+nerrors;
	END; --exception

   ------------ acl insert -----------
   BEGIN
      FORALL i IN 1..inACl.COUNT save EXCEPTIONS
         INSERT INTO acl
            VALUES inAcl(i);
   EXCEPTION
      WHEN bulk_errors THEN
         nbulkErr := SQL%BULK_EXCEPTIONS.COUNT;
         outerrors.EXTEND(nbulkErr);
         FOR i IN 1..nbulkErr LOOP
            errGuid:=inAcl(SQL%BULK_EXCEPTIONS(i).ERROR_INDEX).guid;
            outerrors(nerrors+i):=
               returnO
                  (
                  entries(aclMarkers(errGuid).c),
                  aclMarkers(errGuid).c,
                  Permission_Util.generalInternalBulkErrorVal,
                  SUBSTR(SQLERRM(-SQL%BULK_EXCEPTIONS(i).ERROR_CODE)|| ' at index '||
                     SQL%BULK_EXCEPTIONS(i).ERROR_INDEX,1,1024)
                  );
            -- in the future we may still keep such entry as valid (and not back-delete)
            -- nowadays:
            DELETE FROM fc WHERE guid=errGuid;
         END LOOP;
         nerrors:=nbulkErr+nerrors;
	END; --exception
  -- end of inserts
	return ret;

end;

--SEIndex methods
/*  public String[] listSEbyLFN(String lfn) throws DBException, NotExistsException, AuthorizationException */
function listSEbyLFN(securityToken in principalToken,
          lfn in varchar2,
          outEntries OUT NOCOPY lfnT,
          outerrors out nocopy returnOT)
  return returnO
AS
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

  /* counters */
  j pls_integer;



	/* return */
	ret returnO:=returnO(0);
BEGIN
  outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);

--chck
--select s.surl from rc_surl s where s.guid=( select f.guid from fc_file " +
--                    "f where level=fireman.getNoOfPathDelimitersRoot(:1)+1 start with parent_id=0 " +
--                    "and file_id=1 connect by prior file_id = parent_id and " +
--                    "name = fireman.getpathtoken(level-1,:2) )");

 select
	case permitted
	WHEN 1 THEN  surl
	  ELSE NULL END,
	  case permitted
	  	  WHEN 0 THEN returnO
              (lfn,0,
               Permission_Util.accessDeniedErrVal,  --no right to write
               Permission_Util.accessDeniedErrStr
              )
   		else
   		null
   		end case
      	BULK COLLECT INTO outEntries,outErrors
   		from (
   SELECT /*+ cardinality( i 5 ) cardinality(inP 10) */
        surl,permitted
    FROM rcs join
    (  select guid,
        Permission_Util.isOkAcl(userperm,groupperm,otherperm,
      		principal_id,userPrincipal,
          group_id,
          principals,
          guid,permission_util.readFileOp)
        permitted
        from fc
        	WHERE LEVEL=getNoOfPathDelimitersRoot(lfn)+1
        	START WITH parent_id=0 and file_id=1
          CONNECT BY PRIOR file_id = parent_id
          AND NAME =getpathtoken(LEVEL-1,lfn)
          )
          using (guid))
          ;
    --pamietaj o nie istniejacej sciezce

  j:=outEntries.count;
  if j=0 and outErrors.count=0 then
    outErrors.extend();
    outErrors(1):=returnO
              (lfn,0,
               Permission_Util.NoSuchFileVal, --no parent directory exists
               Permission_Util.NoSuchFileStr
              );
  end if;

    FOR i IN 1..j LOOP
      IF outEntries.exists(i) and outEntries(i) IS NULL THEN
        outEntries.DELETE(i);
      ELSE
        outerrors.DELETE(i);
      END IF;
    END LOOP;

	RETURN ret;

END;


/* public String[] listSEbyGUID(String guid) throws DBException, NotExistsException */
function listSEbyGUID(securityToken in principalToken,
          iguid in varchar2,
          outEntries OUT NOCOPY lfnT,
          outerrors out nocopy returnOT)
  return returnO
as
	/* permissions */
	userPrincipal PERM_PRINCIPAL.principal_id%TYPE;
	principals PermStringIntPairOT;

	/* serializer */
	parentCheck FC_FILE.file_id%TYPE;
  parentCheckParent FC_FILE.file_id%TYPE;
  parentCheckVal PLS_INTEGER;

  j PLS_INteger;

	/* return */
	ret returnO:=returnO(0);
BEGIN
  outerrors:=returnOT();
	Permission_Util.getPrincipalIdsO(securityToken,userPrincipal,principals,0);

  select
	case permitted
	WHEN 1 THEN  surl
	  ELSE NULL END,
	  case permitted
	  	  WHEN 0 THEN returnO
              (iguid,0,
               Permission_Util.accessDeniedErrVal,  --no right to write
               Permission_Util.accessDeniedErrStr
              )
   		else
   		null
   		end case
      	BULK COLLECT INTO outEntries,outErrors
   		from (
   SELECT /*+ cardinality( i 5 ) cardinality(inP 10) */
   (Permission_Util.isOkAcl(userperm,groupperm,otherperm,
      		principal_id,userPrincipal,
          group_id,
          principals,
          guid,permission_util.readFileOp))
        permitted,
        surl
    FROM rcs join fc using (guid)
    where guid=iguid
    )
   ;

  j:=outEntries.count;
  if j=0 and outErrors.count=0 then
    outErrors.extend();
    outErrors(1):=returnO
              (iguid,0,
               Permission_Util.NoSuchFileVal, --no parent directory exists
               Permission_Util.NoSuchFileStr
              );
  end if;

    FOR i IN 1..j LOOP
      IF outEntries.exists(i) and outEntries(i) IS NULL THEN
        outEntries.DELETE(i);
      ELSE
        outerrors.DELETE(i);
      END IF;
    END LOOP;

	RETURN ret;

END;

/* TODO: special version for cycles should be here*/
function getSymlink (path in varchar2)
	return ResolverO
	is
	ret ResolverO;
	rtypef fc.type%type;
	rdata fc.data%type;
  looper fileIdT;
	begin

	    select
        resolverO(0,parent_id,file_id,
            case type when 's' then data else path end,
            guid,type,ctime,mtime,
            file_size,file_checksum,LFN_LIFETIME,status,
            principal_id,group_id,userperm,groupperm,otherperm,
            dangling_ok,
            data,-1)
         ,
         type,
         case type when 's' then data else null end
	    collect into ret, rtypef,rdata
	    from fc
		where level=fireman.getNoOfPathDelimitersRoot(path)+1
		start with parent_id=0 and file_id=1
		connect by
		prior file_id=parent_id
		and name = fireman.getpathtoken(level-1,path)
		;
		if rtypef = 's' then
      looper(ret.file_id):=1;
			ret:=fireman.getSymlinkInt(rdata,looper);
		end if;
		return ret;
  EXCEPTION
      WHEN NO_DATA_FOUND THEN
      ret:=resolverO(0,path);
      RETURN ret;

	end;
  
  
function getSymlinkInt (path in varchar2, looper in out nocopy fileIdT)
	return ResolverO
	is
	ret ResolverO;	
	rdata fc.data%type;
	begin
  debugOn();

	    select
        resolverO(0,parent_id,file_id,
            case type when 's' then data else path end,
            guid,type,ctime,mtime,
            file_size,file_checksum,LFN_LIFETIME,status,
            principal_id,group_id,userperm,groupperm,otherperm,
            dangling_ok,
            data,-1)
         ,         
         case type when 's' then data else null end
	    collect into ret, rdata
	    from fc
		where level=fireman.getNoOfPathDelimitersRoot(path)+1
		start with parent_id=0 and file_id=1
		connect by
		prior file_id=parent_id
		and name = fireman.getpathtoken(level-1,path)
		;
		if ret.ftype = 's' then
      if looper.exists(ret.file_id) then 
        return resolverO(0,0,path,permission_util.loopInSymlinksDetectedVal); 
      end if;
      looper(ret.file_id):=1;
			ret:=fireman.getSymlinkInt(rdata,looper);
		end if;
		return ret;
  EXCEPTION
      WHEN NO_DATA_FOUND THEN
      ret:=resolverO(0,path);
      RETURN ret;

	end;




end;
